/******************************************************************************

  c2c.h
  
  Written by:  Michael Pellegrino
	       Indiana Institute of Technology

  The purpose of this file is to define needed structures and constants for
  c2c communications as seen by the user.

  Date last edited:  August 4, 1995

  Version:  .12

******************************************************************************/
#ifndef C2C_HEADER		/* Protect against multiple definition errors */
#define C2C_HEADER

#ifdef __cplusplus              /* Put in for C++ compilers */
extern "C" {
#endif

/* Define all stream types */
#define S_HEAD_TRACKER		-101	/* Head tracker information */
#define S_WAND_TRACKER		-102	/* Wand tracker information */
#define S_WAND_BUTTON		-103	/* Button information */
#define S_WAND_JOYSTICK		-104	/* Joystick information */
#define S_WORLD_POSITION	-105	/* CAVE location in the world */
#define S_CAVE_USER		-106	/* User information (trackers/buttons etc... all in one) */

/* Define all session types */
#define ST_CAVE			0x01	/* CAVE session */
#define ST_CAVE_SIMULATOR	0x02	/* CAVE simulator session */
#define ST_IMMERSADESK		0x03	/* Immersadesk session */
#define ST_SIMULATION		0x04	/* Data simulation of some sort */

/* Define some data types */
#define DT_CHAR			0x01	/* Character data */
#define DT_INT			0x02	/* Integer data */
#define DT_LONG			0x03	/* Long integer data */
#define DT_FLOAT		0x04	/* Float data */
#define DT_DOUBLE		0x05	/* Double data */

/* Define some machine manufacturers */
#define MT_SGI			0x01	/* Silicon Graphics */
#define MT_IBM			0x02	/* I.B.M. */
#define MT_SUN			0x03	/* Sun */
#define MT_NEXT			0x04	/* NeXT */
#define MT_HP			0x05	/* Hewlett-Packard */
#define MT_DEC			0x06	/* Digital Equipment Corporation */
#define MT_CRAY			0x07	/* Cray */
#define MT_INTEL		0x08	/* Intel */
#define MT_THINK_MACHINE	0x09	/* Thinking Machines Corporation */

/* Define machine models */
#define MM_ONYX			0x01	/* SGI Onyx */
#define MM_INDY			0x02	/* SGI Indy */
#define MM_IRIS			0x03	/* SGI Iris */
#define MM_INDIGO		0x04	/* SGI Indigo */
#define MM_CHALLENGE		0x05	/* SGI POWER Challenge */

/* Define some operating system types */
#define OS_IRIX_53		0x01	/* SGI Irix 5.3 */
#define OS_IRIX_52		0x02	/* SGI Irix 5.2 */
#define OS_AIX_41		0x03	/* IBM AIX 4.1 */
#define OS_AIX_32		0x04	/* IBM AIX 3.2 */
#define OS_SUNOS_41		0x05	/* SunOS 4.1 */

/* Define some constants for drawing a CAVE
   Multiple values can be used by logically ORing the constants together */
#define DC_USER			0x01	/* Draw the user */
#define DC_WALLS		0x02	/* Draw the walls of the environment */
#define DC_CAVE			0x04	/* Draw a CAVE environment */
#define DC_IDESK		0x08	/* Draw an Immersadesk environment */

/* Define all error messages as seen by user */
#define E_SUCCESS		0	/* No error at all */
#define E_BROKER_ATTACH		-1	/* Attach to the broker failed */
#define E_SUBSCRIBE_NO_STREAM	-2	/* Subscribe to non-existant stream */
#define E_INVALID_HOSTID	-3	/* Bad host id given */
#define E_INVALID_CALLBACK	-4	/* Bad callback given */
#define E_INVALID_STREAM	-5	/* Stream never registeged */
#define E_STREAM_REGISTERED	-6	/* Stream already registered */
#define E_INVALID_HOST		-7	/* Invalid host id given */
#define E_INVALID_CAVEID	-8	/* Invalid CAVE id given */
#define E_INVALID_BROKER_URL    -9      /* Invalid broker URL given */
#define E_INVALID_CLIENT        -10     /* Invalid client name given */
#define E_INVALID_SESSION       -11     /* Invalid session name given */
#define E_CLIENT_EXISTS         -12     /* Client already registered on broker */
#define E_SESSION_EXISTS        -13     /* Session already created on broker */
#define E_OPEN_CONFIG_FILE      -14     /* Error opening configuration file */

/* Various array sizes */
#define C2C_NAME_SIZE		128	/* Size of session string */
#define C2C_URL_SIZE		128	/* Size of a URL string */
#define C2C_PATH_SIZE		128	/* Size of path string */
#define C2C_EXE_SIZE		128	/* Size of exe string */
#define C2C_ARGS_SIZE		128	/* Size of argument string */

/* Miscellaneous constant definitions */
#define C2C_CONFIG_FILE ".c2cConfig"	/* Default config file to open if none is given */
#define CONFIG_FILE_LINE_SIZE	1024	/* Max size of a file line in configuration file */
#define CONFIG_OPTION_SIZE	512	/* Max size of an option in configuration file */
#define C2C_SUBSCRIBE		0	/* Flag for world subscribe operation */
#define C2C_UNSUBSCRIBE		1	/* Flag for world unsubscribe operation */

/* Miscellaneous type definitions */
typedef int  CAVEId;		/* Identifier for a CAVE */
typedef int  HostId;		/* Identifier for a session broker */
typedef int  SessionId;	/* Identifier of a session in broker database */
typedef int  StreamType;	/* Identifier for stream types */
typedef char URL[C2C_URL_SIZE];	/* URL for various hosts */

/* Structure describing a CAVE session */
typedef struct cavesession
{
  CAVEId owner;		        /* Who owns the session on the broker */
  char name[C2C_NAME_SIZE];	/* Unique name describing the session */
  char pathname[C2C_PATH_SIZE];	/* Where to find the application */
  char execname[C2C_EXE_SIZE];	/* Executable name */
  char args[C2C_ARGS_SIZE];	/* Command line arguments */
  int id;                       /* Id of the session */
}CaveSession;

/* Structure defining CAVE environment */
typedef struct caveenv
{
  char name[C2C_NAME_SIZE];	/* Unique name for this CAVE */
  int type;			/* Type of session this is (CAVE, I-Desk, etc...) */
  int id;			/* Id of the client */
}CaveEnv;

/* Structure defining CAVE user */
typedef struct caveuser
{
  float headx,heady,headz,	/* Head tracker location */
        heada,heade,headr,	/* Head tracker orientation */
	wandx,wandy,wandz,	/* Wand tracker location */
	wanda,wande,wandr,	/* Wand tracker orientation */
	worldx,worldy,worldz,	/* World location */
	worlda,worlde,worldr;	/* World orientation */
  int   but1,but2,		/* Button information */
        but3,but4;
  float joyx,joyy;		/* joystick information */
}CAVEUser;

/* Structure defining a CAVE-to-CAVE session */
typedef struct c2cconfig
{
  URL broker;			/* URL to connect to broker with */
  CaveEnv env;			/* Local environment parameters */
  CaveSession session;		/* Session info for this CAVE app */
}c2cConfig;

/* Structure for data stream processing */
typedef struct c2cbuffer c2cbufferptr;

typedef struct c2cbuffer
{
  int type;			/* Type of the data being sent */
  int size;			/* Size of the data */
  void *data;			/* Pointer to the data */
  c2cbufferptr *next;		/* Next node in the list */
}c2cBuffer;

/* Function pointer definitions for callbacks */
typedef void (*TrackerP)(float,float,float,float,float,float,int); /* Locations,Angles, and sender id */
typedef void (*ButtonP)(int,int,int,int,int); /* S_WAND_BUTTON callback */
typedef void (*JoystickP)(float,float,int); /* S_WAND_JOYSTICK callback */
typedef void (*CAVEUserP)(CAVEUser *,int); /* S_CAVE_USER callback */
typedef void (*UserDataP)(void *,int);	 /* User data callback */
typedef void (*StreamP)(int,int); /* Stream callback */

/* CAVE-2-CAVE library function prototypes */
extern int	c2cBrokerAttach(URL url);
extern int	c2cBrokerDetach(HostId host);
extern int	c2cBrokerKill(HostId host);
extern int	c2cDrawUser(char *user);
extern int	c2cGetClients(HostId host,int *num_clients,CaveEnv **client_list);
extern int      c2cGetDebugLevel(void);
extern int	c2cGetSessions(HostId host,int *num_sessions,CaveSession **session_list);
extern int	c2cReadConfigFile(char *filename,c2cConfig *config);
extern int	c2cRegisterStream(StreamType stream,void *callback);
extern int	c2cSendStream(c2cBuffer **buffer,StreamType stream);
extern int	c2cSendTracker(float x,float y,float z,float a,float e,float r,int stream);
extern int	c2cSendUser(CAVEUser *user);
extern int	c2cSendWandButtons(int b1,int b2,int b3,int b4);
extern int	c2cSendWandJoystick(float joyx,float joyy);
extern int	c2cSessionAttach(HostId host,SessionId session, CAVEId cave);
extern int	c2cSessionDetach(HostId host,CAVEId cave);
extern int	c2cSetAutoFrequency(StreamType stream,int action);
extern int	c2cSetSkipFrequency(StreamType stream,int frequency);
extern int	c2cSubscribe(HostId host,CAVEId datasource,StreamType stream,void *callback);
extern int	c2cTrackUserInit(char *user,void *callback);
extern int	c2cTrackUserExit(char *user,void *callback);
extern int	c2cUnregister(HostId host,CAVEId cave);
extern int	c2cUnregisterStream(StreamType stream);
extern int	c2cUnsubscribe(HostId host,CAVEId datasource,StreamType stream,void *callback);
extern int	c2cWorldSubscribe(char *user,StreamType stream,void *callback,int action);
extern void	c2cCAVECast(void *user_args);
extern void	c2cCAVEMan(float Hx,float Hy,float Hz,float Ha,float He,float Hr,
                           float wx,float wy,float wz,float wa,float we,float wr);
extern void	c2cDrawAllUsers(void);
extern void	c2cDrawSomeUsers(int count,char *users[]);
extern void	c2cDrawStickMan(CAVEUser *user,CAVEId id);
extern void	c2cFreeGetBuffer(void *buffer);
extern void	c2cFreePackBuffer(c2cBuffer **buffer);
extern void	c2cGetChar(void *buffer,char *data,int size);
extern void	c2cGetDouble(void *buffer,double *data,int size);
extern void	c2cGetFloat(void *buffer,float *data,int size);
extern void	c2cGetInt(void *buffer,int *data,int size);
extern void	c2cGetLong(void *buffer,long *data,int size);
extern void	c2cInit(int *argc, char *argv[],CaveEnv *env);
extern void	c2cInitPackBuffer(c2cBuffer **buffer);
extern void	c2cKillSession(HostId host,SessionId session,CAVEId cave);
extern void	c2cPackBuffer(c2cBuffer **buffer,void *data,int length,int type);
extern void     c2cPrintf(int level, char *format, ...);
extern void     c2cSetDebugLevel(int level);
extern void     *c2cSharedAlloc(int size);
extern void     c2cSharedExit(void);
extern void     c2cSharedFree(void *ptr);
extern void     c2cSharedInit(void);
extern void	c2cTerminate(void);
extern void     c2cTrackUserCallback(CAVEUser *user,int id);
extern void     c2cWorldDataInit(void);
extern void	c2cWorldExit(void);
extern void	c2cWorldInit(int *argc,char *argv[]);
extern CAVEId   c2cGetClientId(HostId host,char *name);
extern CAVEId	c2cRegister(HostId host,CaveEnv *env);
extern CAVEId   c2cWorldGetClientId(char *name);
extern SessionId c2cGetSessionId(HostId host,char *name);
extern SessionId c2cSessionCreate(HostId host,CaveSession *ses);
extern SessionId c2cWorldGetSessionId(char *name);

/* Define some useful macros */
#define c2cSendHeadTracker(x,y,z,a,e,r) \
	{ \
	  c2cSendTracker(x,y,z,a,e,r,S_HEAD_TRACKER); \
	} \

#define c2cSendWandTracker(x,y,z,a,e,r) \
	{ \
	  c2cSendTracker(x,y,z,a,e,r,S_WAND_TRACKER); \
	} \

#define c2cSendWorldPosition(x,y,z,a,e,r) \
	{ \
	  c2cSendTracker(x,y,z,a,e,r,S_WORLD_POSITION); \
	} \

#define c2cWorldDataSubscribe(user,stream,callback) \
	{ \
	  c2cWorldSubscribe(user,stream,callback,C2C_SUBSCRIBE); \
	} \

#define c2cWorldDataUnsubscribe(user,stream,callback) \
	{ \
	  c2cWorldSubscribe(user,stream,callback,C2C_UNSUBSCRIBE); \
	} \

#define c2cSetHeadFrequency(frequency) \
	{ \
	  c2cSetSkipFrequency(S_HEAD_TRACKER,frequency); \
	}

#define c2cSetWandFrequency(frequency) \
	{ \
	  c2cSetSkipFrequency(S_WAND_TRACKER,frequency); \
	}

#define c2cSetButtonFrequency(frequency) \
	{ \
	  c2cSetSkipFrequency(S_WAND_BUTTON,frequency); \
	}

#define c2cSetJoystickFrequency(frequency) \
	{ \
	  c2cSetSkipFrequency(S_WAND_JOYSTICK,frequency); \
	}

#define c2cSetWorldFrequency(frequency) \
	{ \
	  c2cSetSkipFrequency(S_WORLD_POSITION,frequency); \
	}

#define c2cSetUserFrequency(frequency) \
	{ \
	  c2cSetSkipFrequency(S_CAVE_USER,frequency); \
	}

#define c2cPackChar(buffer,data,size) \
	{ \
	  c2cPackBuffer(buffer,data,size,DT_CHAR); \
	}

#define c2cPackInt(buffer,data,size) \
	{ \
	  c2cPackBuffer(buffer,data,size,DT_INT); \
	}

#define c2cPackLong(buffer,data,size) \
	{ \
	  c2cPackBuffer(buffer,data,size,DT_LONG); \
	}

#define c2cPackFloat(buffer,data,size) \
	{ \
	  c2cPackBuffer(buffer,data,size,DT_FLOAT); \
	}

#define c2cPackDouble(buffer,data,size) \
	{ \
	  c2cPackBuffer(buffer,data,size,DT_DOUBLE); \
	}

/* Declare CAVE-to-CAVE library global variables */
extern CaveEnv c2cEnv;				/* Local environment */

#ifdef __cplusplus                      /* Put in for C++ compilers */
}
#endif

#endif			/* Protection against multiple definition errors */
