/******************************************************************************

  c2c.c
  
  Written by:  Michael Pellegrino
	       Indiana Institute of Technology

  The purpose of this file is to provide the user interface for c2c
  communications.

  Date last edited:  August 4, 1995

  Version:  .12
  
******************************************************************************/
#include <malloc.h>
#include <nexus.h>
#include <stdio.h>
#include <string.h>
#include <c2c.h>
#include "c2cdevice.h"

/* Declare all global variables */
nexus_node_t *c2cNodes;				/* Local nexus nodes allocated */
int c2cNumNodes;				/* Number of nexus nodes running */
Conntab c2cBrokers;				/* Everyone connected */
Castlist *c2cDataCast;				/* Cast list for all streams */
CastControl c2cCastControl;			/* Control structure for CAVE casting */
TrackUser **c2cUserList;			/* List of all users to track */
Streamlist *c2cStreams;				/* Stream callback list for all streams */
Handler hRegister;				/* Register handler */
Handler hSession;				/* Session query handler */
Handler hUser;					/* Client query handler */
Handler hNewSession;				/* Sesssion create handler */
Handler hSubscribe;				/* Subscribe handler */
Handler hUnregister;				/* Unregister handler */
Handler hUnsubscribe;				/* Unsubscribe handler */
CaveEnv c2cEnv;					/* Local environment */
CaveSession c2cSession;				/* Local session */
c2cWorldInfo c2cWorld;				/* World information structure */
int c2cDebugLevel;                              /* Debugging information value */

/*****************************************************************************  
  
          Params:  Argument count
                   Argument list
		   Name of session
		   Type of session
         Returns:  Nothing
  Global Effects:  None

  This function initializes all c2c functionality
 
******************************************************************************/
void c2cInit(int *argc, char *argv[],CaveEnv *env)
{ /* c2cInit */
  unsigned short attachport = 0;	/* Port to connect to in future */
  char *attachhost;			/* Name of host where port is */
  int i;                                /* Loop counter */

  c2cSetDebugLevel(0);                  /* Turn off debugging */
  c2cPrintf(110,"I am %s\n",env->name);
  c2cPrintf(110,"begin processing c2c arguments\n");
  for(i=0;i<*argc;i++)                   /* Parse arguments */
  { /* For */
    c2cPrintf(120,"processing argument %s\n",argv[i]);
    if(strcmp(argv[i],"-c2cdbg") == 0)   /* Match on the debugging arg */
    { /* If */
      c2cSetDebugLevel(atoi(argv[++i])); /* Assign the level */
      c2cPrintf(120,"printing debugging values up to %d\n",c2cDebugLevel);
    } /* If */
  } /* For */
  c2cPrintf(110,"finished processing c2c arguments\n");
  c2cBrokers.count = 0;			/* Initialize broker table */
  c2cBrokers.nextid = 0;
  c2cBrokers.addr = NULL;
  c2cStreams = NULL;			/* Initialize stream list */
  c2cDataCast = NULL;			/* Initialize cast list */
  memcpy(&c2cEnv,env,sizeof(CaveEnv)); /* Copy data into local structure */
  c2cPrintf(110,"initializing nexus transport layer\n");
  nexus_init(argc,&argv,"NEXUS_ARGS","nx",NULL,NULL,NULL,NULL,
             &c2cNodes,&c2cNumNodes);	/* Initialize all nexus stuff */
  c2cPrintf(110,"starting nexus transport layer\n");
  nexus_start();			/* Start all nexus stuff */
  c2cPrintf(110,"allowing attachments\n");
  nexus_allow_attach(&attachport,&attachhost,NexusAllowAttachments);
} /* c2cInit */

/*****************************************************************************  

          Params:  Nothing
         Returns:  Nothing
  Global Effects:  None

  All c2c functionality is terminated
  
******************************************************************************/
void c2cTerminate(void)
{ /* c2cTerminate */
  NexusAddr *addr;			/* Work pointer */

  c2cPrintf(110,"starting c2c termination\n");
  c2cPrintf(110,"killing all broker connections\n");
  for(addr = c2cBrokers.addr;		/* For everyone connected */
      addr != NULL;
      addr = addr->next)
  { /* For loop */
    c2cPrintf(120,"killing a connection\n");
    nexus_destroy_global_pointer(&addr->gp); /* Kill global pointer */
  } /* For loop */
/*  nexus_destroy_current_context(NEXUS_FALSE);	 Kill context for clean exit */
} /* c2cTerminate */

/*****************************************************************************  

          Params:  A Broker structure
         Returns:  Updated Broker structure
  Global Effects:  None

  This function attaches to the broker listed in the structure.

******************************************************************************/
int c2cBrokerAttach(URL url) 
{ /* c2cBrokerAttach */
  nexus_global_pointer_t gp;            /* GP buffer */
  NexusAddr *addr;			/* Work node */
  URL base = "x-nexus";			/* Base nexus URL */
  char *ipinfo;				/* Pointer to just IP name in URL */
  int stat;                             /* Status of the broker attach */

  ipinfo = url + strlen("c2cBroker");	/* Point to IP stuff only */
  strcat(base,ipinfo);	/* Remove c2c reference and replace with nexus reference */
  c2cPrintf(110,"attempting to attach to broker %s\n",base);
  stat = nexus_attach_with_url(base,&gp); /* Connect to broker */
  c2cPrintf(110,"result of attachment to broker: %d\n",stat);
  if(stat == 0)
  { /* If */
    c2cBrokers.count++;			/* Count the new entry */
    addr = (NexusAddr *)malloc(sizeof(NexusAddr));
    addr->next = c2cBrokers.addr;	/* Insert into list */
    c2cBrokers.addr = addr;
    addr->rank = c2cBrokers.nextid++;	/* Set the rank of */
    memcpy(&addr->gp,&gp,sizeof(nexus_global_pointer_t)); /* Copy global pointer in */
    c2cPrintf(110,"inserted into broker list with an id of %d\n",addr->rank);
    return(addr->rank);			/* Send back host id */
  } /* If */
  return(E_BROKER_ATTACH);              /* Send back error message */
} /* c2cBrokerAttach */

/*****************************************************************************  

          Params:  A Broker structure
         Returns:  Nothing
  Global Effects:  None

  This function detaches from the broker given by id.

******************************************************************************/
int c2cBrokerDetach(HostId id) 
{ /* c2cBrokerDetach */
  NexusAddr *curaddr,*prevaddr;		/* Work nodes */

  c2cPrintf(110,"attempting to detach from broker\n");
  prevaddr = c2cBrokers.addr;		/* Point to beginning of table */
  if(prevaddr != NULL)                  /* Any data in list */
  { /* If */
    if(prevaddr->rank == id)		/* First in list */
    { /* If */
      c2cBrokers.count--;		/* Remove one from the list */
      c2cBrokers.addr = prevaddr->next;	/* Update links */
      curaddr = prevaddr;                 /* Point to address for one memory deletion routine */
    } /* If */
    else					/* Traverse the table list */
    { /* Else */
      for(curaddr = prevaddr->next;	/* Prepare for list traversal */
          (curaddr != NULL) && (curaddr->rank != id);
          curaddr = curaddr->next, prevaddr = prevaddr->next);
      if(curaddr != NULL)		/* Found it-->erase it */
      { /* If */
        c2cBrokers.count--;		/* Remove one from the list */
        prevaddr->next = curaddr->next;	/* Remove links from table */
      } /* If */
      else				/* Not found in list */
      { /* Else */
        c2cPrintf(110,"host %d not found in the broker list\n",id);
        return(E_INVALID_HOST);         /* Host not found */
      } /* Else */
    } /* Else */
    c2cPrintf(110,"removing host %d from the broker list\n",id);
    nexus_destroy_global_pointer(&curaddr->gp); /* Kill all contact to broker */
    free(curaddr);			/* Free memory */
    return(E_SUCCESS);                  /* Return success */
  } /* If */
  else                                  /* Not found in list */
  { /* Else */
    c2cPrintf(110,"host %d not found in the broker list\n",id);
    return(E_INVALID_HOST);         /* Host not found */
  } /* Else */
} /* c2cBrokerDetach */

/*****************************************************************************  

          Params:  A Broker structure
         Returns:  Nothing
  Global Effects:  None

  This function kills a particular broker

******************************************************************************/
int c2cBrokerKill(HostId id) 
{ /* c2cBrokerKill */
  nexus_buffer_t buffer;		/* Buffer for data */
  NexusAddr *addr;			/* Address of broker to kill */

  c2cPrintf(110,"attempting to kill broker\n");
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != id);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found it-->erase it */
  { /* If */
    c2cPrintf(120,"sending message to broker to die\n");
    nexus_init_remote_service_request(&buffer,&addr->gp,
				      BROKER_KILL_STRING,
				      BROKER_KILL_HASH);
    nexus_send_remote_service_request(&buffer); /* Send along to broker */
    c2cBrokerDetach(id);		/* Remove from broker list */
  } /* If */
  else				/* Not found in list */
  { /* Else */
    c2cPrintf(110,"host %d not found in the broker list\n",id);
    return(E_INVALID_HOST);           /* Invalid host given */
  } /* Else */
  c2cPrintf(110,"killed the host %d\n",id);
  return(E_SUCCESS);                    /* Return success */
} /* c2cBrokerKill */

/*****************************************************************************  

          Params:  Broker
	           CAVE environment
         Returns:  CAVE id for this broker
  Global Effects:  None

  This function informs session manager of your presence and capabilities and
  the session manager registers this info in its database.
  
******************************************************************************/
CAVEId c2cRegister(HostId host,CaveEnv *env)
{ /* c2cRegister */
  nexus_buffer_t buffer;		/* Data buffer */
  nexus_global_pointer_t gp;		/* Client id */
  int ngp;				/* Size of global pointer */
  NexusAddr *addr;			/* Work pointer */
  CAVEId id;				/* Id of client on broker */

  c2cPrintf(110,"attempting to register as %s on %d\n",env->name,host);
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending registration information to host\n");
    NexusInit(hRegister);
    nexus_global_pointer(&gp,NULL);
    nexus_init_remote_service_request(&buffer,&addr->gp,
		                      BROKER_REGISTER_STRING,
				      BROKER_REGISTER_HASH);
    nexus_set_buffer_size(&buffer,nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp) +
  			  nexus_sizeof_char(&buffer,C2C_NAME_SIZE) +
			  nexus_sizeof_int(&buffer,2),
			  ngp + 3);
    nexus_put_global_pointer(&buffer,&gp,1);
    nexus_put_char(&buffer,env->name,C2C_NAME_SIZE);
    nexus_put_int(&buffer,&env->type,1);
    nexus_put_int(&buffer,&env->id,1);
    nexus_send_remote_service_request(&buffer); /* Send along to broker */
    c2cPrintf(120,"waiting for broker to respond\n");
    NexusWait(hRegister);			/* Wait for a response from broker */
    nexus_get_stashed_int(&hRegister.buffer,&id,1); /* Read the id */
    nexus_free_stashed_buffer(&hRegister.buffer);	/* Free all stash memory */
    c2cPrintf(110,"response from the broker:  %d\n",id);
  } /* If */
  else					/* Host not found in list */
  { /* Else */
    c2cPrintf(110,"host %d is not in the broker list\n",host);
    id = E_INVALID_HOST;                /* Invalid host given */
  } /* Else */
  return(id);				/* Send id back */
} /* c2cRegister */

/*****************************************************************************  

          Params:  Broker
	           CAVE environment
         Returns:  CAVE id for this broker
  Global Effects:  None

  This function informs session manager to remove me from the client database.
  
******************************************************************************/
int c2cUnregister(HostId host,CAVEId cave)
{ /* c2cUnregister */
  nexus_buffer_t buffer;		/* Data buffer */
  nexus_global_pointer_t gp;		/* Local global pointer buffer */
  int ngp;				/* Size of local global pointer */
  NexusAddr *addr;			/* Work pointer */
  int stat;                             /* Unregister status */

  c2cPrintf(110,"attempting to unregister on host %d\n",host);
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending unregister message to the broker\n");
    NexusInit(hUnregister);             /* Initialize reply */
    nexus_global_pointer(&gp,NULL);	/* Get local id */
    nexus_init_remote_service_request(&buffer,&addr->gp, /* Send request */
				      BROKER_UNREGISTER_STRING,
				      BROKER_UNREGISTER_HASH);
    nexus_set_buffer_size(&buffer,nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp) +
			  nexus_sizeof_int(&buffer,1),
			  ngp + 2);
    nexus_put_global_pointer(&buffer,&gp,1);	/* Pack the buffer */
    nexus_put_int(&buffer,&cave,1);
    nexus_send_remote_service_request(&buffer); /* Send along to broker */
    c2cPrintf(120,"waiting to hear from thr broker\n");
    NexusWait(hUnregister);			/* Wait for reply from broker */
    nexus_get_stashed_int(&hUnregister.buffer,&stat,1); /* Read reply status */
  } /* If */
  else					        /* Host not found in list */
  { /* Else */
    stat = E_INVALID_HOST;              /* Invalid host given */
  } /* Else */
  c2cPrintf(110,"returning from unregistration with status of %d\n",stat);
  return(stat);                         /* Return status of operation */
} /* c2cUnregister */

/*****************************************************************************  

          Params:  Host to query
	           Number of clients in the database
		   Array of the clients environments
         Returns:  Number of clients
	           Array of the clients environments
  Global Effects:  None

  This function calls the specified broker and retrieves the capabilities of
  all CAVES in the database.
  
******************************************************************************/
int c2cGetClients(HostId host,int *num_clients,CaveEnv **clientlist)
{ /* c2cGetClients */
  nexus_buffer_t buffer;			/* Message buffer */
  nexus_global_pointer_t gp;			/* Reply gp buffer */
  int ngp;					/* Size of global pointer */
  int i;					/* Loop counter */
  CaveEnv *env;					/* Pointer to incoming data */
  NexusAddr *addr;				/* Work pointer */

  c2cPrintf(110,"attempting to query clients on host %d\n",host);
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)				/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending query message to broker\n");
    NexusInit(hUser);             /* Initialize reply */
    nexus_global_pointer(&gp,NULL);		/* Get my global pointer */
    nexus_init_remote_service_request(&buffer,&addr->gp,
				      BROKER_CLIENT_STRING,
				      BROKER_CLIENT_HASH);
    nexus_set_buffer_size(&buffer,nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp),ngp + 1);
    nexus_put_global_pointer(&buffer,&gp,1);
    nexus_send_remote_service_request(&buffer);
    c2cPrintf(120,"waiting to hear back from broker\n");
    NexusWait(hUser);				/* Wait for reply from broker */
    nexus_get_stashed_int(&hUser.buffer,num_clients,1);
    c2cPrintf(110,"received back %d client's environments\n",*num_clients);
    env = *clientlist = (CaveEnv *)malloc(sizeof(CaveEnv)*(*num_clients));
    for(i=0;i<(*num_clients);i++)
    { /* For */
      c2cPrintf(120,"reading entry number %d from the broker\n",i);
      nexus_get_stashed_char(&hUser.buffer,env->name,C2C_NAME_SIZE);
      nexus_get_stashed_int(&hUser.buffer,&env->type,1);
      nexus_get_stashed_int(&hUser.buffer,&env->id,1);
      env++;					/* Advance to next location */
    } /* For */
    nexus_free_stashed_buffer(&hUser.buffer);
    c2cPrintf(110,"finished client query of the broker\n");
    return(E_SUCCESS);                          /* Return success */
  } /* If */
  else						/* Host not found in list */
  { /* Else */
    c2cPrintf(110,"host %d is not in the broker list\n",host);
    return(E_INVALID_HOST);                     /* Return error */
  } /* Else */
} /* c2cGetClients */

/*****************************************************************************  

          Params:  The host where the session resides
		   My CAVEId in the host
		   Session structure to add
         Returns:  Unique session id on that host
  Global Effects:  None

  This function adds a local session to the broker's table
  
******************************************************************************/
SessionId c2cSessionCreate(HostId host,CaveSession *session_info)
{ /* c2cSessionCreate */
  SessionId id;				/* Id of session on broker */
  nexus_global_pointer_t gp;
  nexus_buffer_t buffer;
  int ngp;
  NexusAddr *addr;				/* Work pointer */

  c2cPrintf(110,"attempting to create session on host %d\n",host);
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending session create message to broker\n");
    NexusInit(hNewSession);             /* Initialize reply */
    nexus_global_pointer(&gp,NULL);
    nexus_init_remote_service_request(&buffer,&addr->gp,
				      BROKER_NEWSESSION_STRING,
				      BROKER_NEWSESSION_HASH);
    nexus_set_buffer_size(&buffer,nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp) +
			  nexus_sizeof_int(&buffer,1) +
  			  nexus_sizeof_char(&buffer,C2C_NAME_SIZE) +
  			  nexus_sizeof_char(&buffer,C2C_PATH_SIZE) +
  			  nexus_sizeof_char(&buffer,C2C_EXE_SIZE) +
  			  nexus_sizeof_char(&buffer,C2C_ARGS_SIZE) +
			  nexus_sizeof_int(&buffer,1),
			  ngp + 6);
    nexus_put_global_pointer(&buffer,&gp,1);
    nexus_put_int(&buffer,&session_info->owner,1);
    nexus_put_char(&buffer,session_info->name,C2C_NAME_SIZE);
    nexus_put_char(&buffer,session_info->pathname,C2C_PATH_SIZE);
    nexus_put_char(&buffer,session_info->execname,C2C_EXE_SIZE);
    nexus_put_char(&buffer,session_info->args,C2C_ARGS_SIZE);
    nexus_send_remote_service_request(&buffer);
    c2cPrintf(120,"waiting to hear reply from the broker\n");
    NexusWait(hNewSession);			/* Wait for reply from broker */
    nexus_get_stashed_int(&hNewSession.buffer,&id,1);
    nexus_free_stashed_buffer(&hNewSession.buffer);
  } /* If */
  else					/* Host not found in list */
  { /* Else */
    id = E_INVALID_HOST;                /* Set error */
  } /* Else */
  c2cPrintf(110,"returning from session create with status %d\n",id);
  return(id);				/* Send id back */
} /* c2cSessionCreate */

/*****************************************************************************  

          Params:  Host to query
	           Number of sessions in the database
		   Array of the sessions environments
         Returns:  Number of clients
	           Array of the clients environments
  Global Effects:  None

  This function attaches to the broker listed in the structure and retrieves
  the capabilities of all CAVES in the database.

******************************************************************************/
int c2cGetSessions(HostId host,int *num_sessions,CaveSession **sessionlist)
{ /* c2cGetClients */
  nexus_buffer_t buffer;		/* Data buffer for sends */
  nexus_global_pointer_t gp;		/* Send along global pointer */
  int ngp;				/* Size of global pointer */
  int i;				/* Loop counter */
  CaveSession *ses;			/* Work pointer */
  NexusAddr *addr;			/* Work pointer */

  c2cPrintf(110,"attempting to query for session information onhost %d\n");
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending query message to the broker\n");
    NexusInit(hSession);             /* Initialize reply */
    nexus_global_pointer(&gp,NULL);	/* Get local global pointer */
    nexus_init_remote_service_request(&buffer,&addr->gp,
				      BROKER_SESSION_STRING,
				      BROKER_SESSION_HASH);
    nexus_set_buffer_size(&buffer,
                          nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp),
			  ngp);
    nexus_put_global_pointer(&buffer,&gp,1);
    nexus_send_remote_service_request(&buffer);
    c2cPrintf(120,"waiting to hear reply from broker\n");
    NexusWait(hSession);		/* Wait for a reply from broker */
    nexus_get_stashed_int(&hSession.buffer,num_sessions,1);
    c2cPrintf(110,"received %d session's from broer\n",*num_sessions);
    ses = *sessionlist = (CaveSession *)malloc(sizeof(CaveSession)*(*num_sessions));
    for(i=0;i<(*num_sessions);i++)	/* Read data for all sessions registeres */
    { /* For */
      c2cPrintf(120,"reading session %d from broker\n");
      nexus_get_stashed_int(&hSession.buffer,&ses->owner,1);
      nexus_get_stashed_char(&hSession.buffer,ses->name,C2C_NAME_SIZE);
      nexus_get_stashed_char(&hSession.buffer,ses->pathname,C2C_PATH_SIZE);
      nexus_get_stashed_char(&hSession.buffer,ses->execname,C2C_EXE_SIZE);
      nexus_get_stashed_char(&hSession.buffer,ses->args,C2C_ARGS_SIZE);
      nexus_get_stashed_int(&hSession.buffer,&ses->id,1);
      ses++;				/* Advance to next location */
    } /* For */
    nexus_free_stashed_buffer(&hSession.buffer); /* Get rid of stash buffer */
    c2cPrintf(110,"finished query of broker for sessions\n");
    return(E_SUCCESS);                  /* Success */
  } /* If */
  else					/* Host not found in list */
  { /* Else */
    c2cPrintf(110,"host %d is not found in the broker list\n",host);
    return(E_INVALID_HOST);             /* Return error */
  } /* Else */
} /* c2cGetClients */

/*****************************************************************************  

          Params:  Host where session is located
		   CAVE to connect to
		   Type of data to receive
		   Callback to execute when data arrives
         Returns:  None
  Global Effects:  Data is streamed to process

  This function subscribved to a remote stream

******************************************************************************/
int c2cSubscribe(HostId host,CAVEId datasource,StreamType stream,void *callback)
{ /* c2cSubscribe */
  nexus_buffer_t buffer;		/* Data buffer for sends */
  nexus_global_pointer_t gp;		/* Send along global pointer */
  int ngp;				/* Size of global pointer */
  int status;				/* Status indicator from remote subscribe */
  Streamlist *list;			/* Work pointer to stream list manipulation */
  Streamlist *node;			/* Buffer for new nodes */
  NexusAddr *addr;			/* Work pointer */

  c2cPrintf(110,"attempting to subscribe to %d for stream %d on host %d with callback %p\n",
            datasource,stream,host,callback);
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending subscribe message to the broker\n");
    NexusInit(hSubscribe);              /* Initialize reply */
    nexus_global_pointer(&gp,NULL);	/* Derive local global pointer */
    nexus_init_remote_service_request(&buffer,&addr->gp,
				      BROKER_SUBSCRIBE_STRING,
				      BROKER_SUBSCRIBE_HASH);
    nexus_set_buffer_size(&buffer,nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp) +
				  nexus_sizeof_int(&buffer,1) +
				  nexus_sizeof_int(&buffer,1),
				  ngp + 2);
    nexus_put_global_pointer(&buffer,&gp,1);	/* Pack the buffer */
    nexus_put_int(&buffer,&datasource,1);	/* Who to subscribe to */
    nexus_put_int(&buffer,&stream,1);		/* What to subscribe to */
    nexus_send_remote_service_request(&buffer);	/* Send along everything */
    c2cPrintf(120,"waiting for the broker to reply\n");
    NexusWait(hSubscribe);			/* Wait for status from remote process */
    nexus_get_stashed_int(&hSubscribe.buffer,&status,1); /* Read status from remote process */
    nexus_free_stashed_buffer(&hSubscribe.buffer); /* Get rid of stash buffer */
    c2cPrintf(110,"return status from the broker %d\n",status);
    if(status == E_SUCCESS)		/* Stream is there to attach to */
    { /* If */
      for(list=c2cStreams;		/* Find our stream node process */
          (list != NULL) && (list->type!=stream);
          list=list->nexttype);
      if(list == NULL)			/* Not in list-->add it */
      { /* If */
        list = c2cStreams;		/* Point to beginning of list */
        node = (Streamlist *)malloc(sizeof(Streamlist)); /* Get some memory */
        node->nextfunc = NULL;		/* Initialize the node info */
        node->nexttype = list;
        c2cStreams = node;
      } /* If */
      else				/* In list-->add to sub list */
      { /* Else */
        node = (Streamlist *)malloc(sizeof(Streamlist));	/* Get some memory */
        node->nextfunc = list->nextfunc;	/* Initialize the node info */
        list->nextfunc = node;
        node->nexttype = NULL;
      } /* Else */
      node->type = stream;		/* Initialize the node info */
      node->callback = callback;
      node->source = datasource;
      c2cPrintf(120,"added the data subscription to stream list\n");
    } /* If */
  } /* If */
  else					/* Host not found in list */
  { /* Else */
    status = E_INVALID_HOST;            /* Set error */
  } /* Else */
  c2cPrintf(110,"status of the data subscribe %d\n",status);
  return(status);                       /* Send back status of operation */
} /* c2cSubscribe */

/*****************************************************************************  

          Params:  Host where session is located
		   CAVE to connect to
		   Type of data to receive
		   Callback executed when data was received
         Returns:  None
  Global Effects:  Data is streamed to process

  This function unsubscribes to a remote stream

******************************************************************************/
int c2cUnsubscribe(HostId host,CAVEId datasource,StreamType stream,void *callback)
{ /* c2cSubscribe */
  nexus_buffer_t buffer;		/* Data buffer for sends */
  nexus_global_pointer_t gp;		/* Send along global pointer */
  int ngp;				/* Size of global pointer */
  Streamlist *list;			/* Work pointer for stream list manipulation */
  Streamlist *temp;			/* Work pointer for stream list manipulation */
  NexusAddr *addr;			/* Work pointer */
  int status;				/* Status of current operation */

  c2cPrintf(110,"attempting to unsubscribe to %d for stream %d on host %d with callback %p\n",
            datasource,stream,host,callback);
  addr = c2cBrokers.addr;               /* Start at beginning of list */
  if(addr != NULL)                      /* Anybody in list */
  { /* If */
    for(/* Starting condition already set */;  /* Prepare for list traversal */
        (addr != NULL) && (addr->rank != host);
        addr = addr->next);
  } /* If */
  if(addr != NULL)			/* Found the host we want */
  { /* If */
    c2cPrintf(120,"sending unsubscribe message to the broker\n");
    NexusInit(hUnsubscribe);             /* Initialize reply */
    nexus_global_pointer(&gp,NULL);	/* Derive local global pointer */
    nexus_init_remote_service_request(&buffer,&addr->gp,
				      BROKER_UNSUBSCRIBE_STRING,
				      BROKER_UNSUBSCRIBE_HASH);
    nexus_set_buffer_size(&buffer,nexus_sizeof_global_pointer(&buffer,&gp,1,&ngp) +
				  nexus_sizeof_int(&buffer,1)+
				  nexus_sizeof_int(&buffer,1),
				  ngp + 2);
    nexus_put_global_pointer(&buffer,&gp,1);	/* Pack the buffer */
    nexus_put_int(&buffer,&datasource,1);
    nexus_put_int(&buffer,&stream,1);
    nexus_send_remote_service_request(&buffer); /* Send along to remote process */
    c2cPrintf(120,"waiting for a reply from the broker\n");
    NexusWait(hUnsubscribe);		/* Wait for reply */
    nexus_get_stashed_int(&hUnsubscribe.buffer,&status,1);
    nexus_free_stashed_buffer(&hUnsubscribe.buffer); /* Get rid of stash buffer */
    c2cPrintf(110,"return status from the broker\n");
    if(status==E_SUCCESS)		/* Stream is there to unsubscribe to */
    { /* If */
      list = c2cStreams;		/* Point to beginning of list */
      if(list != NULL)			/* Stream list is empty */
      { /* If */
        if(list->type == stream)	/* First type match? */
        { /* If */
          if(list->callback == callback)/* First functions match? */
          { /* If */
            if(list->nextfunc==NULL)	/* Only one callback is there */
            { /* If */
              c2cStreams = c2cStreams->nexttype; /* Update pointers */
              free(list);		/* Free memory */
            } /* If */
            else			/* Bring down next child node */
            { /* Else */
              c2cStreams = list->nextfunc; /* Update pointers */
              c2cStreams->nexttype = list->nexttype;
              free(list);		/* Free memory */
	      c2cPrintf(120,"removed the entry from the stream list\n");
            } /* Else */
          } /* If */
	  else				/* Search the first sub-list */
	  { /* Else */
            for(temp = list->nextfunc;	/* Prepare for list traversal */
                (temp != NULL) && (temp->callback != callback);
	        temp = temp->nextfunc, list = list->nextfunc);
            if(temp != NULL)		/* Found it-->erase it */
            { /* If */
              list->nextfunc = temp->nextfunc; /* Remove links from table */
              free(temp);		/* Free memory */
	      c2cPrintf(120,"removed the entry from the stream list\n");
            } /* If */
	    else			/* Not found in list */
            { /* Else */
              status = E_INVALID_CALLBACK; /* Set error status */
            } /* Else */        
	  } /* Else */
        } /* If */
        else				/* Not first node in list--search */
        { /* Else */
          for(temp=list->nexttype;	/* Find our stream node process */
              (temp != NULL) && (temp->type!=stream);
              temp=temp->nexttype,list=list->nexttype);
          if(temp != NULL)		/* In list so remove it */
          { /* If */
	    if(temp->callback == callback) /* First node in func list? */
	    { /* If */
	      if(temp->nextfunc == NULL) /* First node only node? */
	      { /* If */
	        list->nexttype = temp->nexttype; /* Update pointers */
	        free(temp);		/* Free memory */
	        c2cPrintf(120,"removed the entry from the stream list\n");
	      } /* If */
	      else			/* Bring down another node */
	      { /* Else */
	        list->nexttype = temp->nextfunc; /* Update pointers */
	        list->nexttype->nexttype = temp->nexttype;
	        free(temp);		/* Free memory */
	        c2cPrintf(120,"removed the entry from the stream list\n");
	      } /* Else */
	    } /* If */
	    else			/* Start searching */
	    { /* Else */
              for(temp = list->nextfunc; /* Prepare for list traversal */
                  (temp != NULL) && (temp->callback != callback);
	          temp = temp->nextfunc, list = list->nextfunc);
              if(temp != NULL)		/* Found it-->erase it */
              { /* If */
	        list->nextfunc = temp->nextfunc; /* Remove links from table */
	        free(temp);		/* Free memory */
	        c2cPrintf(120,"removed the entry from the stream list\n");
              } /* If */
	      else			/* Not found in list */
              { /* Else */
                status = E_INVALID_CALLBACK; /* Set error status */
              } /* Else */        
	    } /* Else */        
          } /* If */
	  else
	  { /* Else */
            status = E_INVALID_STREAM; /* Set error status */
	  } /* Else */
        } /* Else */
      } /* If */
      else				/* Not in the list at all */
      { /* Else */
        status = E_INVALID_STREAM; /* Set error status */
      } /* Else */
    } /* If */
  } /* If */
  else					/* Host not found in list */
  { /* Else */
    status = E_INVALID_HOST;		/* Set error status */
  } /* Else */
  c2cPrintf(110,"status of the unsubscribe operation %d\n",status);
  return(status);                       /* Return status of operation */
} /* c2cUnsubscribe */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Particular callback is executed for client data

  This registers a particular stream in the c2cDataCast list.  All data
  streams must be registered or the data cannot be subscribed to by remote
  processes.
    
******************************************************************************/
int c2cRegisterStream(StreamType stream,void *callback)
{ /* c2cRegisterStream */
  Castlist *list;			/* Work pointer */
  int status = E_SUCCESS;		/* Flag for success/failure */

  c2cPrintf(110,"attempting to register stream %d with callback %p\n",
            stream,callback);
  list = c2cDataCast;			/* Initialize work pointer */
  if(list != NULL)			/* Search to see if already registered */
  { /* If */
    for( /* Start condition already set */ ; /* Start at the beginning... */
        (list != NULL) && (list->type != stream);
	list = list->next);
  } /* If */
  if(list == NULL)			/* Good, not in list */
  { /* If */
    list = (Castlist *)malloc(sizeof(Castlist)); /* Get some memory */
    list->next = c2cDataCast;		/* Insert into linked list */
    c2cDataCast = list;
    list->type = stream;		/* Assign stream type to node */
    list->conntab.count = 0;		/* Initialize all cast stuff */
    list->conntab.addr = NULL;
    list->frequency = 0;
    list->callback = callback;
  } /* If */
  else
  { /* Else */
    status = E_STREAM_REGISTERED;       /* Set error flag */
  } /* Else */
  c2cPrintf(110,"status of the stream registration is %d\n",status);
  return(status);			/* Return status of operation */
} /* c2cRegisterStream */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Particular callback is executed for client data

  This unregisters a particular stream in the c2cDataCast list.  All data
  streams unregistered cannot be subscribed to by remote processes.  All
  current subscriptions are deleted upon this call.
    
******************************************************************************/
int c2cUnregisterStream(StreamType stream)
{ /* c2cUnregisterStream */
  Castlist *curlist,*prevlist;		/* Work pointers */
  int status = E_SUCCESS;		/* Success flag */

  c2cPrintf(110,"attempting to unregister stream %d\n",stream);
  if(c2cDataCast != NULL)		/* A list exists to process */
  { /* If */
    prevlist = c2cDataCast;		/* Start at the beginning */
    if(prevlist->type == stream)	/* First node the match? */
    { /* If */
      c2cDataCast = prevlist->next;	/* Update pointers */
      free(prevlist);			/* Free memory */
      c2cPrintf(120,"removing the stream %d from the stream list\n",stream);
    } /* If */
    else				/* Start searchin... */
    { /* Else */
      for(curlist=c2cDataCast->next;	/* Start at beginning... */
          (curlist != NULL) && (curlist->type != stream);
          curlist = curlist->next, prevlist = prevlist->next);
      if(curlist != NULL)			/* Stream is found */
      { /* If */
        prevlist->next = curlist->next;	/* Update pointers */
	free(curlist);			/* Free associated memory */
        c2cPrintf(120,"removing the stream %d from the stream list\n",stream);
      } /* If */
      else
      { /* Else */
	status = E_INVALID_STREAM;	/* Set error flag */
      } /* Else */  
    } /* Else */
  } /* If */
  else
  { /* Else */
      status = E_INVALID_STREAM;	/* Set error flag */
  } /* Else */
  c2cPrintf(110,"exiting the unregistration with a status of %d\n",status);
  return(status);			/* Send back result */
} /* c2cUnregisterStream */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Particular callback is executed for client data

  This process sends a particular set of data to all in a particular conntab
    
******************************************************************************/
int c2cSendStream(c2cBuffer **buffer,StreamType stream)
{ /* SendStream */
  nexus_buffer_t buf;			/* Data buffer */
  int ngp;				/* Global pointer size buffer */
  Castlist *list;			/* Pointer to streams to cast */
  NexusAddr *addr;			/* Node to send to */
  c2cBuffer *curbuf;			/* Work pointer */
  int size;				/* Size of data to send */

  c2cPrintf(110,"sending a stream %d to the world\n",stream);
  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Stream found -->start streaming */
  { /* If */
    for(addr=list->conntab.addr;	/* Start at beginning of list */
	addr != NULL;			/* Stop when end of list is reached */
	addr = addr->next)		/* Increment each node */
    { /* For loop */
      c2cPrintf(120,"sending stream to connection\n");
      nexus_init_remote_service_request(&buf,&addr->gp,
				      STREAM_CALLBACK_STRING,STREAM_CALLBACK_HASH);
      nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&addr->gp,1,&ngp) +
			    nexus_sizeof_int(&buf,1) +
			    nexus_sizeof_int(&buf,1),
			    ngp + 2);
      nexus_put_global_pointer(&buf,&addr->gp,1); /* Pack buffer to sender */
      nexus_put_int(&buf,&addr->caveid,1);	/* ID of sending process */
      nexus_put_int(&buf,&stream,1);		/* Stream type sent */
      for(curbuf=*buffer;curbuf != NULL;curbuf = curbuf->next)
      { /* For loop */
        c2cPrintf(130,"processing data type %d\n",curbuf->type);
        switch(curbuf->type)		/* What type of data? */
        { /* Switch */
	  case DT_CHAR:
	    c2cPrintf(130,"processing character data for send\n");
	    size = nexus_sizeof_char(&buf,curbuf->size);
            nexus_check_buffer_size(&buf,size,size); /* Avoid buffer overrun */
	    nexus_put_char(&buf,(char *)curbuf->data,curbuf->size);
	    break;
	  case DT_INT:
	    c2cPrintf(130,"processing integer data for send\n");
	    size = nexus_sizeof_int(&buf,curbuf->size);
            nexus_check_buffer_size(&buf,size,size); /* Avoid buffer overrun */
	    nexus_put_int(&buf,(int *)curbuf->data,curbuf->size);
	    break;
	  case DT_LONG:
	    c2cPrintf(130,"processing long data for send\n");
	    size = nexus_sizeof_long(&buf,curbuf->size);
            nexus_check_buffer_size(&buf,size,size); /* Avoid buffer overrun */
	    nexus_put_long(&buf,(long *)curbuf->data,curbuf->size);
	    break;
	  case DT_FLOAT:
	    c2cPrintf(130,"processing float data for send\n");
	    size = nexus_sizeof_float(&buf,curbuf->size);
            nexus_check_buffer_size(&buf,size,size); /* Avoid buffer overrun */
	    nexus_put_float(&buf,(float *)curbuf->data,curbuf->size);
	    break;
	  case DT_DOUBLE:
	    c2cPrintf(130,"processing double data for send\n");
	    size = nexus_sizeof_double(&buf,curbuf->size);
            nexus_check_buffer_size(&buf,size,size); /* Avoid buffer overrun */
	    nexus_put_double(&buf,(double *)curbuf->data,curbuf->size);
	    break;
	} /* Switch */
      } /* For loop */
      nexus_send_remote_service_request(&buf);	/* Send it all along */
      c2cPrintf(120,"finished sending stream to connection\n");
    } /* For loop */
    c2cPrintf(110,"finished broadcasting stream to world\n");
    return(E_SUCCESS);                  /* Send success back */
  } /* If */
  else					/* Stream type unknown */
  { /* Else */
    c2cPrintf(110,"stream %d is invalid\n",stream);
    return(E_INVALID_STREAM);           /* Return error */
  } /* Else */
} /* SendStream */

/*****************************************************************************  

          Params:  CAVE user
         Returns:  Nothing
  Global Effects:  None

  This process casts a user to the world
    
******************************************************************************/
int c2cSendTracker(float x,float y,float z,float a,float e,float r,int stream)
{ /* c2cSendTracker */
  nexus_buffer_t buf;			/* Data buffer */
  int ngp;				/* Global pointer size buffer */
  Castlist *list;			/* Pointer to streams to cast */
  NexusAddr *addr;			/* Node to send to */

  c2cPrintf(110,"broadcasting tracker data of type %d to world\n",stream);
  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Stream found -->start streaming */
  { /* If */
    for(addr=list->conntab.addr;	/* Start at beginning of list */
	addr != NULL;			/* Stop when end of list is reached */
	addr = addr->next)		/* Increment each node */
    { /* For loop */
      c2cPrintf(120,"sending tracker data to connection\n");
      nexus_init_remote_service_request(&buf,&addr->gp,
				      STREAM_CALLBACK_STRING,STREAM_CALLBACK_HASH);
      nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&addr->gp,1,&ngp) +
			    nexus_sizeof_int(&buf,2) +
			    nexus_sizeof_float(&buf,6),
			    ngp + 8);
      nexus_put_global_pointer(&buf,&addr->gp,1); /* Pack buffer to sender */
      nexus_put_int(&buf,&addr->caveid,1);	/* ID of sending process */
      nexus_put_int(&buf,&stream,1);		/* Stream type sent */
      nexus_put_float(&buf,&x,1);
      nexus_put_float(&buf,&y,1);
      nexus_put_float(&buf,&z,1);
      nexus_put_float(&buf,&a,1);
      nexus_put_float(&buf,&e,1);
      nexus_put_float(&buf,&r,1);
      nexus_send_remote_service_request(&buf);	/* Send it all along */
      c2cPrintf(120,"finished sending tracker data to connection\n");
    } /* For loop */
    c2cPrintf(110,"finished broadcasting tracker data to the world\n");
    return(E_SUCCESS);                  /* Send back success */
  } /* If */
  else					/* Stream type unknown */
  { /* Else */
    c2cPrintf(110,"stream %d is invalid\n",stream);
    return(E_INVALID_STREAM);           /* Stream not registered */
  } /* Else */
} /* c2cSendTracker */

/*****************************************************************************  

          Params:  CAVE wand buttons
         Returns:  Error code
  Global Effects:  None

  This process casts wand buttons to the world
    
******************************************************************************/
int c2cSendWandButtons(int b1,int b2,int b3,int b4)
{ /* c2cSendWandButtons */
  nexus_buffer_t buf;			/* Data buffer */
  int ngp;				/* Global pointer size buffer */
  Castlist *list;			/* Pointer to streams to cast */
  NexusAddr *addr;			/* Node to send to */
  int stream = S_WAND_BUTTON;		/* Type of stream we're sending */

  c2cPrintf(110,"sending wand buttons to the world\n");
  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Stream found -->start streaming */
  { /* If */
    for(addr=list->conntab.addr;	/* Start at beginning of list */
	addr != NULL;			/* Stop when end of list is reached */
	addr = addr->next)		/* Increment each node */
    { /* For loop */
      c2cPrintf(120,"sending wand buttons to a connection\n");
      nexus_init_remote_service_request(&buf,&addr->gp,
				      STREAM_CALLBACK_STRING,STREAM_CALLBACK_HASH);
      nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&addr->gp,1,&ngp) +
			    nexus_sizeof_int(&buf,6),
			    ngp + 6);
      nexus_put_global_pointer(&buf,&addr->gp,1); /* Pack buffer to sender */
      nexus_put_int(&buf,&addr->caveid,1);	/* ID of sending process */
      nexus_put_int(&buf,&stream,1);		/* Stream type sent */
      nexus_put_int(&buf,&b1,1);
      nexus_put_int(&buf,&b2,1);
      nexus_put_int(&buf,&b3,1);
      nexus_put_int(&buf,&b4,1);
      nexus_send_remote_service_request(&buf);	/* Send it all along */
      c2cPrintf(120,"finished sending wand buttons to a connection\n");
    } /* For loop */
    c2cPrintf(110,"finished broadcasting wand buttons to world\n");
    return(E_SUCCESS);                  /* Return success */    
  } /* If */
  else					/* Stream type unknown */
  { /* Else */
    c2cPrintf(110,"S_WAND_BUTTON is not registered\n");
    return(E_INVALID_STREAM);           /* Return error */
  } /* Else */
} /* c2cSendWandButtons */

/*****************************************************************************  

          Params:  CAVE user
         Returns:  Nothing
  Global Effects:  None

  This process casts joystick info to the world
    
******************************************************************************/
int c2cSendWandJoystick(float joyx,float joyy)
{ /* c2cSendWandJoystick */
  nexus_buffer_t buf;			/* Data buffer */
  int ngp;				/* Global pointer size buffer */
  Castlist *list;			/* Pointer to streams to cast */
  NexusAddr *addr;			/* Node to send to */
  int stream = S_WAND_JOYSTICK;		/* Type of stream we're sending */

  c2cPrintf(110,"sending joystick values to the world\n");
  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Stream found -->start streaming */
  { /* If */
    for(addr=list->conntab.addr;	/* Start at beginning of list */
	addr != NULL;			/* Stop when end of list is reached */
	addr = addr->next)		/* Increment each node */
    { /* For loop */
      c2cPrintf(120,"sending joystick information to connection\n");
      nexus_init_remote_service_request(&buf,&addr->gp,
				      STREAM_CALLBACK_STRING,STREAM_CALLBACK_HASH);
      nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&addr->gp,1,&ngp) +
			    nexus_sizeof_float(&buf,2) +
			    nexus_sizeof_int(&buf,2),
			    ngp + 4);
      nexus_put_global_pointer(&buf,&addr->gp,1); /* Pack buffer to sender */
      nexus_put_int(&buf,&addr->caveid,1);	/* ID of sending process */
      nexus_put_int(&buf,&stream,1);		/* Stream type sent */
      nexus_put_float(&buf,&joyx,1);
      nexus_put_float(&buf,&joyy,1);
      nexus_send_remote_service_request(&buf);	/* Send it all along */
      c2cPrintf(120,"finished sending joystick information to connection\n");
    } /* For loop */    
    c2cPrintf(110,"finished broadcasting joystick values to the world\n");
    return(E_SUCCESS);                  /* Send back success */
  } /* If */
  else					/* Stream type unknown */
  { /* Else */
    c2cPrintf(110,"S_WAND_JOYSTICK is not registered\n");
    return(E_INVALID_STREAM);           /* Send back error */
  } /* Else */
} /* c2cSendWandJoystick */

/*****************************************************************************  

          Params:  CAVE user
         Returns:  Nothing
  Global Effects:  None

  This process casts a user to the world
    
******************************************************************************/
int c2cSendUser(CAVEUser *user)
{ /* c2cSendUser */
  nexus_buffer_t buf;			/* Data buffer */
  int ngp;				/* Global pointer size buffer */
  Castlist *list;			/* Pointer to streams to cast */
  NexusAddr *addr;			/* Node to send to */
  int stream = S_CAVE_USER;		/* Type of stream we're sending */

  c2cPrintf(110,"sending CAVE user information to world\n");
  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Stream found -->start streaming */
  { /* If */
    for(addr=list->conntab.addr;	/* Start at beginning of list */
	addr != NULL;			/* Stop when end of list is reached */
	addr = addr->next)		/* Increment each node */
    { /* For loop */
      c2cPrintf(120,"sending CAVE user to connection\n");
      nexus_init_remote_service_request(&buf,&addr->gp,
				      STREAM_CALLBACK_STRING,STREAM_CALLBACK_HASH);
      nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&addr->gp,1,&ngp) +
			    nexus_sizeof_int(&buf,1) +
			    nexus_sizeof_int(&buf,1) +
			    nexus_sizeof_char(&buf,sizeof(CAVEUser)),
			    ngp + 3);
      nexus_put_global_pointer(&buf,&addr->gp,1); /* Pack buffer to sender */
      nexus_put_int(&buf,&addr->caveid,1);	/* ID of sending process */
      nexus_put_int(&buf,&stream,1);		/* Stream type sent */
      nexus_put_char(&buf,(char *)user,sizeof(CAVEUser)); /* Send along CAVE user */
      nexus_send_remote_service_request(&buf);	/* Send it all along */
      c2cPrintf(120,"finished sending CAVE user to connection\n");
    } /* For loop */
    c2cPrintf(110,"finished broadcasting CAVE user to world\n");
    return(E_SUCCESS);                  /* Return success */
  } /* If */
  else					/* Stream type unknown */
  { /* Else */
    c2cPrintf(110,"S_CAVE_USER is not registered\n");
    return(E_INVALID_STREAM);           /* Return error */
  } /* Else */
} /* c2cSendUser */

/******************************************************************************

         Params:  Stream type
		  Frequency of skips
   Return Value:  Nothing
 Global Effects:  None

 This function assigns the number of skips of a particular stream in the
 cast thread

******************************************************************************/
int c2cSetAutoFrequency(StreamType stream,int action)
{ /* c2cSetAutoFrequency */
  Castlist *list;			/* Work pointer */

  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Found the stream */
  { /* If */
    list->autofrequency = action;	/* Assign list frequency */
    return(E_SUCCESS);                  /* Return success */
  } /* If */
  else					/* Stream not found in list */
  { /* Else */
    return(E_INVALID_STREAM);           /* Invalid stream given */
  } /* Else */
} /* c2cSetAutoFrequency */

/******************************************************************************

         Params:  Stream type
		  Frequency of skips
   Return Value:  Nothing
 Global Effects:  None

 This function assigns the number of skips of a particular stream in the
 cast thread

******************************************************************************/
int c2cSetSkipFrequency(StreamType stream,int frequency)
{ /* c2cSetSkipFrequency */
  Castlist *list;			/* Work pointer */

  list = c2cDataCast;                   /* Start at beginning of list */
  if(list != NULL)                      /* Any list to traverse? */
  { /* If */
    for( /* Starting condition already set */; /* Find our stream table */
        (list != NULL) && (list->type != stream);
        list = list->next);
  } /* If */
  if(list != NULL)			/* Found the stream */
  { /* If */
    list->frequency = frequency;	/* Assign list frequency */
    return(E_SUCCESS);                  /* Return success */
  } /* If */
  else					/* Stream not found in list */
  { /* Else */
    return(E_INVALID_STREAM);           /* Return error */
  } /* Else */
} /* c2cSetSkipFrequency */

/******************************************************************************

         Params:  Data buffer
		  Data to put into buffer
		  Length of data
		  Type of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function adds a block of data to a c2cBuffer for later broadcasting

******************************************************************************/
void c2cPackBuffer(c2cBuffer **buffer,void *data,int length,int type)
{ /* c2cPackBuffer */
  c2cBuffer *buf;			/* Work buffer */
  c2cBuffer *prevbuf;                   /* Work pointers */
  c2cBuffer *curbuf;

  buf = (c2cBuffer *)malloc(sizeof(c2cBuffer));	/* Get some memory */
  buf->type = type;			/* Assign parameters */
  buf->size = length;
  buf->data = data;
  buf->next = NULL;
  if(*buffer != NULL)                   /* More in the list so search */
  { /* If */
    curbuf = prevbuf = *buffer;         /* Start traversal at beginning of list */
    while(prevbuf != NULL)              /* Go to end of list */
    { /* While */
      curbuf = prevbuf;                 /* Point to this node */
      prevbuf = prevbuf->next;          /* Go to next node */
    } /* While */
    curbuf->next = buf;			/* Insert into list */
  } /* If */
  else                                  /* Insert into beginning of list */
  { /* Else */
    *buffer = buf;                      /* Insert */
  } /* Else */
} /* c2cPackBuffer */

/******************************************************************************

         Params:  Data buffer
		  Buffer to put data into
		  Length of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function gets a block of data from a buffer

******************************************************************************/
void c2cGetChar(void *buffer,char *data,int size)
{ /* c2cGetChar */
  c2cPrintf(130,"getting data of size %d and putting in address %p\n",size,data);
  nexus_get_stashed_char((nexus_stashed_buffer_t *)buffer,data,size);
} /* c2cGetChar */

/******************************************************************************

         Params:  Data buffer
		  Buffer to put data into
		  Length of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function gets a block of data from a buffer

******************************************************************************/
void c2cGetInt(void *buffer,int *data,int size)
{ /* c2cGetInt */
  c2cPrintf(130,"getting data of size %d and putting in address %p\n",size,data);
  nexus_get_stashed_int((nexus_stashed_buffer_t *)buffer,data,size);
} /* c2cGetInt */

/******************************************************************************

         Params:  Data buffer
		  Buffer to put data into
		  Length of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function gets a block of data from a buffer

******************************************************************************/
void c2cGetLong(void *buffer,long *data,int size)
{ /* c2cGetLong */
  c2cPrintf(130,"getting data of size %d and putting in address %p\n",size,data);
  nexus_get_stashed_long((nexus_stashed_buffer_t *)buffer,data,size);
} /* c2cGetLong */

/******************************************************************************

         Params:  Data buffer
		  Buffer to put data into
		  Length of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function gets a block of data from a buffer

******************************************************************************/
void c2cGetFloat(void *buffer,float *data,int size)
{ /* c2cGetFloat */
  c2cPrintf(130,"getting data of size %d and putting in address %p\n",size,data);
  nexus_get_stashed_float((nexus_stashed_buffer_t *)buffer,data,size);
} /* c2cGetFloat */

/******************************************************************************

         Params:  Data buffer
		  Buffer to put data into
		  Length of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function gets a block of data from a buffer

******************************************************************************/
void c2cGetDouble(void *buffer,double *data,int size)
{ /* c2cGetDouble */
  c2cPrintf(130,"getting data of size %d and putting in address %p\n",size,data);
  nexus_get_stashed_double((nexus_stashed_buffer_t *)buffer,data,size);
} /* c2cGetDouble */

/******************************************************************************

         Params:  Data buffer
		  Buffer to put data into
		  Length of data
   Return Value:  Updated buffer
 Global Effects:  None

 This function gets rid of a data buffer used in a callback

******************************************************************************/
void c2cFreeGetBuffer(void *buffer)
{ /* c2cFreeGetBuffer */
  nexus_free_stashed_buffer((nexus_stashed_buffer_t *)buffer);	/* Get rid of stashed buffer */
} /* c2cFreeGetBuffer */

/******************************************************************************

         Params:  Data buffer to empty
   Return Value:  Emptied buffer
 Global Effects:  None

 This function gets rid of a pack data buffer

******************************************************************************/
void c2cFreePackBuffer(c2cBuffer **buffer)
{ /* c2cFreePackBuffer */
  c2cBuffer *buf;			/* Work pointer */

  while(*buffer != NULL)		/* For length of buffer */
  { /* while */
    buf = (*buffer)->next;		/* Point to next node */
    free(*buffer);			/* Get rid of current node */
    *buffer = buf;			/* Update list */
  } /* while */
} /* c2cFreePackBuffer */

/******************************************************************************

         Params:  Data buffer to initialize
   Return Value:  Initialized buffer
 Global Effects:  None

 This function gets rid of a pack data buffer

******************************************************************************/
void c2cInitPackBuffer(c2cBuffer **buffer)
{ /* c2cInitPackBuffer */
  *buffer = NULL;			/* Initialize buffer */
} /* c2cInitPackBuffer */

/******************************************************************************

         Params:  Id of the broker to check
	          Name of the client to query
   Return Value:  Initialized buffer
 Global Effects:  None

 This function returns the id of a client name on a specified broker

******************************************************************************/
CAVEId c2cGetClientId(HostId host,char *name)
{ /* c2cGetClientId */
  int i;				/* Loop counter */
  int numclients;			/* Number of clientss on the broker */
  CaveEnv *env;		                /* List of all clients on broker */
  CAVEId id;				/* Id of client to query */
  int stat;                             /* Status of c2c operations */

  c2cPrintf(110,"attempting to retrieve client id from host %d\n",host);
  stat = c2cGetClients(host,&numclients,&env); /* Get list of everyone registered */
  if(stat < 0)                          /* Error on GetSessions */
  { /* If */
    return(stat);                       /* Send back the error */
  } /* If */
  else
  { /* Else */
    id = (-1);				/* Plant a known elephant in Cairo */
    for(i=0;i<numclients;i++)		/* Search for our session */
    { /* For */
      if(strcmp(env[i].name,name) == 0)	/* Found our match */
      { /* If */
        id = env[i].id;			/* Get the id of the session */
        break;				/* Get out of here... */
      } /* If */
    } /* For */
    if(id != (-1))                      /* Found the client */
    { /* If */
      return(id);                       /* Return the id of the found client */
    } /* If */
    else				/* Error because client wasn't registered */
    { /* Else */
      stat = E_INVALID_CLIENT;          /* Set error flag */
    } /* Else */
  } /* Else */
  c2cPrintf(110,"returning with a status of %d\n");
  return(stat);                         /* Send back error */
} /* c2cGetClientId */

/******************************************************************************

         Params:  Id of the broker to check
	          Name of the session to query
   Return Value:  Id of the session or an error
 Global Effects:  None

 This function returns the id of a client name on a specified broker

******************************************************************************/
CAVEId c2cGetSessionId(HostId host,char *name)
{ /* c2cGetClientId */
  int i;				/* Loop counter */
  int numsessions;			/* Number of sessions on the broker */
  CaveSession *ses;		        /* List of all sessions on broker */
  SessionId id;				/* Id of session to query */
  int stat;                             /* Status of c2c operations */

  c2cPrintf(110,"attempting to retrieve session id from host %d\n",host);
  stat = c2cGetSessions(host,&numsessions,&ses); /* Get list of everyone registered */
  if(stat < 0)                          /* Error on GetSessions */
  { /* If */
    return(stat);                       /* Send back the error */
  } /* If */
  else
  { /* Else */
    id = (-1);				/* Plant a known elephant in Cairo */
    for(i=0;i<numsessions;i++)		/* Search for our session */
    { /* For */
      if(strcmp(ses[i].name,name) == 0)	/* Found our match */
      { /* If */
        id = ses[i].id;			/* Get the id of the session */
        break;				/* Get out of here... */
      } /* If */
    } /* For */
    if(id != (-1))                      /* Found the client */
    { /* If */
      return(id);                       /* Return the id of the found client */
    } /* If */
    else				/* Error because client wasn't registered */
    { /* Else */
      stat = E_INVALID_SESSION;         /* Set error flag */
    } /* Else */
  } /* Else */
  c2cPrintf(110,"returning from the operation with status of %d\n",stat);
  return(stat);                         /* Send back error */
} /* c2cGetClientId */
