/******************************************************************************

  c2cbroker.c
  
  Written by:  Michael Pellegrino
	       Indiana Institute of Technology

  The file contains all code necessary for a device-independent c2c library
  broker.

  Date last edited:  August 4, 1995

  Version:  .12

******************************************************************************/
#include <malloc.h>
#include <nexus.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <c2c.h>
#include "c2cdevice.h"

typedef struct Client clientptr;

typedef struct Client			/* Description of client connected to broker */
{
  NexusAddr addr;			/* How to contact client */
  CaveEnv env;			        /* Environment of client */
  clientptr *next;			/* Next one in list */
}client;


typedef struct Session sessionptr;

typedef struct Session			/* Description of session on broker */
{
  SessionId id;			/* Id of the particular session */
  CaveSession info;			/* Session description */
  Conntab *attached;			/* All connected to session */
  sessionptr *next;			/* Next one in list */
}session;

typedef struct clienttable		/* Table of all clients registered */
{
  int count;				/* Number of clients connected */
  int nextid;				/* Next id issued in table */
  client *clients;			/* All clients registered */
}ClientTable;

typedef struct sessiontable		/* Table of all sessions created */
{
  int count;				/* Number of sessions */
  int nextid;				/* Next id issued in table */
  session *sessions;			/* All sessions created */
}SessionTable;

void BrokerRegisterCallback(void *address, nexus_buffer_t *buffer);
void BrokerUnregisterCallback(void *address, nexus_buffer_t *buffer);
void BrokerClientCallback(void *address, nexus_buffer_t *buffer);
void BrokerSessionCallback(void *address, nexus_buffer_t *buffer);
void BrokerNewSessionCallback(void *address, nexus_buffer_t *buffer);
void BrokerSubscribeCallback(void *address, nexus_buffer_t *buffer);
void BrokerUnsubscribeCallback(void *address, nexus_buffer_t *buffer);
void BrokerKillCallback(void *address, nexus_buffer_t *buffer);

static nexus_handler_t NexusHandlers[] =
{ /* Handler definitions */
  {
    BROKER_REGISTER_STRING,BROKER_REGISTER_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerRegisterCallback
  },   
  {
    BROKER_UNREGISTER_STRING,BROKER_UNREGISTER_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerUnregisterCallback
  },   
  {
    BROKER_CLIENT_STRING,BROKER_CLIENT_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerClientCallback
  },
  {
    BROKER_SESSION_STRING,BROKER_SESSION_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerSessionCallback
  },
  {
    BROKER_SUBSCRIBE_STRING,BROKER_SUBSCRIBE_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerSubscribeCallback
  },
  {
    BROKER_UNSUBSCRIBE_STRING,BROKER_UNSUBSCRIBE_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerUnsubscribeCallback
  },
  {
    BROKER_NEWSESSION_STRING,BROKER_NEWSESSION_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerNewSessionCallback
  },
  {
    BROKER_KILL_STRING,BROKER_KILL_HASH,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)BrokerKillCallback
  },
  {
    (char *)NULL,0,
    NEXUS_HANDLER_TYPE_NON_THREADED,(nexus_handler_func_t)NULL
  }
}; /* Handler definitions */

/* Declare all global variables */
ClientTable c2cClients;			/* Table of all clients */
SessionTable c2cSessions;		/* Table of all sessions */
Handler hExit;				/* Handler for broker exit */

/*****************************************************************************  

          Params:  None
         Returns:  N_SUCCESS
  Global Effects:  None

  This function is called when a nexus context is created
    
******************************************************************************/
int NexusBoot(void)
{ /* NexusBoot */
  nexus_register_handlers(NexusHandlers);	/* Register all handlers */
  return(0);					/* Send back as per protocol */
} /* NexusBoot */

/*****************************************************************************  

          Params:  A url
		   A global pointer
         Returns:  N_SUCCESS
  Global Effects:  None

  This function acts as a callback for the nexus_allow_attachments.  It
  broadcasts a global pointer to everyone who attaches.
      
******************************************************************************/
int NexusAllowAttachments(char *url, nexus_global_pointer_t *gp)
{ /* NexusAllowAttachments */
  nexus_global_pointer(gp, NULL);		/* Send back a global pointer */
  return(0);					/* Send along as per protocol */
} /* NexusAllowAttachments */

/*****************************************************************************  

          Params:  Address of global pointer
		   Data buffer
		   Handler name invoked
		   Handler hash
         Returns:  Nothing
  Global Effects:  None

  This callback is called when a handler that is referenced is not registered
  with nexus.
    
******************************************************************************/
void NexusUnknownHandler(void *address,nexus_buffer_t *buffer,
			 char *handler_name,int handler_id)
{ /* NexusUnknownHandler */
  printf("Warning --> unknown handler activated from within c2c broker\n\n",
  	 "Handler name: %s\n",
	 "  Hash value: %d\n",handler_name,handler_id);
} /* NexusUnknownHandler */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes session inquiries on the client side
    
******************************************************************************/
void BrokerKillCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerKillCallback */
  c2cPrintf(100,"getting message to die\n");
  NexusWake(hExit);				/* Wake up process */
  c2cPrintf(100,"awakening broker process\n");
} /* BrokerKillCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes register requests on the client side
    
******************************************************************************/
void BrokerRegisterCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerRegisterCallback */
  nexus_buffer_t buf;			/* Data buffer */
  client *newclient;			/* Work pointers */
  client *tempclient;
  int stat;                             /* Status of the opertaion */

  c2cPrintf(100,"interrupted for client registration\n");
  newclient = (client *)malloc(sizeof(client));	/* Get some memory */
  nexus_get_global_pointer(buffer,&newclient->addr.gp,1);
  nexus_get_char(buffer,newclient->env.name,C2C_NAME_SIZE);
  nexus_get_int(buffer,&newclient->env.type,1);
  nexus_get_int(buffer,&newclient->env.id,1);
  c2cPrintf(100,"attempting to add %s to client list\n",newclient->env.name);
  tempclient = c2cClients.clients;
  if(tempclient != NULL)                /* Make sure list isn't empty */
  { /* If */
    c2cPrintf(100,"checking table for duplicate client name\n");
    for( tempclient = c2cClients.clients;
        (tempclient != NULL) &&
        (strcmp(tempclient->env.name,newclient->env.name) != 0);
        tempclient = tempclient->next)
    { /* For loop */
      c2cPrintf(120,"comparing %s to %s for match\n",tempclient->env.name,
                newclient->env.name);
    } /* For loop */
  } /* If */
  if(tempclient == NULL)                /* Good --> not in list */
  { /* If */
    c2cClients.count++;			/* Count this new client in table */  
    newclient->next = c2cClients.clients;	/* Insert into the list */
    c2cClients.clients = newclient;
    stat = newclient->env.id = newclient->addr.rank = 
           c2cClients.nextid++; /* Assign id to client */
    c2cPrintf(100,"no duplicates --> client registered as %d\n",stat);
  } /* If */
  else
  { /* Else */
    c2cPrintf(100,"client with name %s already registered\n",newclient->env.name);
    stat = E_CLIENT_EXISTS;             /* Set error flag */
  } /* Else */
  c2cPrintf(100,"sending status %d back to registering process\n",stat);
  nexus_init_remote_service_request(&buf,&newclient->addr.gp,
				    CLIENT_REGISTER_STRING,
				    CLIENT_REGISTER_HASH);
  nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,1),1);
  nexus_put_int(&buf,&stat,1);
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sendng status to registering process\n");
  if(stat == E_CLIENT_EXISTS)           /* Remove old memory */
  { /* If */
    free(newclient);                    /* Get rid of memory */
  } /* If */
} /* BrokerRegisterCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes unregister requests on the client side
    
******************************************************************************/
void BrokerUnregisterCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerUnregisterCallback */
  client *curclient,*prevclient;	/* Work pointers */
  nexus_buffer_t buf;			/* Data buffer */
  nexus_global_pointer_t gp;		/* Calling client id */
  CAVEId id;				/* ID of unregistering client */
  int status;				/* Status of the unregister */

  c2cPrintf(100,"interrupted for client unregistration\n");
  nexus_get_global_pointer(buffer,&gp,1); /* Get calling id */
  nexus_get_int(buffer,&id,1);		/* Get id of unregistering client */
  prevclient = c2cClients.clients;	/* Prepare for removal from list */
  c2cPrintf(100,"client %d attempting to unregister\n",id);
  if(prevclient != NULL)		/* Is there anyone there to remove? */
  { /* If */
    if(prevclient->next == NULL)	/* First entry in list? */
    { /* If */
      c2cClients.count--;		/* Forget about one entry */
      c2cClients.clients = prevclient->next; /* Update pointers */
      nexus_destroy_global_pointer(&prevclient->addr.gp); /* Remove connection */
      free(prevclient);			/* Free memory */
      c2cPrintf(100,"removed client %d from table\n",id);
    } /* If */
    else				/* Start searching */
    { /* Else */
      for(curclient = prevclient->next;	/* Find our client */
          (curclient != NULL) && (curclient->addr.rank != id);
          curclient = curclient->next) ;
      if(curclient != NULL)		/* Found client to unregister */
      { /* If */
        c2cClients.count--;		/* Forget about one entry */
        prevclient->next = curclient->next; /* Update all pointers */
        nexus_destroy_global_pointer(&curclient->addr.gp); /* Remove connection */
        free(curclient);		/* Remove node from memory */
        c2cPrintf(100,"removed client %d from table\n",id);
      } /* If */
      else				/* Error-->client not found on broker */
      { /* Else */
         status = E_INVALID_CAVEID;     /* Set flag */
         c2cPrintf(100,"client %d not found in table table\n",id);
      } /* Else */
    } /* Else */
  } /* If */
  else					/* Error-->client not found on broker */
  { /* Else */
    status = E_INVALID_CAVEID;          /* Set flag */
    c2cPrintf(100,"client %d not found in table table\n",id);
  } /* Else */
  c2cPrintf(100,"sending status %d back to unregistering process\n",stat);
  nexus_init_remote_service_request(&buf,&gp,
				    CLIENT_UNREGISTER_STRING,
				    CLIENT_UNREGISTER_HASH);
  nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,1),1);
  nexus_put_int(&buf,&status,1);
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sending status to unregistering process\n");
} /* BrokerUnregisterCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes register requests on the client side
    
******************************************************************************/
void BrokerClientCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerClientCallback */
  nexus_buffer_t buf;			/* Data buffer */
  nexus_global_pointer_t gp;		/* Global pointer of calling client */
  client *curclient;			/* Work pointer */
  
  c2cPrintf(100,"interrupted by client for client query\n");
  nexus_get_global_pointer(buffer,&gp,1); /* Get caller id */
  c2cPrintf(100,"sending client table back to client\n");
  nexus_init_remote_service_request(&buf,&gp, /* Respond to client */
				    CLIENT_USER_STRING,
				    CLIENT_USER_HASH);
  nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,c2cClients.count) +
  			     nexus_sizeof_int(&buf,c2cClients.count) +
  			     nexus_sizeof_int(&buf,1) +
  			     nexus_sizeof_char(&buf,c2cClients.count*C2C_NAME_SIZE),
			     3*c2cClients.count + 1);
  nexus_put_int(&buf,&c2cClients.count,1); /* Send along count of clients */
  for(curclient = c2cClients.clients;	/* Loop for everyone in the list */
      curclient != NULL;		/* Stop when list end is reached */
      curclient = curclient->next)	/* Send every node */
  { /* For */
    c2cPrintf(120,"sending a client along\n");
    nexus_put_char(&buf,curclient->env.name,C2C_NAME_SIZE);
    nexus_put_int(&buf,&curclient->env.type,1);
    nexus_put_int(&buf,&curclient->env.id,1);
  } /* For */
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sending client table to client\n");
} /* BrokerClientCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes register requests on the client side
    
******************************************************************************/
void BrokerNewSessionCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerNewSessionCallback */
  nexus_buffer_t buf;			/* Data buffer */
  nexus_global_pointer_t gp;		/* Global pointer of calling client */
  session *newsession;			/* Work pointers */
  session *tempsession;
  int stat;                             /* Status of the operation */

  c2cPrintf(100,"interrupted by client for session creation\n");
  newsession = (session *)malloc(sizeof(session));	/* Get some memory */
  nexus_get_global_pointer(buffer,&gp,1); /* Get caller id */
  nexus_get_int(buffer,&newsession->info.owner,1); /* Read all info */
  nexus_get_char(buffer,newsession->info.name,C2C_NAME_SIZE);
  nexus_get_char(buffer,newsession->info.pathname,C2C_PATH_SIZE);
  nexus_get_char(buffer,newsession->info.execname,C2C_EXE_SIZE);
  nexus_get_char(buffer,newsession->info.args,C2C_ARGS_SIZE);
  c2cPrintf(100,"client %d attempting to create session %s",
            newsession->info.owner,newsession->info.name);
  c2cPrintf(100,"checking for duplicates in session table\n");
  tempsession = c2cSessions.sessions;   /* Start at beginning of list */
  if(tempsession != NULL)               /* Search for duplicates */
  { /* If */
    for( /* Starting condition already set */ ;
        (tempsession != NULL) &&
        (strcmp(tempsession->info.name,newsession->info.name) != 0);
        tempsession = tempsession->next)
    { /* For loop */
      c2cPrintf(120,"comparing %s to %s\n",tempsession->info.name,
                newsession->info.name);
    } /* For loop */
  } /* If */
  if(tempsession == NULL)               /* Good --> it's not in the list */
  { /* If */
    c2cSessions.count++;		     /* Count the new session in table */
    newsession->next = c2cSessions.sessions; /* Insert into the list */
    c2cSessions.sessions = newsession;
    newsession->id = c2cSessions.nextid++;   /* Assign id to session */
    newsession->attached = NULL;	     /* Clear all attachment flags */
    stat = newsession->info.id = newsession->id; /* Assign the session its id */
    c2cPrintf(100,"no duplicates found --> session added with id %d\n",stat);
  } /* If */
  else
  { /* Else */
    c2cPrintf(100,"session %s already exists on broker\n",newsession->info.name);
    stat = E_SESSION_EXISTS;            /* Set error flag */
    c2cPrintf(100,"session %s already exists in table\n",newsession->info.name);
  } /* Else */
  c2cPrintf(100,"sending status %d back to client\n",stat);
  nexus_init_remote_service_request(&buf,&gp, /* Give reply to client */
				    CLIENT_NEWSESSION_STRING,
				    CLIENT_NEWSESSION_HASH);
  nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,1),2);
  nexus_put_int(&buf,&newsession->id,1); /* Send back id */
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sending status to client\n");
  if(stat == E_SESSION_EXISTS)          /* Free old memory */
  { /* If */
    free(newsession);                   /* Get rid of memory */
  } /* If */
} /* BrokerNewSessionCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes register requests on the client side
    
******************************************************************************/
void BrokerSessionCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerSessionCallback */
  nexus_buffer_t buf;			/* Data buffer */
  nexus_global_pointer_t gp;		/* Caller id */
  session *cursession;			/* Work pointer */
  
  c2cPrintf(100,"interrupted by client for session query\n");
  nexus_get_global_pointer(buffer,&gp,1); /* Get calling client */
  c2cPrintf(100,"sending session table to client\n");
  nexus_init_remote_service_request(&buf,&gp,	/* Send back response */
				    CLIENT_SESSION_STRING,
				    CLIENT_SESSION_HASH);
  nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,2*c2cSessions.count + 1) +
  			     nexus_sizeof_char(&buf,c2cSessions.count*C2C_NAME_SIZE) +
  			     nexus_sizeof_char(&buf,c2cSessions.count*C2C_PATH_SIZE) +
  			     nexus_sizeof_char(&buf,c2cSessions.count*C2C_EXE_SIZE) +
  			     nexus_sizeof_char(&buf,c2cSessions.count*C2C_ARGS_SIZE),
			     2*c2cSessions.count + 1);
  nexus_put_int(&buf,&c2cSessions.count,1);
  for(cursession = c2cSessions.sessions; /* Loop for everyone in the list */
      cursession != NULL;		/* Stop when list end is reached */
      cursession = cursession->next)	/* Send every node */
  { /* For */
    c2cPrintf(120,"sending a session along\n");
    nexus_put_int(&buf,&cursession->info.owner,1);
    nexus_put_char(&buf,cursession->info.name,C2C_NAME_SIZE);
    nexus_put_char(&buf,cursession->info.pathname,C2C_PATH_SIZE);
    nexus_put_char(&buf,cursession->info.execname,C2C_EXE_SIZE);
    nexus_put_char(&buf,cursession->info.args,C2C_ARGS_SIZE);
    nexus_put_int(&buf,&cursession->info.id,1);
  }
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sending session table to client\n");
} /* BrokerSessionCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes register requests on the client side
    
******************************************************************************/
void BrokerSubscribeCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerSubscribeCallback */
  nexus_buffer_t buf;			/* Data buffer */
  nexus_global_pointer_t gp;		/* Caller id */
  int ngp;				/* Size of a global pointer */
  CAVEId datasource;			/* Subscription source */
  StreamType stream;			/* Subscription stream */
  client *curclient;			/* Work pointer */
  int status = E_INVALID_CAVEID;	/* Error status */

  c2cPrintf(100,"interrupted by client for a subscription\n");
  nexus_get_global_pointer(buffer,&gp,1); /* Read client id */
  nexus_get_int(buffer,&datasource,1);	/* Read subscription source */
  nexus_get_int(buffer,&stream,1);	/* Read subscription stream */
  curclient = c2cClients.clients;       /* Start at beginning of list */
  c2cPrintf(100,"attempting to subscribe to client %d for stream %d\n",
            datasource,stream);
  if(curclient != NULL)
  { /* If */
    c2cPrintf(100,"searching for client %d in table\n",datasource);
    for(curclient = c2cClients.clients;	/* Find the requested source */
        (curclient != NULL) && (curclient->addr.rank != datasource);
        curclient = curclient->next);
  } /* If */
  if(curclient != NULL)			/* Source exists */
  { /* If */
    c2cPrintf(100,"sending subscription notice to %d\n",datasource);
    nexus_init_remote_service_request(&buf,&curclient->addr.gp,
				      SUBSCRIBE_STRING,
				      SUBSCRIBE_HASH);
    nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&gp,1,&ngp) +
			       nexus_sizeof_int(&buf,1) +
			       nexus_sizeof_int(&buf,1),
			       ngp + 2);
    nexus_put_global_pointer(&buf,&gp,1);
    nexus_put_int(&buf,&datasource,1);
    nexus_put_int(&buf,&stream,1);
  } /* If */
  else					/* Error-->send back to calling client */
  { /* Else */
    c2cPrintf(100,"client %d not found --> sending error notice to client\n",datasource);
    nexus_init_remote_service_request(&buf,&gp,
				      CLIENT_SUBSCRIBE_STRING,
				      CLIENT_SUBSCRIBE_HASH);
    nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,1),1);
    nexus_put_int(&buf,&status,1);
  } /* Else */
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sending notice to client\n");
} /* BrokerSubscribeCallback */

/*****************************************************************************  

          Params:  Address portion of global pointer
		   Data buffer
         Returns:  None
  Global Effects:  Entry is added client table

  This callback processes register requests on the client side
    
******************************************************************************/
void BrokerUnsubscribeCallback(void *address, nexus_buffer_t *buffer)
{ /* BrokerUnsubscribeCallback */
  nexus_buffer_t buf;			/* Data buffer */
  nexus_global_pointer_t gp;		/* Caller id */
  int ngp;				/* Size of a global pointer */
  CAVEId datasource;			/* Subscription source */
  StreamType stream;			/* Subscription stream */
  client *curclient;			/* Work pointer */
  int status = E_INVALID_CAVEID;	/* Error status */

  c2cPrintf(100,"interrupted by client for unsibscribe\n");
  nexus_get_global_pointer(buffer,&gp,1); /* Get calling client id */
  nexus_get_int(buffer,&datasource,1);	/* Get subscription source */
  nexus_get_int(buffer,&stream,1);	/* Get subscription stream type */
  curclient = c2cClients.clients;       /* Check list for client */
  if(curclient != NULL)                 /* List to check? */
  { /* If */
    for(curclient = c2cClients.clients;	/* Find the requested source */
        (curclient != NULL) && (curclient->addr.rank != datasource);
        curclient = curclient->next);
  } /* If */
  if(curclient != NULL)			/* Source exists */
  { /* If */
    c2cPrintf(100,"sending unsubscription messag to client %d\n",datasource);
    nexus_init_remote_service_request(&buf,&curclient->addr.gp,
				      UNSUBSCRIBE_STRING,
				      UNSUBSCRIBE_HASH);
    nexus_set_buffer_size(&buf,nexus_sizeof_global_pointer(&buf,&gp,1,&ngp) +
			       nexus_sizeof_int(&buf,1),
			       ngp + 1);
    nexus_put_global_pointer(&buf,&gp,1);
    nexus_put_int(&buf,&stream,1);
  } /* If */
  else					/* Error-->send back to calling client */
  { /* Else */
    c2cPrintf(100,"client %d not found --> sending error notice to client\n",datasource);
    nexus_init_remote_service_request(&buf,&gp,
				      CLIENT_UNSUBSCRIBE_STRING,
				      CLIENT_UNSUBSCRIBE_HASH);
    nexus_set_buffer_size(&buf,nexus_sizeof_int(&buf,1),1);
    nexus_put_int(&buf,&status,1);  
  } /* Else */
  nexus_send_remote_service_request(&buf); /* Send along to client */
  c2cPrintf(100,"finished sending notice to client\n");
} /* BrokerUnsubscribeCallback */

/*****************************************************************************  

          Params:  Argument count
                   Argument array
         Returns:  None
  Global Effects:  Broker is started

  This is the main function of the program
    
******************************************************************************/
void main(int argc,char *argv[])
{ /* main */
  unsigned short attachport = 0;	/* Port of broker on host */
  char *attachhost;			/* Name of host broker is running on */
  nexus_node_t *c2cNodes;		/* Local nexus nodes allocated */
  int c2cNumNodes;			/* Number of nexus nodes running */
  NexusAddr *addr;			/* Work pointer */
  int stat;				/* Daemon status indicator */
  int i;                                /* Loop counter */

  stat = fork();			/* Fork the daemon process */
  if(stat != 0)				/* Parent process */
  { /* If */
    exit(0);				/* Exit from parent process */
  } /* If */

  c2cPrintf(110,"begin processing c2c arguments\n");
  for(i=0;i<argc;i++)                   /* Parse arguments */
  { /* For */
    c2cPrintf(120,"processing argument %s\n",argv[i]);
    if(strcmp(argv[i],"-c2cdbg") == 0)   /* Match on the debugging arg */
    { /* If */
      c2cSetDebugLevel(atoi(argv[++i])); /* Assign the level */
      c2cPrintf(120,"printing debugging values up to %d\n",c2cGetDebugLevel());
    } /* If */
  } /* For */

  c2cPrintf(100,"initializing nexus transport layer\n");
  nexus_init(&argc,&argv,"NEXUS_ARGS","nx", /* Initialize nexus stuff */
             NULL,NULL,NULL,NULL,&c2cNodes,&c2cNumNodes);
  c2cPrintf(100,"starting nexus transport layer\n");
  nexus_start();			/* Start nexus stuff */
  c2cPrintf(100,"allowing attachments\n");
  nexus_allow_attach(&attachport,&attachhost,&NexusAllowAttachments);

  printf("\nc2cBroker://%s:%hu/\n",attachhost,attachport); /* Write URL to screen */

  c2cSessions.count = 0;		/* Initialize session table */
  c2cSessions.nextid = 0;
  c2cSessions.sessions = NULL;
  c2cClients.count = 0;			/* Initialize client table */
  c2cClients.nextid = 0;
  c2cClients.clients = NULL;

  c2cPrintf(100,"sleeping waiting for control requests\n");
  NexusInit(hExit);                     /* Prepare for wait */
  NexusWait(hExit);			/* Wait for an exit event to occur */
  c2cPrintf(100,"broker awakened and exiting\n");

  c2cPrintf(100,"killing all connections\n");
  for(addr = &c2cClients.clients->addr;	/* Destroy all connections */
      addr != NULL;			/* Stop at the end of the list */
      addr = addr->next)		/* For each node */
  { /* For */
    c2cPrintf(120,"killing a connection\n");
    nexus_destroy_global_pointer(&addr->gp); /* Kill global pointer */
  } /* For */
} /* main */
