/******************************************************************************

  c2cdevice.h
  
  Written by:  Michael Pellegrino
	       Indiana Institute of Technology

  The purpose of this file is to provide the interface between the API and the
  communications device used.  This file is specific to each particular
  communcations device so Nexus has its own, MPI has its own, etc...

  Date last edited:  August 4, 1995

  Version:  .12

******************************************************************************/
#include <nexus.h>		/* Used for all nexus particular stuff */

/* This macro sets up a particular handler */
#define NexusInit(handler) \
        { \
	  nexus_mutex_init(&handler.mutex,(nexus_mutexattr_t *)NULL); \
	  nexus_cond_init(&handler.cond,(nexus_condattr_t *)NULL); \
	  handler.done = NEXUS_FALSE; \
	}

/* This macro waits on a particular condition and then destroys handler */
#define NexusWait(handler) \
	{ \
	  nexus_mutex_lock(&handler.mutex); \
	  while(!handler.done) \
	  { \
	    nexus_cond_wait(&handler.cond,&handler.mutex); \
	  } \
	  nexus_mutex_unlock(&handler.mutex); \
	  nexus_mutex_destroy(&handler.mutex); \
	  nexus_cond_destroy(&handler.cond); \
	}

/* This macro awakens a process waiting on a particular condition */
#define NexusWake(handler) \
	{ \
	  nexus_mutex_lock(&handler.mutex); \
	  handler.done = NEXUS_TRUE; \
	  nexus_cond_signal(&handler.cond); \
	  nexus_mutex_unlock(&handler.mutex); \
	}

/* Define hashes for the nexus library callbacks */
#define BROKER_CLIENT_STRING		"BrokerClientCallback"
#define BROKER_CLIENT_HASH		980
#define BROKER_KILL_STRING		"BrokerKillCallback"
#define BROKER_KILL_HASH		647
#define BROKER_NEWSESSION_STRING	"BrokerNewSessionCallback"
#define BROKER_NEWSESSION_HASH		390
#define BROKER_REGISTER_STRING		"BrokerRegisterCallback"
#define BROKER_REGISTER_HASH		189
#define BROKER_SESSION_STRING		"BrokerSessionCallback"
#define BROKER_SESSION_HASH		92
#define BROKER_SUBSCRIBE_STRING		"BrokerSubscribeCallback"
#define BROKER_SUBSCRIBE_HASH		282
#define BROKER_UNREGISTER_STRING	"BrokerUnregisterCallback"
#define BROKER_UNREGISTER_HASH		416
#define BROKER_UNSUBSCRIBE_STRING	"BrokerUnsubscribeCallback"
#define BROKER_UNSUBSCRIBE_HASH		509
#define CLIENT_NEWSESSION_STRING	"ClientNewSessionCallback"
#define CLIENT_NEWSESSION_HASH		384
#define CLIENT_REGISTER_STRING		"ClientRegisterCallback"
#define CLIENT_REGISTER_HASH		183
#define CLIENT_SESSION_STRING		"ClientSessionCallback"
#define CLIENT_SESSION_HASH		86
#define CLIENT_SUBSCRIBE_STRING		"ClientSubscribeCallback"
#define CLIENT_SUBSCRIBE_HASH		276
#define CLIENT_UNREGISTER_STRING	"ClientUnregisterCallback"
#define CLIENT_UNREGISTER_HASH		410
#define CLIENT_UNSUBSCRIBE_STRING	"ClientUnsubscribeCallback"
#define CLIENT_UNSUBSCRIBE_HASH		503
#define CLIENT_USER_STRING		"ClientUserCallback"
#define CLIENT_USER_HASH		782
#define STREAM_CALLBACK_STRING		"StreamCallback"
#define STREAM_CALLBACK_HASH		380
#define SUBSCRIBE_STRING		"SubscribeCallback"
#define SUBSCRIBE_HASH			690
#define UNSUBSCRIBE_STRING		"UnsubscribeCallback"
#define UNSUBSCRIBE_HASH		917

/* Miscellaneous defines */
#define SHARED_MEM_SIZE		10000	/* Shared memory pool size */
#ifndef FALSE
#define FALSE			0	/* Flag values */
#endif
#ifndef TRUE
#define TRUE			!FALSE
#endif
#define WAND_END_OFFSET   0.10		/* Wand drawing information */
#define WAND_START_OFFSET 0.05		/* Used for the wand routine written by */
#define WAND_LENGTH       0.25		/* Dave Pape (see worlddraw.c) */

/* Structure designed for contacting a remote process */
typedef struct nexusaddr NexusAddrPtr;

typedef struct nexusaddr
{
  nexus_global_pointer_t gp;		/* Communication pointer to broker*/
  int rank;				/* Rank of node in table */
  int caveid;				/* CAVEID of local process as seen by remote process */
  NexusAddrPtr *next;			/* Next node in list of addresses */
}NexusAddr;

/* List of process to contact */
typedef struct conntab ConntabPtr;

typedef struct conntab
{
  int count;				/* Number of people connected */
  int nextid;				/* Next id issued in table */
  NexusAddrPtr *addr;			/* Nexus address of the rank */
}Conntab;

/* Structure maintains all stream callbacks */
typedef struct streamlist StreamlistPtr;

typedef struct streamlist
{
  int type;				/* Streamtype of particular callback */
  int source;				/* Id of the the sender of the stream */
  void *callback;			/* Pointer to particular function t execute*/
  StreamlistPtr *nexttype;			/* Next node in the linked list */
  StreamlistPtr *nextfunc;			/* Next local function in the list */
}Streamlist;

/* Structure maintains all cast information for all streams */
typedef struct castlist CastlistPtr;

typedef struct castlist
{
  int type;				/* Stream type of broadcast to perform */
  Conntab conntab;			/* All connections for particular stream type */
  int frequency;			/* Frequency of skips */
  int autofrequency;			/* Flag for automatic frequency setting */
  CastlistPtr *next;			/* Next node in the linked list */
  void *callback;                       /* Callback to execute when stream is subscribed to */
}Castlist;

/* Contains all pieces to allow for a process to sleep on a condition */
typedef struct handler
{
  nexus_mutex_t mutex;			/* Mutex flag */
  nexus_cond_t cond;			/* Condition flag */
  int done;				/* Completion flag */
  nexus_stashed_buffer_t buffer;	/* Buffer for data */
}Handler;

/* Contains all pieces to allow for higher level c2c functionality */
typedef struct c2cworldinfo
{
  int hostid;				/* Id of the host the world connects to */
  int caveid;				/* CAVE id of local process on host */
  int memid;				/* Shared memory pool id */
  void *memptr;				/* Pointer to shared memory pool */
}c2cWorldInfo;

/* Structure for tracking a user */
typedef struct trackuser TrackUserPtr;

typedef struct trackuser
{
  void *callback;		/* Callback for the drawing function */
  CaveEnv env;			/* Environment of the tracked user */
  CAVEUser user;		/* User information on remote CAVE */
  TrackUserPtr *next;		/* Next node in linked list */
}TrackUser;

/* Control information for data casting */
typedef struct castcontrol
{
  nexus_mutex_t mutex;		/* Mutual exclusion flag */
  int done;			/* Completion flag */
  CAVEUser user;		/* Local CAVE user */
}CastControl;

/* Nexus specific function prototypes */
extern int NexusBoot(void);
extern int NexusAllowAttachments(char *url, nexus_global_pointer_t *gp);
extern void NexusUnknownHandler(void *address,nexus_buffer_t *buffer,char *handler_name,int handler_id);

/* CAVE-to-CAVE utility function prototypes */
extern void c2cLowerCase(char *str);
extern void c2cTrackUserCallback(CAVEUser *user,int id);
