/******************************************************************************

  c2cworld.c
  
  Written by:  Michael Pellegrino
	       Indiana Institute of Technology

  The purpose of this file is to provide the user interface for c2c
  world communications.

  Date last edited:  August 4, 1995

  Version:  .12

******************************************************************************/
#ifdef C2C_CAVE_PRESENT
#include <cave.h>
#endif
#include <malloc.h>
#include <math.h>
#include <nexus.h>
#include <stdio.h>
#include <string.h>
#include "c2c.h"
#include "c2cdevice.h"

/* Declare all global variables */
extern CastControl c2cCastControl;		/* Control structure for CAVE casting */
extern TrackUser **c2cUserList;			/* List of all users to track */
extern Streamlist *c2cStreams;			/* Stream callback list for all streams */
extern CaveSession c2cSession;			/* Local session */
extern c2cWorldInfo c2cWorld;			/* World information structure */

/*****************************************************************************  
  
          Params:  Argument count
                   Argument list
         Returns:  Nothing
  Global Effects:  None

  This function initializes all c2c world functionality
 
******************************************************************************/
void c2cWorldInit(int *argc,char *argv[])
{ /* c2cWorldInit */
  c2cConfig config;			/* c2c configuration information */
  nexus_thread_t thread;		/* Thread for sub-process */
  int stat;                             /* Result of nexus thread create */

  c2cPrintf(100,"starting c2cWorldInit\n");
  c2cPrintf(100,"reading configuration file\n");
  c2cReadConfigFile(NULL,&config);	/* Get setup information */
  c2cPrintf(100,"starting c2c functionality\n");
  c2cInit(argc,argv,&config.env); /* Start c2c functionality */
  c2cPrintf(100,"attaching to broker\n");
  c2cWorld.hostid = c2cBrokerAttach(config.broker); /* Attach to broker */
  c2cPrintf(100,"registering with broker as %s\n",c2cEnv.name);
  c2cWorld.caveid = c2cRegister(c2cWorld.hostid,&c2cEnv); /* Register with broker */
  if( (c2cEnv.type == ST_CAVE) || (c2cEnv.type == ST_IMMERSADESK) || /* Create tracker thread */
      (c2cEnv.type == ST_CAVE_SIMULATOR) )
  { /* If */
    c2cPrintf(100,"creating casting thread for application\n");
    if(c2cUserList != NULL)             /* Make sure c2cWorldDataInit was called */
    { /* If */
      c2cRegisterStream(S_HEAD_TRACKER,NULL);	/* Register head tracker stream */
      c2cRegisterStream(S_WAND_TRACKER,NULL);	/* Register wand tracker stream */
      c2cRegisterStream(S_WAND_BUTTON,NULL);	/* Register wand button  stream */
      c2cRegisterStream(S_WAND_JOYSTICK,NULL);	/* Register wand joystick stream */
      c2cRegisterStream(S_WORLD_POSITION,NULL); /* Register world position stream */
      c2cRegisterStream(S_CAVE_USER,NULL);	/* Register CAVE user stream */
      nexus_mutex_init(&c2cCastControl.mutex,   /* Initialize mutual exclution flag */
                       (nexus_mutexattr_t *)NULL);
      stat = nexus_thread_create(&thread,NULL,(nexus_thread_func_t)c2cCAVECast,NULL); /* Create a sub-process */
      if(stat != 0)			/* Error on thread creation */
      { /* If */
        printf("Warning --> Casting thread was not created.  Return code is %d.\n",stat);
      } /* If */
    } /* If */
    else                                /* Give warning that there will be problems */
    { /* Else */
      printf("Warning --> c2cWorldDataInit never called!\n",
	     "            To use the CAVE-to-CAVE tracking functions,\n",
	     "            you must make a call to c2cWorldDataInit or,\n",
	     "            your program might crash.\n");
    } /* Else */
  } /* If */
  c2cPrintf(100,"completed c2cWorldInit\n");
} /* c2cWorldInit */

/*****************************************************************************  
  
          Params:  Argument count
                   Argument list
         Returns:  Nothing
  Global Effects:  None

  This function terminates all c2c world functionality
 
******************************************************************************/
void c2cWorldExit(void)
{ /* c2cWorldExit */
  c2cPrintf(100,"attempting to close c2c via world exit\n");
  c2cUnregister(c2cWorld.hostid,c2cWorld.caveid);	/* Unregister */
  c2cBrokerDetach(c2cWorld.hostid);			/* Forget about the broker */
  if( (c2cEnv.type == ST_CAVE) || (c2cEnv.type == ST_IMMERSADESK) || /* Tell thread to go away */
      (c2cEnv.type == ST_CAVE_SIMULATOR) )
  { /* If */
    c2cPrintf(100,"signalling the casting thread to die\n");
    nexus_mutex_lock(&c2cCastControl.mutex);		/* Enforce mutual exclusion */
    c2cCastControl.done = TRUE;				/* Tell thread to exit */
    nexus_mutex_unlock(&c2cCastControl.mutex);		/* Cancel mutual exclusion */
    c2cPrintf(100,"finished signalling the casting thread to die\n");
  } /* If */
  c2cTerminate();					/* Exit c2c application */
  c2cPrintf(100,"terminating c2c world functionality\n");
} /* c2cWorldExit */

/*****************************************************************************  

          Params:  User arguments (not used)
         Returns:  None
  Global Effects:  CAVE data is streamed everywhere

  This process casts all CAVE streams to all subscribed to them
    
******************************************************************************/
void c2cCAVECast(void *user_args)
{ /* c2cCAVECast */
#ifdef C2C_CAVE_PRESENT
  static float worldorigin[3] = {0,0,0}; /* World location in CAVE */
  static float worldorient[3] = {1,1,1}; /* World orientation vector */
  float worldpos[3];                     /* World location */
  float worldangle[3];                   /* World orientation angles */
#endif
  c2cPrintf(120,"beginning the casting of tracker data\n");
  while(!c2cCastControl.done)		/* Spin until parent exits */
  { /* While */
    c2cPrintf(110,"deriving CAVE library tracker values\n");
    nexus_mutex_lock(&c2cCastControl.mutex);	/* Enforce mutual exclusion */
#ifdef C2C_CAVE_PRESENT
    CAVEGetHead(c2cCastControl.user.headx,c2cCastControl.user.heady, /* Get head values */
                c2cCastControl.user.headz);
    CAVEGetWand(c2cCastControl.user.wandx,c2cCastControl.user.wandy, /* Get wand values */
                c2cCastControl.user.wandz);
    CAVEGetHeadOrientation(c2cCastControl.user.heada,c2cCastControl.user.heade, /* Get head values */
                           c2cCastControl.user.headr);
    CAVEGetWandOrientation(c2cCastControl.user.wanda,c2cCastControl.user.wande, /* Get wand values */
                           c2cCastControl.user.wandr);
    c2cCastControl.user.but1 = CAVEBUTTON1;	/* Get button values */
    c2cCastControl.user.but2 = CAVEBUTTON2;	/* Get button values */
    c2cCastControl.user.but3 = CAVEBUTTON3;	/* Get button values */
    c2cCastControl.user.joyx = CAVE_JOYSTICK_X; /* Get joystick values */
    c2cCastControl.user.joyy = CAVE_JOYSTICK_Y; /* Get joystick values */
    CAVENavConvertCAVEToWorld(worldorigin,worldpos); /* Calculate world location */
    c2cCastControl.user.worldx = worldpos[0];   /* Assign world location */
    c2cCastControl.user.worldy = worldpos[1];
    c2cCastControl.user.worldz = worldpos[2];
    CAVENavConvertVectorCAVEToWorld(worldorient,worldangle); /* Get world vector */
/*    c2cCastControl.user.worlda = atan2f(worldangle[2],worldangle[0]) * (180/M_PI);
    c2cCastControl.user.worlde = atan2f(worldangle[1],worldangle[2]) * (180/M_PI);
    c2cCastControl.user.worldr = atan2f(worldangle[1],worldangle[0]) * (180/M_PI); */
    c2cPrintf(110,"finished deriving CAVE library tracker values\n");
#endif
    c2cSendHeadTracker(c2cCastControl.user.headx,c2cCastControl.user.heady,
		       c2cCastControl.user.headz,c2cCastControl.user.heada,
		       c2cCastControl.user.heade,c2cCastControl.user.headr);
    c2cSendWandTracker(c2cCastControl.user.wandx,c2cCastControl.user.wandy,
		       c2cCastControl.user.wandz,c2cCastControl.user.wanda,
		       c2cCastControl.user.wande,c2cCastControl.user.wandr);
    c2cSendWandButtons(c2cCastControl.user.but1,c2cCastControl.user.but2,
		       c2cCastControl.user.but3,c2cCastControl.user.but4);
    c2cSendWandJoystick(c2cCastControl.user.joyx,c2cCastControl.user.joyy);
    c2cSendWorldPosition(c2cCastControl.user.worldx,c2cCastControl.user.worldy,
			 c2cCastControl.user.worldz,c2cCastControl.user.worlda,
			 c2cCastControl.user.worlde,c2cCastControl.user.worldr);
    c2cSendUser(&c2cCastControl.user);
    nexus_mutex_unlock(&c2cCastControl.mutex);	/* Cancel mutual exclusion */
  } /* While */
} /* c2cCAVECast */

/*****************************************************************************  

          Params:  None
         Returns:  None
  Global Effects:  None

  This function draws all users tracked via c2cTrackUserInit
    
******************************************************************************/
void c2cDrawAllUsers(void)
{ /* c2cDrawAllUsers */
  CAVEUserP UserCallback;		/* CAVEUser callback */
  TrackUser *user;			/* Work pointer */

  c2cPrintf(100,"drawing all users in the tracking list\n");
  for(user = *c2cUserList;		/* Draw all users */
      user != NULL;
      user = user->next)
  { /* For */
    c2cPrintf(120,"drawing a user\n");
    UserCallback = (CAVEUserP)user->callback;	/* Cast the callback */
    UserCallback(&user->user,user->env.id);
  } /* For */
  c2cPrintf(100,"finished drawing all users\n");
} /* c2cDrawAllUsers */

/*****************************************************************************  

          Params:  Name of a CAVE to draw
         Returns:  None
  Global Effects:  None

  This function draws a user specified by his registered client name
    
******************************************************************************/
int c2cDrawUser(char *user)
{ /* c2cDrawUser */
  CAVEUserP UserCallback;		/* CAVEUser callback */
  TrackUser *curuser;			/* Work pointer */

  c2cPrintf(100,"attempting to draw user %s\n",user);
  for(curuser = *c2cUserList;	/* Find our user */
      (curuser != NULL) && (strcmp(user,curuser->env.name) != 0);
      curuser = curuser->next);
  if(curuser != NULL)			/* Found it */
  { /* If */
    c2cPrintf(100,"found the user to draw\n");
    UserCallback = (CAVEUserP)curuser->callback;/* Cast the callback */
    UserCallback(&curuser->user,curuser->env.id);
  } /* If */
  else					/* Not found in list */
  { /* Else */
    c2cPrintf(100,"user %s is not found in the track list\n",user);
    return(E_INVALID_CLIENT);           /* Return error */
  } /* Else */
  c2cPrintf(100,"finished drawing user\n");
  return(E_SUCCESS);                    /* Return success */
} /* c2cDrawUser */

/*****************************************************************************  

          Params:  Name of a CAVE to draw
         Returns:  None
  Global Effects:  None

  This function draws a list of specified users
    
******************************************************************************/
void c2cDrawSomeUsers(int count,char *users[])
{ /* c2cDrawSomeUsers */
  int i;				/* Loop counter */

  c2cPrintf(100,"drawing a list of users\n");
  for(i=0;i<count;i++)			/* For every user in list */
  { /* For */
    c2cDrawUser(users[i]);		/* Draw the user */
  } /* For */
  c2cPrintf(100,"finished drawing a list of users\n");
} /* c2cDrawUsers */

/*****************************************************************************  

          Params:  Name of a user to track
                   Drawing function callback
         Returns:  None
  Global Effects:  User of the remote CAVE is tracke

  This function adds a user to list of users to track.  The remote
  environment will start streaming appropriate data to be drawn by
  callback.  If callback is NULL, a default drawing function is used.
    
******************************************************************************/
int c2cTrackUserInit(char *user,void *callback)
{ /* c2cTrackUserInit */
  int i;				/* Loop counter */
  int numclients;			/* Number of clients on the broker */
  CaveEnv *env;				/* List of all environments on broker */
  CAVEId id;				/* Id of CAVE user to track */
  TrackUser *newuser;			/* Work pointer */
  int stat;				/* Error flag */

  c2cPrintf(100,"attempting to track user %s for drawing with %p\n",user,callback);
  c2cGetClients(c2cWorld.hostid,&numclients,&env); /* Get list of everyone registered */
  id = (-1);				/* Plant a known elephant in Cairo */
  for(i=0;i<numclients;i++)		/* Search for our user */
  { /* For */
    if(strcmp(env[i].name,user) == 0)	/* Found our match */
    { /* If */
      id = env[i].id;			/* Get the rank of the client */
      break;				/* Get out of here... */
    } /* If */
  } /* For */
  if(id != (-1))			/* Found our client */
  { /* If */
    c2cPrintf(110,"adding user to track list\n");
    newuser = (TrackUser *)c2cSharedAlloc(sizeof(TrackUser));	/* Get some memory */
    newuser->next = *c2cUserList;
    *c2cUserList = newuser;		/* Insert into list */
    memcpy(&newuser->env,&env[i],sizeof(CaveEnv)); /* Copy it into node */
    if(callback == NULL)		/* NULL --> Use default procedure */
    { /* If */
      c2cPrintf(110,"using default c2cDrawStickMan for user rendering\n");
      newuser->callback = c2cDrawStickMan;	/* Set pointer appropriately */
    } /* If */
    else					/* Assign passed callback */
    { /* Else */
      c2cPrintf(110,"using user supplied callback at address %p\n",callback);
      newuser->callback = callback;		/* Point to callback */
    } /* Else */
    stat = c2cSubscribe(c2cWorld.hostid,id,     /* Subscribe to user stream */
                        S_CAVE_USER,c2cTrackUserCallback);
  } /* If */
  else					/* Error because client wasn't registered */
  { /* Else */
    stat = E_INVALID_CLIENT;            /* Set error flag */
  } /* Else */
  c2cPrintf(100,"returning from operations with result %d\n",stat);
  return(stat);                         /* Send back status indicator */
} /* c2cTrackUserInit */

/*****************************************************************************  

          Params:  User arguments (not used)
         Returns:  None
  Global Effects:  CAVE data is streamed everywhere

  This function removes a user from list of users to track
    
******************************************************************************/
int c2cTrackUserExit(char *user,void *callback)
{ /* c2cTrackUserExit */
  TrackUser *prevuser,*curuser;		/* Work pointers */

  c2cPrintf(100,"attempting to remove user %s with callback %p from track list\n",user,callback);
  if(callback == NULL)			/* Check to see if callback is default one */
  { /* If */
    c2cPrintf(110,"using default c2cDrawStickMan for searching\n");
    callback = c2cDrawStickMan;		/* Set to default callback */
  } /* If */
  prevuser = *c2cUserList;		/* Start at beginning */
  if(prevuser != NULL)			/* List has something */
  { /* If */
    if((strcmp(user,prevuser->env.name) == 0) && /* First node? */
       (callback == prevuser->callback) )
    { /* If */
      c2cPrintf(110,"removing the user from the track list\n");
      c2cUnsubscribe(c2cWorld.hostid,prevuser->env.id,S_CAVE_USER,c2cTrackUserCallback); /* Unsubscribe from stream */
      *c2cUserList = prevuser->next;	/* Update pointers */
      c2cSharedFree(prevuser);	       	/* Release node memory */
    } /* If */
    else				/* Start searching */
    { /* Else */
      for(curuser = prevuser->next;
          (curuser != NULL) && (strcmp(user,prevuser->env.name) == 0) && 
	  (callback == curuser->callback);
	  curuser = curuser->next, prevuser = prevuser->next);
      if(curuser != NULL)		/* Found what we want */
      { /* If */
        c2cPrintf(110,"removing the user from the track list\n");
        c2cUnsubscribe(c2cWorld.hostid,curuser->env.id,S_CAVE_USER,c2cTrackUserCallback); /* Unsubscribe from stream */
        prevuser->next = curuser->next;	/* Update pointers */
	c2cSharedFree(curuser);	       	/* Release node memory */
      } /* If */
      else				/* Not in this list */
      { /* Else */
        c2cPrintf(100,"user was not found in the track list\n");
        return(E_INVALID_CLIENT);       /* Send back error */
      } /* Else */
    } /* Else */
  } /* If */
  else					/* List empty */
  { /* Else */
    c2cPrintf(100,"user was not found in the track list\n");
    return(E_INVALID_CLIENT);           /* Send back error */
  } /* Else */
  c2cPrintf(100,"finished removing user from track list\n");
  return(E_SUCCESS);                    /* Send back success */
} /* c2cTrackUserExit */

/*****************************************************************************  

          Params:  User to un/subscribe to (application name)
		   Known stream to un/subscribe to
		   Callback to use upon data receipt
         Returns:  None
  Global Effects:  Subscribed to stream listed

  This function un/subscribes to a stream posted by user
    
******************************************************************************/
int c2cWorldSubscribe(char *user,StreamType stream,void *callback,int action)
{ /* c2cWorldSubscribe */
  int i;				/* Loop counter */
  int numclients;			/* Number of clients on the broker */
  CaveEnv *env;				/* List of all environments on broker */
  CAVEId id;				/* Id of CAVE user to track */
  int stat;                             /* Status of the entire operation */

  c2cPrintf(100,"attempting to subscribe to %s for %d with callback %p and action %d\n",
            user,stream,callback,action);
  c2cGetClients(c2cWorld.hostid,&numclients,&env); /* Get list of everyone registered */
  id = (-1);				/* Plant a known elephant in Cairo */
  for(i=0;i<numclients;i++)		/* Search for our user */
  { /* For */
    if(strcmp(env[i].name,user) == 0)	/* Found our match */
    { /* If */
      id = env[i].id;			/* Get the rank of the client */
      break;				/* Get out of here... */
    } /* If */
  } /* For */
  if(id != (-1))			/* Found our client */
  { /* If */
    if(action == C2C_SUBSCRIBE)		/* Subscribe to the stream */
    { /* If */
      stat =  c2cSubscribe(c2cWorld.hostid,id,stream,callback); /* Subscribe to user stream */
    } /* If */
    else				/* Unsubscribe to stream */
    { /* Else */
      stat = c2cUnsubscribe(c2cWorld.hostid,id,stream,callback); /* Subscribe to user stream */
    } /* Else */
  } /* If */
  else					/* Error because client wasn't registered */
  { /* Else */
    stat = E_INVALID_CLIENT;            /* Set error flag */
  } /* Else */
  c2cPrintf(100,"returning from operation with result of %d\n",stat);
  return(stat);                         /* Send back error flag */
} /* c2cWorldSubscribe */

/*****************************************************************************  

          Params:  None
         Returns:  None
  Global Effects:  Shared memory arena specific to CAVE-to-CAVE library is
                   created

  This function creates a shared memory pool that is used to store track-user
  data.  While this function represents a serious hack on the design side,
  it does provide what is needed to be accomplished.
    
******************************************************************************/
void c2cWorldDataInit(void)
{ /* c2cWorldDataInit */
  c2cPrintf(100,"creating shared memory pool\n");
  c2cSharedInit();                      /* Initialize shared memory */
  c2cPrintf(100,"creating user track list\n");
  c2cUserList = (TrackUser **)c2cSharedAlloc(sizeof(TrackUser *));/* Get some memory */
  *c2cUserList = (TrackUser *)NULL;	/* Initialize user list */
} /* c2cWorldDataInit */

/*****************************************************************************  

          Params:  Client name to query
         Returns:  Id of the client on the broker c2cWorld
  Global Effects:  None

  This function returns the id of the client name on the world broker
    
******************************************************************************/
int c2cWorldGetClientId(char *user)
{ /*( c2cWorldGetClientId */
  return(c2cGetClientId(c2cWorld.hostid,user));
} /* c2cWorldGetClientId */

/*****************************************************************************  

          Params:  Session name to query
         Returns:  Id of the session on the broker c2cWorld
  Global Effects:  None

  This function returns the id of the session name on the world broker
    
******************************************************************************/
int c2cWorldGetSessionId(char *user)
{ /*( c2cWorldGetSessionId */
  return(c2cGetSessionId(c2cWorld.hostid,user));
} /* c2cWorldGetSessionId */
