/*
 * handler.c
 *
 * Handler manipulation routines.
 */

static char *rcsid = "$Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/handler.c,v 1.27 1995/06/12 22:11:16 patton Exp $";

#include "internal.h"

#ifndef NEXUS_LITE
typedef struct _threaded_handler_startup_info_t
{
    nexus_context_t *			context;
    void *				address;
#ifdef NEXUS_PROFILE
    int					node_id;
    int					context_id;
    int					message_length;
#endif	    
    nexus_stashed_buffer_t		buffer;
    nexus_handler_record_t *		rec;
    nexus_bool_t				started;
    nexus_mutex_t			mutex;
    nexus_cond_t			cond;
} threaded_handler_startup_info_t;

static void *threaded_handler_startup(void *arg);
#endif /* NEXUS_LITE */


static nexus_bool_t	arg_got_hash;
static int		handler_table_size;

#ifdef NEXUS_PROFILE
static nexus_bool_t	arg_got_rsr_hash;
static int		profile_rsr_table_size;
#endif

static void insert_handler_record(nexus_handler_record_t **handler_table,
				  char *name,
				  int id,
				  nexus_handler_type_t type,
				  nexus_handler_func_t func);


/*
 * _nx_handler_usage_message()
 */
void _nx_handler_usage_message()
{
    printf("    -hash <integer>           : Handler hash table size.  (Do not use this\n");
    printf("                                unless you know what you are doing...)\n");
#ifdef NEXUS_PROFILE
    printf("    -rsr_hash <integer>       : Remote service request profile hash\n");
    printf("                                table size.\n");
#endif /* NEXUS_PROFILE */
} /* _nx_handler_usage_message() */


#ifndef NEXUS_LITE
/*
 * _nx_handler_new_process_params()
 */
int _nx_handler_new_process_params(char *buf, int size)
{
    char tmp_buf1[1024];
#ifdef NEXUS_PROFILE
    char tmp_buf2[1024];
#endif    
    int n_added;

    tmp_buf1[0] = '\0';

    if (arg_got_hash)
    {
	sprintf(tmp_buf1, "-hash %d ", handler_table_size);
    }

#ifdef NEXUS_PROFILE
    if (arg_got_rsr_hash)
    {
	sprintf(tmp_buf2, "-rsr_hash %d ", profile_rsr_table_size);
	strcat(tmp_buf1, tmp_buf2);
    }
#endif /* NEXUS_PROFILE */

    n_added = strlen(tmp_buf1);
    if (n_added > size)
    {
	nexus_fatal("_nx_handler_new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }

    strcpy(buf, tmp_buf1);
    
    return (n_added);

} /* _nx_handler_new_process_params() */
#endif /* NEXUS_LITE */


/*
 * _nx_handler_init()
 *
 * Initialize the handler table for the passed 'context'.
 */
void _nx_handler_init(int *argc,
		      char ***argv)
{
    int arg_num;
    
    if ((arg_num = nexus_find_argument(argc, argv, "hash", 2)) >= 0)
    {
	handler_table_size = atoi((*argv)[arg_num + 1]);
	arg_got_hash = NEXUS_TRUE;
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	handler_table_size = NEXUS_HANDLER_HASH_TABLE_SIZE;
	arg_got_hash = NEXUS_FALSE;
    }
    
#ifdef NEXUS_PROFILE
    if ((arg_num = nexus_find_argument(argc, argv, "rsr_hash", 2)) >= 0)
    {
	profile_rsr_table_size = atoi((*argv)[arg_num + 1]);
	arg_got_rsr_hash = NEXUS_TRUE;
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	profile_rsr_table_size = NEXUS_PROFILE_RSR_HASH_TABLE_SIZE;
	arg_got_rsr_hash = NEXUS_FALSE;
    }
#endif
} /* _nx_handler_init() */


/*
 * _nx_handler_init_for_context()
 *
 * Initialize the handler table for the passed 'context'.
 */
void _nx_handler_init_for_context(nexus_context_t *context)
{
    int i;
    nexus_handler_record_t **handler_table;
    
    /* Initialize the handler table */
    NexusMalloc(_nx_handler_init_for_context(),
		handler_table,
		nexus_handler_record_t **,
		(sizeof(nexus_handler_record_t *) * handler_table_size) );
    for (i = 0; i < handler_table_size; i++)
    {
	handler_table[i] = (nexus_handler_record_t *) NULL;
    }

    /* Add the handler table to the context */
    context->handler_table = handler_table;

} /* _nx_handler_init_for_context() */


/*
 * _nx_handler_destroy()
 *
 * Free up the handler table for the passed context.
 */
void _nx_handler_destroy(nexus_context_t *context)
{
    nexus_handler_record_t **handler_table;
    nexus_handler_record_t *rec, *next_rec;
    int i;

    handler_table = context->handler_table;
    
    for (i = 0; i < handler_table_size; i++)
    {
	rec = handler_table[i];
	while (rec)
	{
	    next_rec = rec->next;
	    NexusFree(rec->name);
	    NexusFree(rec);
	    rec = next_rec;
	}
    }

    NexusFree(handler_table);
} /* _nx_handler_destroy() */


/*
 * insert_handler_record()
 *
 * Malloc and fill in a new handler record.
 * Add it to the 'id' bucket of the handler_table.
 */
static void insert_handler_record(nexus_handler_record_t **handler_table,
				  char *name,
				  int id,
				  nexus_handler_type_t type,
				  nexus_handler_func_t func)
{
    nexus_handler_record_t *rec;
    
    NexusAssert2(((id >= 0) && (id < handler_table_size)),
		 ("insert_handler_record(): Handler id (%d) out of range (%d-%d).\n", id, 0, handler_table_size) );
    
    NexusMalloc(insert_handler_record(), rec, nexus_handler_record_t *,
		sizeof(nexus_handler_record_t) );
    
    rec->name = _nx_copy_string(name);
    rec->id = id;
    rec->type = type;
    rec->func = func;
    rec->next = (nexus_handler_record_t *) NULL;
#ifdef NEXUS_PROFILE
    rec->profile_table = (nexus_profile_rsr_record_t **) NULL;
#endif    

    rec->next = handler_table[id];
    handler_table[id] = rec;

} /* insert_handler_record() */


/*
 * nexus_handler_hash()
 */
int nexus_handler_hash(char *name)
{
    int hash;
    char *s;

    NexusAssert2((name != (char *) NULL),
		 ("nexus_handler_hash(): Passed a NULL handler name\n"));
    hash = 0;
    for (s = name; *s != '\0'; s++)
	hash += (int) *s;
    hash = hash % NEXUS_HANDLER_HASH_TABLE_SIZE;
    
    return (hash);
} /* nexus_handler_hash() */


/*
 * nexus_register_handlers()
 */
void nexus_register_handlers(nexus_handler_t *handlers)
{
    int i;
    nexus_context_t *context;
    nexus_handler_record_t **handler_table;

    _nx_context(&context);
    handler_table = context->handler_table;
    
    for (i = 0; handlers[i].func; i++)
    {
#ifdef NEXUS_LITE
	if (handlers[i].type != NEXUS_HANDLER_TYPE_NON_THREADED)
	{
	    nexus_fatal("nexus_register_handlers(): All handlers must be non-threaded with NexusLite.  %s is not.\n", handlers[i].name);
	}
#endif /* NEXUS_LITE */
	insert_handler_record(handler_table,
			      handlers[i].name,
			      handlers[i].id,
			      handlers[i].type,
			      handlers[i].func);
	nexus_debug_printf(2, ("nexus_register_handlers(): Handler:[%s] registered\n",handlers[i].name));
    }
} /* nexus_register_handlers() */


/*
 * nexus_substitute_handler()
 */
void nexus_substitute_handler(char *name,
			      int id,
			      nexus_handler_type_t new_type,
			      nexus_handler_func_t new_func,
			      nexus_handler_type_t *old_type,
			      nexus_handler_func_t *old_func)
{
    nexus_handler_record_t *rec;
    nexus_context_t *context;
    nexus_handler_record_t **handler_table;

    NexusAssert2(((id >= 0) && (id < handler_table_size)),
		 ("nexus_substitute_handler(): Handler id (%d) out of range (%d-%d).\n", id, 0, handler_table_size) );

#ifdef NEXUS_LITE
    if (new_type != NEXUS_HANDLER_TYPE_NON_THREADED)
    {
	nexus_fatal("nexus_substitute_handler(): All handlers must be non-threaded with NexusLite.  %s is not.\n", name);
    }
#endif /* NEXUS_LITE */

    _nx_context(&context);
    handler_table = context->handler_table;
    
    for (rec = handler_table[id];
	 (   (rec != (nexus_handler_record_t *) NULL)
	  && (strcmp(name,rec->name) != 0) );
	 rec = rec->next)
	;

    if (rec != (nexus_handler_record_t *) NULL)
    {
	*old_type = rec->type;
	*old_func = rec->func;
	rec->type = new_type;
	rec->func = new_func;
    }
    else
    {
	*old_func = (nexus_handler_func_t) NULL;
	insert_handler_record(handler_table, name, id, new_type, new_func);
    }
} /* nexus_substitute_handler() */


/*
 * _nx_lookup_handler()
 *
 * Lookup the handler ('handler_name', 'handler_id') in the handler
 * table for the 'context'.
 *
 * Return:	Function pointer to handler function,
 *		or NULL if it doens't exist.
 */
void _nx_lookup_handler(unsigned long context_arg,
			char *handler_name,
			int handler_id,
			nexus_handler_type_t *type,
			nexus_handler_func_t *func)
{
    nexus_handler_record_t *rec;
    nexus_context_t *context = (nexus_context_t *) context_arg;

    NexusAssert2(((handler_id >= 0) && (handler_id < handler_table_size)),
		 ("_nx_lookup_handler(): Handler id (%d) out of range (%d-%d).\n", handler_id, 0, handler_table_size) );

    /*
     * in case this rsr was received from a machine that did not
     * know of any context here but just a node number
     * (such as an mp machine like the SP2) use a valid context
     */

    nexus_debug_printf(2, ("_nx_lookup_handler(): Looking up:\"%s\":%d\n",
			   handler_name, handler_id));

    if (context_arg == 0)
    {
	nexus_debug_printf(2, ("_nx_lookup_handler(): Got NULL context. Using _nx_initial_context:%x\n",_nx_initial_context));
	context_arg = (unsigned long) _nx_initial_context;
    }

    context = (nexus_context_t *) context_arg;

    for (rec = context->handler_table[handler_id];
	 (   (rec != (nexus_handler_record_t *) NULL)
	  && (strcmp(handler_name,rec->name) != 0) );
	 rec = rec->next)
	;

    if (rec != (nexus_handler_record_t *) NULL)
    {
	*type = rec->type;
	*func = (nexus_handler_func_t) rec;
    }
    else
    {
	*type = NEXUS_HANDLER_TYPE_NON_THREADED;
	*func = (nexus_handler_func_t) NULL;
    }

    nexus_debug_printf(2, ("_nx_lookup_handler(): Found handler\n"));
    
} /* _nx_lookup_handler() */


/*
 * _nx_handle_message()
 *
 * Call the handler function, 'func' function with 'address' and 'buffer'
 * as arguments.  Set the thread to the given 'context' before
 * doing so.
 *
 * If type==NEXUS_HANDLER_TYPE_THREADED, then the function will be
 * called in a newly thread, and the assumed types are:
 *	nexus_stashed_buffer_t buffer
 *	nexus_threaded_handler_func_t func
 * If type==NEXUS_HANDLER_TYPE_NON_THREADED, then the function will be
 * called in this thread, and the assumed types are:
 *	nexus_buffer_t buffer
 *	nexus_non_threaded_handler_func_t func
 *
 * Note: The buffer that is passed to a threaded handler is actually of
 * type nexus_stashed_buffer_t*, which means it really wants the address
 * of the buffer argument (&buffer).  However, we need to make
 * sure that the &buffer is done from within the newly created
 * thread for that handler, so that it doesn't dereference through
 * another thread's stack.  This was a nasty bug to find...
 *
 * If 'handler_func' is NULL, then NexusUnknownHandler() is called
 * using the 'handler_name' and 'handler_id' arguments.
 * If 'handler_func' is not NULL, then the 'handler_name'
 * and 'handler_id' arguments are ignored.
 * This was done so that you could avoid copying the handler_name
 * unless _nx_lookup_handler() failed to return a function pointer.
 *
 * If the handler is non-threaded, then free the buffer
 * using nexus_free_buffer(), after the handler completes.
 * If the handler is threaded, then do not free the buffer.
 */
void _nx_handle_message(char *handler_name,
			int handler_id,
			nexus_handler_type_t type,
			nexus_handler_func_t func_arg,
			unsigned long context_arg,
			unsigned long address,
#ifdef NEXUS_PROFILE
			int node_id,
			int context_id,
			int message_length,
#endif
			void *buffer)
{
    nexus_context_t *context;
    nexus_handler_record_t *rec = (nexus_handler_record_t *) func_arg;

    /*
     * in case this rsr was received from a machine that did not
     * know of any context here but just a node number
     * (such as an mp machine like the SP2) use a valid context
     */

    if (context_arg == 0)
    {
	context_arg = (unsigned long) _nx_initial_context;
    }

    context = (nexus_context_t *) context_arg;

    if (rec == (nexus_handler_record_t *) NULL)
    {
	/* Call NexusUnknownHandler() */
	nexus_context_t *save_context;
	nexus_unknown_handler_func_t NexusUnknownHandler;
	nexus_buffer_t buf = (nexus_buffer_t) buffer;
	
	_nx_lookup_handler((unsigned long) context_arg,
			   "NexusUnknownHandler",
			   NEXUS_UNKNOWN_HANDLER_HASH,
			   &type,
			   (nexus_handler_func_t *) (&rec) );
	
	nexus_debug_printf(1, ("_nx_handle_message(): Handler:[%s] not found. %s\n",
			       handler_name,
			       (rec ? "Calling unknown handler" :
				"Ignoring (unknown handler not registered)")));
	
	_nx_context(&save_context);
	_nx_set_context(context);
	
#ifdef NEXUS_PROFILE	
	_nx_pablo_log_remote_service_request_receive(node_id,
						     context_id,
						     handler_name,
						     handler_id,
						     message_length);
#endif /* NEXUS_PROFILE */	    
	
	if (rec)
	{
	    NexusAssert2((type == NEXUS_HANDLER_TYPE_NON_THREADED),
			 ("_nx_handle_message(): NexusUnknownHandler() must be a non-threaded handler\n") );

	    NexusUnknownHandler = (nexus_unknown_handler_func_t) rec->func;
	    
	    (*NexusUnknownHandler)((void *) address,
				   &buf,
				   handler_name,
				   handler_id);
	}
	_nx_set_context(save_context);
	nexus_free_buffer(&buf);
    }
    else
    {
	/* Call the handler */
#ifndef NEXUS_LITE
	if (type == NEXUS_HANDLER_TYPE_THREADED)
	{
	    threaded_handler_startup_info_t info;
	    nexus_thread_t thread;

	    info.context = (nexus_context_t *) context;
	    info.address = (void *) address;
#ifdef NEXUS_PROFILE
	    info.node_id = node_id;
	    info.context_id = context_id;
	    info.message_length = message_length;
#endif	    
	    info.buffer = (nexus_stashed_buffer_t) buffer;
	    info.rec = rec;
	    info.started = NEXUS_FALSE;
	    nexus_mutex_init(&(info.mutex), (nexus_mutexattr_t *) NULL);
	    nexus_cond_init(&(info.cond), (nexus_condattr_t *) NULL);

	    nexus_debug_printf(2, ("_nx_handle_message(): creating thread to handle buffer:%x\n",buffer));

	    nexus_thread_create(&thread,
				(nexus_thread_attr_t *) NULL,
				threaded_handler_startup,
				(void *) (&info) );

	    nexus_debug_printf(2, ("_nx_handle_message(): thread created\n"));

	    nexus_mutex_lock(&(info.mutex));
	    while (!info.started)
	    {
		nexus_cond_wait(&(info.cond), &(info.mutex));
	    }
	    nexus_mutex_unlock(&(info.mutex));
	    nexus_mutex_destroy(&(info.mutex));
	    nexus_cond_destroy(&(info.cond));
	}
	else /* (type == NEXUS_HANDLER_TYPE_NON_THREADED) */
#endif /* NEXUS_LITE */
	{
	    nexus_non_threaded_handler_func_t handler_func
		= (nexus_non_threaded_handler_func_t) rec->func;
	    nexus_context_t *save_context;
	    nexus_buffer_t buf = (nexus_buffer_t) buffer;
	    
	    _nx_context(&save_context);
	    _nx_set_context(context);

#ifdef NEXUS_PROFILE	
	    _nx_pablo_log_remote_service_request_receive(node_id,
							 context_id,
							 rec->name,
							 handler_id,
							 message_length);
	    if (_nx_pablo_count_remote_service_requests())
	    {
		_nx_accumulate_rsr_profile_info(context,
						rec,
						node_id,
						context_id,
						message_length);
	    }
#endif /* NEXUS_PROFILE */
	    nexus_debug_printf(2, ("Beginning non-threaded handler call with address:%x and buffer:%x\n",address, &buf));
	    (*handler_func)((void *) address, &buf);
	    nexus_debug_printf(2, ("Ending non-threaded handler call\n"));
	    nexus_free_buffer(&buf);
	    nexus_debug_printf(2, ("Buffer freed\n"));
	    _nx_set_context(save_context);
	}
    }

} /* _nx_handle_message() */


#ifndef NEXUS_LITE
/*
 * threaded_handler_startup()
 *
 * Startup a threaded handler.  The info about the handler is
 * in arg.
 */
static void *threaded_handler_startup(void *arg)
{
    threaded_handler_startup_info_t *info
	= (threaded_handler_startup_info_t *) arg;
    nexus_context_t *context;
    void *address;
    nexus_stashed_buffer_t buffer;
    nexus_handler_record_t *rec;
    nexus_threaded_handler_func_t func;
#ifdef NEXUS_PROFILE    
    int node_id;
    int context_id;
    int message_length;
#endif    

    context = info->context;
    address = info->address;
#ifdef NEXUS_PROFILE
    node_id = info->node_id;
    context_id = info->context_id;
    message_length = info->message_length;
#endif
    buffer = info->buffer;
    rec = info->rec;
    func = (nexus_threaded_handler_func_t) rec->func;
    
    _nx_set_context(context);
    
    nexus_mutex_lock(&(info->mutex));
    info->started = NEXUS_TRUE;
    nexus_cond_signal(&(info->cond));
    nexus_mutex_unlock(&(info->mutex));

#ifdef NEXUS_PROFILE	
    _nx_pablo_log_remote_service_request_receive(node_id,
						 context_id,
						 rec->name,
						 rec->id,
						 message_length);
    if (_nx_pablo_count_remote_service_requests())
    {
	_nx_accumulate_rsr_profile_info(context,
					rec,
					node_id,
					context_id,
					message_length);
    }
#endif /* NEXUS_PROFILE */

#ifdef NEXUS_DEBUG
    if(NexusDebug(2)) {
	nexus_printf("threaded_handler_startup(): calling handler for buffer:%x\n",buffer);
    }
#endif

    (*func)(address, &buffer);

    return (NULL);
    
} /* threaded_handler_startup() */
#endif /* NEXUS_LITE */


#ifdef NEXUS_PROFILE
/*
 * _nx_accumulate_rsr_profile_info()
 *
 * Accumulate the count and message length of a message destined
 * for the handler described by 'rec', for the given source 'node_id'
 * and 'context_id'.
 */
void _nx_accumulate_rsr_profile_info(nexus_context_t *context,
				     nexus_handler_record_t *rec,
				     int node_id,
				     int context_id,
				     int message_length)
{
    nexus_profile_rsr_record_t *profile_rec;
    int i;
    int hash;

    _nx_context_lock(context);
    
    /* Allocate the profile table, if there isn't one already */
    if (rec->profile_table == (nexus_profile_rsr_record_t **) NULL)
    {
	NexusMalloc(_nx_accumulate_rsr_profile_info(),
		    rec->profile_table,
		    nexus_profile_rsr_record_t **,
		    (profile_rsr_table_size
		     * sizeof(nexus_profile_rsr_record_t *)) );
	for (i = 0; i < profile_rsr_table_size; i++)
	{
	    rec->profile_table[i] = (nexus_profile_rsr_record_t *) NULL;
	}
    }

    /* Find the appropriate profile record in the table */
    hash = (node_id + context_id) % profile_rsr_table_size;
    for (profile_rec = rec->profile_table[hash];
	 (   (profile_rec != (nexus_profile_rsr_record_t *) NULL)
	  && (   (profile_rec->node_id != node_id)
	      || (profile_rec->context_id != context_id)));
	 profile_rec = profile_rec->next)
	;

    if (profile_rec == (nexus_profile_rsr_record_t *) NULL)
    {
	/* Didn't find an entry, so allocate one and store the data */
	NexusMalloc(_nx_accumulate_rsr_profile_info(),
		    profile_rec,
		    nexus_profile_rsr_record_t *,
		    (sizeof(nexus_profile_rsr_record_t)) );
	profile_rec->node_id = node_id;
	profile_rec->context_id = context_id;
	profile_rec->count = 1;
	profile_rec->size = message_length;
	profile_rec->next = rec->profile_table[hash];
	rec->profile_table[hash] = profile_rec;
    }
    else
    {
	/* Found and entry, so bump the data */
	profile_rec->count++;
	profile_rec->size += message_length;
    }

    /* See if we need to dump the current counts */
    context->rsr_profile_count++;
    if (_nx_pablo_should_dump_remote_service_request_counts(context->rsr_profile_count))
    {
	_nx_dump_rsr_profile_info(context);
    }
    
    _nx_context_unlock(context);
    
} /* _nx_accumulate_rsr_profile_info() */


/*
 * _nx_dump_rsr_profile_info()
 *
 * Dump and reset the rsr profile records for the current context.
 */
void _nx_dump_rsr_profile_info(nexus_context_t *context)
{
    nexus_handler_record_t *handler_rec;
    nexus_profile_rsr_record_t *profile_rec;
    int i, j;
    int my_node_id, my_context_id;
    int snapshot_id;
    char *handler_name;
    int handler_id;

    _nx_node_id(&my_node_id);
    my_context_id = context->id;
    snapshot_id = context->snapshot_id;
    
    for (i = 0; i < handler_table_size; i++)
    {
	for (handler_rec = context->handler_table[i];
	     handler_rec;
	     handler_rec = handler_rec->next)
	{
	    if (handler_rec->profile_table)
	    {
		handler_name = handler_rec->name;
		handler_id = handler_rec->id;
		for (j = 0; j < profile_rsr_table_size; j++)
		{
		    for (profile_rec = handler_rec->profile_table[j];
			 profile_rec;
			 profile_rec = profile_rec->next)
		    {
			if (profile_rec->count > 0)
			{
			    _nx_pablo_log_remote_service_request_count(
						profile_rec->count,
						snapshot_id,
						profile_rec->node_id,
						profile_rec->context_id,
						my_node_id,
						my_context_id,
						handler_name,
						handler_id,
						profile_rec->size);
			    profile_rec->count = 0;
			    profile_rec->size = 0;
			}
		    }
		}
	    }
	}
    }

    context->snapshot_id++;
    context->rsr_profile_count = 0;
    
} /* _nx_dump_rsr_profile_info() */

#endif /* NEXUS_PROFILE */
