/*
 * Nexus
 * Authors:     Steven Tuecke and Robert Olson
 *              Argonne National Laboratory
 *
 * pr_local.c		- local protocol module
 */

static char *rcsid = "$Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/pr_local.c,v 1.24 1995/06/20 16:10:55 tuecke Exp $";

#include "internal.h"


/*
 * Only one thread is allowed to be in the local protocol code (and thus
 * mucking with data structures) at a time.
 */
#ifdef NEXUS_LITE
#define local_enter()
#define local_exit()
#else  /* NEXUS_LITE */
static nexus_mutex_t	local_mutex;
#define local_enter()	nexus_mutex_lock(&local_mutex);
#define local_exit()	nexus_mutex_unlock(&local_mutex);
#endif /* NEXUS_LITE */

/*
 * Some forward typedef declarations...
 */
typedef struct _local_buffer_t	local_buffer_t;
typedef struct _local_proto_t	local_proto_t;


#ifdef NEXUS_DEBUG
static char tmpbuf1[10240];
static char tmpbuf2[10240];
#endif

/*
 * local_proto_t
 *
 * This is an overload of nexus_proto_t.  It adds the
 * local specific information to that structure.
 */
struct _local_proto_t
{
    nexus_proto_type_t	type;	/* NEXUS_PROTO_TYPE_LOCAL */
    nexus_proto_funcs_t *funcs;
};


/*
 * local_buffer_t
 *
 * This is an overload of nexus_buffer_t.  It adds the
 * local specific information to that structure.
 */
struct _local_buffer_t
{
#ifdef NEXUS_DEBUG
    int				magic;
#endif /* NEXUS_DEBUG */
    nexus_buffer_funcs_t *	funcs;
    local_buffer_t *		next;
    unsigned long		context;
    unsigned long		address;
#ifdef NEXUS_PROFILE
    int				node_id;
    int				context_id;
    int				message_length;
#endif /* NEXUS_PROFILE */    
    unsigned long		size;
    int				n_elements;
    nexus_bool_t		stashed;
    char *			base_pointer;
    char *			current_pointer;
    int				handler_id;
    char			handler_name[NEXUS_MAX_HANDLER_NAME_LENGTH+1];
};

/*
 * A local_buffer_t free list, to avoid malloc calls on the
 * message buffer structure.
 */
static local_buffer_t *	buffer_free_list = (local_buffer_t *) NULL;

#ifdef NEXUS_LITE
#define lock_buffer_free_list()
#define unlock_buffer_free_list()
#else  /* NEXUS_LITE */
static nexus_mutex_t	buffer_free_list_mutex;
#define lock_buffer_free_list()   nexus_mutex_lock(&buffer_free_list_mutex)
#define unlock_buffer_free_list() nexus_mutex_unlock(&buffer_free_list_mutex)
#endif /* NEXUS_LITE */

#ifdef NEXUS_DEBUG
#define MallocLocalBuffer(Routine, Buf) \
{ \
    NexusMalloc(Routine, Buf, local_buffer_t *, \
		sizeof(struct _local_buffer_t)); \
    Buf->magic = NEXUS_BUFFER_MAGIC; \
}
#else  /* NEXUS_DEBUG */
#define MallocLocalBuffer(Routine, Buf) \
{ \
    NexusMalloc(Routine, Buf, local_buffer_t *, \
		sizeof(struct _local_buffer_t)); \
}
#endif /* NEXUS_DEBUG */

#define GetLocalBuffer(Routine, Buf) \
{ \
    lock_buffer_free_list(); \
    if (buffer_free_list) \
    { \
	Buf = buffer_free_list; \
	buffer_free_list = buffer_free_list->next; \
	unlock_buffer_free_list(); \
    } \
    else \
    { \
	unlock_buffer_free_list(); \
	MallocLocalBuffer(Routine, Buf); \
    } \
}
#define FreeLocalBuffer(Buf) \
{ \
    if ((Buf)->base_pointer != (char *) NULL) \
	NexusFree((Buf)->base_pointer); \
    lock_buffer_free_list(); \
    (Buf)->next = buffer_free_list; \
    buffer_free_list = (Buf); \
    unlock_buffer_free_list(); \
}


/*
 * This is a message queue containing all messages that have
 * been sent locally and are waiting to be handled.
 *
 * In the single-threaded version, the sending thread will
 * call handle_enqueued_messages() to handle messages on this queue.
 *
 * In the multi-threaded version, a separate handler thread will
 * call handle_enqueued_messages() to handle messages on this queue.
 * A lock and condition is used to awaken the handler thread when
 * there is a message to be delivered.
 */
static local_buffer_t *	message_queue_head;
static local_buffer_t *	message_queue_tail;

#define EnqueueMessage(buf) \
{ \
    if (message_queue_head == (local_buffer_t *) NULL) \
    { \
	message_queue_head = message_queue_tail = buf; \
    } \
    else \
    { \
	message_queue_tail->next = buf; \
	message_queue_tail = buf; \
    } \
    buf->next = (local_buffer_t *) NULL; \
}
#define DequeueMessage(buf) \
{ \
    /* Assume that message_queue_head != NULL */ \
    buf = message_queue_head; \
    message_queue_head = message_queue_head->next; \
}
#define MessagesEnqueued() (message_queue_head != (local_buffer_t *) NULL)


static nexus_bool_t	handle_in_progress;

static void handle_enqueued_messages();


/*
 * Various forward declarations of procedures
 */
static nexus_proto_type_t	local_proto_type();
static void		local_init(int *argc, char ***argv);
static void		local_shutdown(nexus_bool_t shutdown_others);
static void		local_abort();

static void		local_init_remote_service_request(
					    nexus_buffer_t *buffer,
					    nexus_global_pointer_t *gp,
					    char *handler_name,
					    int handler_id);
static void		local_send_remote_service_request(
						   nexus_buffer_t *buffer);
static void		local_send_urgent_remote_service_request(
						   nexus_buffer_t *buffer);
static void		local_destroy_proto(nexus_proto_t *nexus_proto);
static int		local_compare_protos(nexus_proto_t *proto1,
					     nexus_proto_t *proto2);


static void		local_set_buffer_size(nexus_buffer_t *buffer,
					      int size, int n_elements);
static int		local_check_buffer_size(nexus_buffer_t *buffer,
						int slack, int increment);
static void		local_free_buffer(nexus_buffer_t *buffer);
static void		local_stash_buffer(nexus_buffer_t *buffer,
				       nexus_stashed_buffer_t *stashed_buffer);
static void		local_free_stashed_buffer(
				       nexus_stashed_buffer_t *stashed_buffer);


static int	local_sizeof_float	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_double	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_short	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_u_short	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_int	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_u_int	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_long	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_u_long	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_char	(nexus_buffer_t *buffer, int count);
static int	local_sizeof_u_char	(nexus_buffer_t *buffer, int count);


static void	local_put_float	(nexus_buffer_t *buffer,
				 float *data, int count);
static void	local_put_double(nexus_buffer_t *buffer,
				 double *data, int count);
static void	local_put_short	(nexus_buffer_t *buffer,
				 short *data, int count);
static void	local_put_u_short(nexus_buffer_t *buffer,
				 unsigned short *data, int count);
static void	local_put_int	(nexus_buffer_t *buffer,
				 int *data, int count);
static void	local_put_u_int	(nexus_buffer_t *buffer,
				 unsigned int *data, int count);
static void	local_put_long	(nexus_buffer_t *buffer,
				 long *data, int count);
static void	local_put_u_long(nexus_buffer_t *buffer,
				 unsigned long *data, int count);
static void	local_put_char	(nexus_buffer_t *buffer,
				 char *data, int count);
static void	local_put_u_char(nexus_buffer_t *buffer,
				 unsigned char *data, int count);

static void	local_get_float	(nexus_buffer_t *buffer,
				 float *dest, int count);
static void	local_get_double(nexus_buffer_t *buffer,
				 double *dest, int count);
static void	local_get_short	(nexus_buffer_t *buffer,
				 short *dest, int count);
static void	local_get_u_short(nexus_buffer_t *buffer,
				 unsigned short *dest, int count);
static void	local_get_int	(nexus_buffer_t *buffer,
				 int *dest, int count);
static void	local_get_u_int	(nexus_buffer_t *buffer,
				 unsigned int *dest, int count);
static void	local_get_long	(nexus_buffer_t *buffer,
				 long *dest, int count);
static void	local_get_u_long(nexus_buffer_t *buffer,
				 unsigned long *dest, int count);
static void	local_get_char	(nexus_buffer_t *buffer,
				 char *dest, int count);
static void	local_get_u_char(nexus_buffer_t *buffer,
				 unsigned char *dest, int count);

static void	local_get_stashed_float	(nexus_stashed_buffer_t *buffer,
					 float *dest, int count);
static void	local_get_stashed_double(nexus_stashed_buffer_t *buffer,
					 double *dest, int count);
static void	local_get_stashed_short	(nexus_stashed_buffer_t *buffer,
					 short *dest, int count);
static void	local_get_stashed_u_short(nexus_stashed_buffer_t *buffer,
					 unsigned short *dest, int count);
static void	local_get_stashed_int	(nexus_stashed_buffer_t *buffer,
					 int *dest, int count);
static void	local_get_stashed_u_int	(nexus_stashed_buffer_t *buffer,
					 unsigned int *dest, int count);
static void	local_get_stashed_long	(nexus_stashed_buffer_t *buffer,
					 long *dest, int count);
static void	local_get_stashed_u_long(nexus_stashed_buffer_t *buffer,
					 unsigned long *dest, int count);
static void	local_get_stashed_char	(nexus_stashed_buffer_t *buffer,
					 char *dest, int count);
static void	local_get_stashed_u_char(nexus_stashed_buffer_t *buffer,
					 unsigned char *dest, int count);


static nexus_proto_funcs_t local_proto_funcs =
{
    local_proto_type,
    local_init,
    local_shutdown,
    local_abort,
    NULL /* local_poll */,
    NULL /* local_blocking_poll */,
    local_init_remote_service_request,
    local_destroy_proto,
    NULL /* local_get_my_mi_proto */,
    NULL /* local_construct_from_mi_proto */,
    NULL /* local_get_creator_proto_params */,
    NULL /* local_construct_creator_proto */,
    local_compare_protos,
};

static nexus_buffer_funcs_t local_buffer_funcs =
{
    local_set_buffer_size,
    local_check_buffer_size,
    local_send_remote_service_request,
    local_send_urgent_remote_service_request,
    local_free_buffer,
    local_stash_buffer,
    local_free_stashed_buffer,
    local_sizeof_float,
    local_sizeof_double,
    local_sizeof_short,
    local_sizeof_u_short,
    local_sizeof_int,
    local_sizeof_u_int,
    local_sizeof_long,
    local_sizeof_u_long,
    local_sizeof_char,
    local_sizeof_u_char,
    local_put_float,
    local_put_double,
    local_put_short,
    local_put_u_short,
    local_put_int,
    local_put_u_int,
    local_put_long,
    local_put_u_long,
    local_put_char,
    local_put_u_char,
    local_get_float,
    local_get_double,
    local_get_short,
    local_get_u_short,
    local_get_int,
    local_get_u_int,
    local_get_long,
    local_get_u_long,
    local_get_char,
    local_get_u_char,
    local_get_stashed_float,
    local_get_stashed_double,
    local_get_stashed_short,
    local_get_stashed_u_short,
    local_get_stashed_int,
    local_get_stashed_u_int,
    local_get_stashed_long,
    local_get_stashed_u_long,
    local_get_stashed_char,
    local_get_stashed_u_char,
};


/*
 * _nx_pr_local_info()
 *
 * Return the nexus_proto_funcs_t function table for this protocol module.
 *
 * This procedure is used for bootstrapping the protocol module.
 * The higher level Nexus code needs to call this routine to
 * retrieve the functions it needs to use this protocol module.
 */
void *_nx_pr_local_info()
{
    return((void *) (&local_proto_funcs));
} /* _nx_pr_local_info() */


/*
 * local_proto_type()
 *
 * Return the nexus_proto_type_t for this protocol module.
 */
static nexus_proto_type_t local_proto_type()
{
    return (NEXUS_PROTO_TYPE_LOCAL);
} /* local_proto_type() */


/*
 * local_init()
 *
 * Initialize the LOCAL protocol.
 */
static void local_init(int *argc, char ***argv)
{
    local_proto_t *local_proto;

#ifndef NEXUS_LITE
    nexus_mutex_init(&local_mutex, (nexus_mutexattr_t *) NULL);
    nexus_mutex_init(&buffer_free_list_mutex, (nexus_mutexattr_t *) NULL);
#endif /* NEXUS_LITE */
    message_queue_head = message_queue_tail = (local_buffer_t *) NULL;
    handle_in_progress = NEXUS_FALSE;
    
    /* Setup _nx_local_proto */
    NexusMalloc(local_init(), local_proto, local_proto_t *,
		sizeof(local_proto_t) );
    local_proto->type = NEXUS_PROTO_TYPE_LOCAL;
    local_proto->funcs = &local_proto_funcs;
    _nx_local_proto = (nexus_proto_t *) local_proto;
    
} /* local_init() */


/*
 * local_shutdown()
 *
 * This routine is called during normal shutdown of a process.
 */
static void local_shutdown(nexus_bool_t shutdown_others)
{
} /* local_shutdown() */


/*
 * local_abort()
 *
 * This routine is called during the _abnormal_ shutdown of a process.
 */
static void local_abort()
{
} /* local_abort() */

static void local_dump_buffer(local_buffer_t *buf)
{
    char *ptr;
    char *outputbuf;
    char temp2buf[30];
    int curr_printed=0;

    NexusMalloc(local_dump_buffer(), outputbuf, char *,
		(3 * buf->size + 20));


    outputbuf[0]='\0';
    ptr = buf->base_pointer;

    do {
	if (ptr+4 >= buf->current_pointer && curr_printed==0) {
	    sprintf(temp2buf, "%8.8x <<%d<< ", *((unsigned long *)ptr), 
		    buf->current_pointer - ptr);
	    strcat(outputbuf, temp2buf);
	    curr_printed=1;
	} else {
	    sprintf(temp2buf, "%8.8x ", *((unsigned long *)ptr));
	    strcat(outputbuf, temp2buf);
	}
	ptr += 4;
    } while(ptr < ((buf->base_pointer) + (buf->size)));

    nexus_printf("local_dump_buffer(): base_pointer:%x current_pointer:%x size:%d data:%s\n", 
		 buf->base_pointer, buf->current_pointer, 
		 buf->size, outputbuf);
}

/*
 * local_init_remote_service_request()
 *
 * Initiate a local remote service request (i.e. to the same context
 * that is calling this).
 *
 * Note: The things pointed to by 'gp' and 'handler_name' should
 * NOT change between this init_rsr call and the subsequent send_rsr.
 * Only the pointers are stashed away, and not the full data.
 *
 * Return: Fill in 'buffer' with a nexus_buffer_t.
 */
static void local_init_remote_service_request(nexus_buffer_t *buffer,
					      nexus_global_pointer_t *gp,
					      char *handler_name,
					      int handler_id)
{
    local_buffer_t *local_buffer;
    int handler_name_length;

    NexusAssert2((gp->proto->type == NEXUS_PROTO_TYPE_LOCAL),
		 ("local_init_remote_service_request(): Internal error: proto_type is not NEXUS_PROTO_TYPE_LOCAL\n"));

    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "local_init_remote_service_request");

    GetLocalBuffer(local_init_remote_service_request(), local_buffer);

    nexus_debug_printf(2, ("local_init_remote_service_request(): to:(%lu,%lu) %s-%d\n",gp->context,gp->address,handler_name,handler_id));

    local_buffer->funcs = &local_buffer_funcs;
    local_buffer->next = (local_buffer_t *) NULL;
    local_buffer->context = gp->context;
    local_buffer->address = gp->address;
#ifdef NEXUS_PROFILE
    /* Set node_id and context_id to the _destination_ node and context */
    local_buffer->node_id = gp->node_id;
    local_buffer->context_id = gp->context_id;
#endif
    local_buffer->size = 0;
    local_buffer->n_elements = -1;
    local_buffer->stashed = NEXUS_FALSE;
    local_buffer->base_pointer = (char *) NULL;
    local_buffer->current_pointer = (char *) NULL;
    local_buffer->handler_id = handler_id;
    
    handler_name_length = strlen(handler_name);
    NexusAssert2((handler_name_length < NEXUS_MAX_HANDLER_NAME_LENGTH),
		 ("local_init_remote_service_request(): Handler name exceeds maximum length (%d): %s\n", NEXUS_MAX_HANDLER_NAME_LENGTH, handler_name));
    strcpy(local_buffer->handler_name,handler_name);

    *buffer = (nexus_buffer_t) local_buffer;
} /* local_init_remote_service_request() */


/*
 * local_send_remote_service_request()
 *
 * Generate a local remote service request for 'nexus_buffer'.
 */
static void local_send_remote_service_request(nexus_buffer_t *nexus_buffer)
{
    local_buffer_t *buf;

    NexusBufferMagicCheck(local_send_remote_service_request, nexus_buffer);

    buf = (local_buffer_t *) *nexus_buffer;

#ifdef NEXUS_PROFILE
    buf->message_length = buf->current_pointer - buf->base_pointer;
    _nx_pablo_log_remote_service_request_send(buf->node_id,
					      buf->context_id,
					      buf->handler_name,
					      buf->handler_id,
					      buf->message_length);
    /* Reset node_id and context_id to the _source_ node and context */
    _nx_node_id(&(buf->node_id));
    _nx_context_id(&(buf->context_id));
#endif /* NEXUS_PROFILE */
    
    /* Reset the current_pointer for the future get operations */
    buf->current_pointer = buf->base_pointer;
#ifdef NEXUS_DEBUG
    if (NexusDebug(2)) {
/*
	int i;
	
	tmpbuf1[0] = '\0';
	for (i = 0; i < buf->size; i++) {
	    sprintf(tmpbuf2, "%u ", (unsigned int)((buf->base_pointer)[i]));
	    strcat(tmpbuf1, tmpbuf2);
	}

	nexus_printf("local_send_remote_service_request(): Beginning of message: %s\n",tmpbuf1);
	*/
	nexus_printf("local_send_remote_service_request(): sending buffer:%x\n",buf->base_pointer);
	if (buf)
	{
	    local_dump_buffer( buf );
	}
    }
#endif

    local_enter();
    EnqueueMessage(buf);
    if (!handle_in_progress)
    {
	/*
	 * Handle the message, if this isn't already being called
	 * from a handler
	 */
	handle_enqueued_messages();
    }
    local_exit();

    nexus_poll();

} /* local_send_remote_service_request() */


/*
 * local_send_urgent_remote_service_request()
 */
static void local_send_urgent_remote_service_request(nexus_buffer_t *buffer)
{
    local_send_remote_service_request(buffer);
} /* local_urgent_remote_service_request() */


/*
 * handle_enqueued_messages()
 *
 * Handle all messages that are enqueued.
 */
static void handle_enqueued_messages()
{
    local_buffer_t *buf;
    nexus_handler_func_t handler_func;
    nexus_handler_type_t handler_type;
    
    handle_in_progress = NEXUS_TRUE;
    while (MessagesEnqueued())
    {
	DequeueMessage(buf);

	nexus_debug_printf(3, ("local_handle_enqueued_messages(): calling %s with data base_pointer:%x and current_pointer:%x\n", buf->handler_name, buf->base_pointer, buf->current_pointer));

	_nx_lookup_handler(buf->context,
			   buf->handler_name,
			   buf->handler_id,
			   &handler_type,
			   &handler_func);
	if (handler_type == NEXUS_HANDLER_TYPE_THREADED)
	{
	    buf->stashed = NEXUS_TRUE;
	}

	local_exit();
	_nx_handle_message(buf->handler_name,
			   buf->handler_id,
			   handler_type,
			   handler_func,
			   buf->context,
			   buf->address,
#ifdef NEXUS_PROFILE
			   buf->node_id,
			   buf->context_id,
			   buf->message_length,
#endif
			   (void *) buf);
	local_enter();
    }
    handle_in_progress = NEXUS_FALSE;
    
} /* handle_enqueued_messages() */


/*
 * local_destroy_proto()
 *
 * Do nothing, since we never want to actually destroy the local proto.
 */
static void local_destroy_proto(nexus_proto_t *nexus_proto)
{
} /* local_destroy_proto() */


/*
 * local_compare_protos()
 */
static int local_compare_protos(nexus_proto_t *proto1,
				nexus_proto_t *proto2)
{
    return (NEXUS_TRUE);
} /* local_compare_protos() */


/*********************************************************************
 * 		Buffer management code
 *********************************************************************/

#ifdef NEXUS_SANITY_CHECK
#define SANITY_TYPE_CHECK_SIZE (sizeof(unsigned long) + sizeof(int))
#else  /* NEXUS_SANITY_CHECK */
#define SANITY_TYPE_CHECK_SIZE 0
#endif /* NEXUS_SANITY_CHECK */

/*
 * local_set_buffer_size()
 * 
 * Allocate message buffer space for 'size' bytes, that will be used to
 * hold 'n_elements'.
 */
static void local_set_buffer_size(nexus_buffer_t *buffer,
				  int size, int n_elements)
{
    local_buffer_t *local_buffer;
    char *storage;

    NexusBufferMagicCheck(local_check_buffer_size, buffer);
    
#ifdef NEXUS_SANITY_CHECK
    if (n_elements != -1) {
	size += SANITY_TYPE_CHECK_SIZE * n_elements;
    }
#endif

    if (size > 0)
    {
	local_buffer = (local_buffer_t *) *buffer;
    
	NexusMalloc(local_set_buffer_size(), storage, char *, size);

	local_buffer->size = size;
	local_buffer->n_elements = n_elements;
	local_buffer->base_pointer = storage;
	local_buffer->current_pointer = local_buffer->base_pointer;
    }

} /* local_set_buffer_size() */


/*
 * local_check_buffer_size()
 *
 * Check that that passed message 'buffer' has at least 'slack'
 * bytes remaining; if not, increase size by 'increment' bytes
 * until there are enough bytes remaining.
 *
 * If no resizing is necessary, leave 'buffer' unchanged and
 * return NEXUS_TRUE.
 * If resizing is successful, modify 'buffer' to a new, larger
 * buffer and return NEXUS_TRUE.
 * Otherwise, if 'increment' is 0 and 'slack' bytes are not
 * available in teh buffer, then leave 'buffer' unchanged and
 * return NEXUS_FALSE.
 */
static int local_check_buffer_size(nexus_buffer_t *buffer,
				   int slack, int increment)
{
    local_buffer_t *local_buffer;
    int used;
    int needed;

    NexusBufferMagicCheck(local_check_buffer_size, buffer);

    local_buffer = (local_buffer_t *) *buffer;
    
    used = (local_buffer->current_pointer - local_buffer->base_pointer);
    needed = used + slack;

    if (local_buffer->size == 0)
    {
	local_set_buffer_size(buffer, slack, -1);
    }
    else if (needed > local_buffer->size)
    {
	char *new_storage;
	int new_size;

	if (increment <= 0)
	    return(NEXUS_FALSE);

	new_size = local_buffer->size;
	while (new_size < needed)
	{
	    new_size += increment;
	}

	NexusMalloc(local_check_buffer_size(), new_storage, char *, new_size);
	
	memcpy(new_storage, local_buffer->base_pointer, used);
	    
	NexusFree(local_buffer->base_pointer);

	local_buffer->size = new_size;
	local_buffer->base_pointer = new_storage;
	local_buffer->current_pointer = (local_buffer->base_pointer + used);
    }
    
    return(NEXUS_TRUE);
} /* local_check_buffer_size() */


/*
 * local_free_buffer()
 *
 * Free the passed nexus_buffer_t.
 *
 * This should be called on the receiving end, after the handler
 * has completed.
 *
 * Note: The stashed flag could be set, since local_stash_buffer()
 * just sets this flag and typecasts a buffer to a stashed buffer.
 * In this case, do not free the buffer.
 */
static void local_free_buffer(nexus_buffer_t *buffer)
{
    local_buffer_t *local_buffer;

    NexusAssert2((buffer),
		 ("local_free_buffer(): Passed a NULL nexus_buffer_t *\n") );

    /* If the buffer was stashed, *buffer will have been set to NULL */
    if (!(*buffer))
	return;

    NexusBufferMagicCheck(local_free_buffer, buffer);

    local_buffer = (local_buffer_t *) *buffer;

    NexusAssert2((!local_buffer->stashed),
		 ("local_free_buffer(): Expected a non-stashed buffer\n"));

    FreeLocalBuffer(local_buffer);

    *buffer = (nexus_buffer_t) NULL;
    
} /* local_free_buffer() */


/*
 * local_stash_buffer()
 *
 * Convert 'buffer' to a stashed buffer.
 */
static void local_stash_buffer(nexus_buffer_t *buffer,
			       nexus_stashed_buffer_t *stashed_buffer)
{
    local_buffer_t *local_buffer;
    
    NexusBufferMagicCheck(local_stash_buffer, buffer);

    local_buffer = (local_buffer_t *) *buffer;
    NexusAssert2((!local_buffer->stashed),
		 ("local_stash_buffer(): Expected an un-stashed buffer\n"));
    
    local_buffer->stashed = NEXUS_TRUE;
		  
    *stashed_buffer = (nexus_stashed_buffer_t) *buffer;

    *buffer = (nexus_buffer_t) NULL;
} /* local_stash_buffer() */


/*
 * local_free_stashed_buffer()
 *
 * Free the passed nexus_stashed_buffer_t that was stashed
 * by local_stash_buffer().
 */
static void local_free_stashed_buffer(nexus_stashed_buffer_t *stashed_buffer)
{
    local_buffer_t *local_buffer;
    
    NexusAssert2((stashed_buffer),
		 ("local_free_stashed_buffer(): Passed a NULL nexus_stashed_buffer_t *\n") );
    NexusBufferMagicCheck(local_free_stashed_buffer,
			  (nexus_buffer_t *) stashed_buffer);
    
    local_buffer = (local_buffer_t *) *stashed_buffer;
    
    NexusAssert2((local_buffer->stashed),
		 ("local_free_stashed_buffer(): Expected a stashed buffer\n"));

    FreeLocalBuffer(local_buffer);

    *stashed_buffer = (nexus_stashed_buffer_t) NULL;
    
} /* local_free_stashed_buffer() */




/*
 * local_sizeof_*()
 *
 * Return the size (in bytes) that 'count' elements of the given
 * type will require to be put into the 'buffer'.
 */
static int local_sizeof_float(nexus_buffer_t *buffer, int count)
{
    return (sizeof(float) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_double(nexus_buffer_t *buffer, int count)
{
    return (sizeof(double) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_short(nexus_buffer_t *buffer, int count)
{
    return (sizeof(short) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_u_short(nexus_buffer_t *buffer, int count)
{
    return (sizeof(u_short) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_int(nexus_buffer_t *buffer, int count)
{
    return (sizeof(int) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_u_int(nexus_buffer_t *buffer, int count)
{
    return (sizeof(u_int) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_long(nexus_buffer_t *buffer, int count)
{
    return (sizeof(long) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_u_long(nexus_buffer_t *buffer, int count)
{
    return (sizeof(u_long) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_char(nexus_buffer_t *buffer, int count)
{
    return (sizeof(char) * count + SANITY_TYPE_CHECK_SIZE);
}

static int local_sizeof_u_char(nexus_buffer_t *buffer, int count)
{
    return (sizeof(u_char) * count + SANITY_TYPE_CHECK_SIZE);
}


#ifdef NEXUS_ARCH_AIX
#define DO_TRACEBACK _nx_traceback()
#else
#define DO_TRACEBACK
#endif


/*
 * local_put_*()
 *
 * Put 'count' elements, starting at 'data', of the given type,
 * into 'buffer'.
 */

#ifdef NEXUS_DEBUG
#define DO_PUT_ASSERTIONS(TYPE) \
    NexusBufferMagicCheck(local_put_ ## TYPE, buf); \
    if (((local_buf->current_pointer - local_buf->base_pointer) + count * sizeof(TYPE) + SANITY_TYPE_CHECK_SIZE) > local_buf->size) \
    { \
        DO_TRACEBACK; \
	nexus_fatal("nexus_put_" #TYPE ": Buffer overrun\n"); \
    } \
    if(NexusDebug(3)) \
    { \
        nexus_printf("local_put_" #TYPE "(): base:%x size:%lu current:%x limit:%x count:%d\n", \
		     local_buf->base_pointer, \
		     local_buf->size, \
		     local_buf->current_pointer, \
		     local_buf->base_pointer + local_buf->size, \
		     count); \
    }

#else /* NEXUS_DEBUG */
#define DO_PUT_ASSERTIONS(TYPE)
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_SANITY_CHECK
#define PUT(TYPE) \
    local_buffer_t *local_buf = (local_buffer_t *) *buf; \
    DO_PUT_ASSERTIONS(TYPE); \
    { \
    unsigned long sanity_flag_long   = 0x10ca100a;  \
    unsigned long sanity_flag_int    = 0x10ca100b;  \
    unsigned long sanity_flag_char   = 0x10ca100c;  \
    unsigned long sanity_flag_double = 0x10ca100d;  \
    unsigned long sanity_flag_short  = 0x10ca100e;  \
    unsigned long sanity_flag_float  = 0x10ca100f;  \
                                                    \
    unsigned long sanity_flag_u_long  = 0x10ca101a; \
    unsigned long sanity_flag_u_int   = 0x10ca101b; \
    unsigned long sanity_flag_u_char  = 0x10ca101c; \
                                                    \
    unsigned long sanity_flag_u_short = 0x10ca101e; \
                                                    \
    memcpy(local_buf->current_pointer, & sanity_flag_ ## TYPE, sizeof(unsigned long)); \
    local_buf->current_pointer += sizeof(unsigned long);                               \
    memcpy(local_buf->current_pointer, & count, sizeof(int));                          \
    local_buf->current_pointer += sizeof(int);                                         \
    memcpy(local_buf->current_pointer, data, count * sizeof(TYPE));                    \
    local_buf->current_pointer += (count * sizeof(TYPE));                              \
    }
#else  /* NEXUS_SANITY_CHECK */
#define PUT(TYPE) \
    local_buffer_t *local_buf = (local_buffer_t *) *buf; \
    DO_PUT_ASSERTIONS(TYPE); \
    memcpy(local_buf->current_pointer, data, count * sizeof(TYPE)); \
    local_buf->current_pointer += count * sizeof(TYPE);
#endif /* NEXUS_SANITY_CHECK */


static void local_put_float(nexus_buffer_t *buf, float *data, int count)
{
    PUT(float);
}

static void local_put_double(nexus_buffer_t *buf, double *data, int count)
{
    PUT(double);
}

static void local_put_short(nexus_buffer_t *buf, short *data, int count)
{
    PUT(short);
}

static void local_put_u_short(nexus_buffer_t *buf, unsigned short *data,
			    int count)
{
    PUT(u_short);
}

static void local_put_int(nexus_buffer_t *buf, int *data, int count)
{
    PUT(int);
}

static void local_put_u_int(nexus_buffer_t *buf, unsigned int *data, int count)
{
    PUT(u_int);
}

static void local_put_long(nexus_buffer_t *buf, long *data, int count)
{
    PUT(long);
}

static void local_put_u_long(nexus_buffer_t *buf, unsigned long *data,
			     int count)
{
    PUT(u_long);
}

static void local_put_char(nexus_buffer_t *buf, char *data, int count)
{
    PUT(char);
}

static void local_put_u_char(nexus_buffer_t *buf, unsigned char *data,
			     int count)
{
    PUT(u_char);
}



/*
 * local_get_*()
 *
 * Get 'count' elements of the given type from 'buffer' and store
 * them into 'data'.
 */

#ifdef NEXUS_DEBUG
#define DO_GET_ASSERTIONS(TYPE) \
    NexusBufferMagicCheck(local_get_ ## TYPE, buf); \
    NexusAssert2(!local_buf->stashed, \
		 ("local_get_stashed_*(): Expected an un-stashed buffer\n")); \
    if (local_buf->current_pointer > (local_buf->base_pointer + local_buf->size)) \
    { \
        DO_TRACEBACK; \
	nexus_fatal("nexus_get_" #TYPE ": Buffer overrun\n"); \
    } \
    if(NexusDebug(3)) \
    { \
        nexus_printf("local_get_" #TYPE "(): base:%x size:%lu current:%x limit:%x count:%d\n", \
		     local_buf->base_pointer, \
		     local_buf->size, \
		     local_buf->current_pointer, \
		     local_buf->base_pointer + local_buf->size, \
		     count); \
    }

#else /* NEXUS_DEBUG */
#define DO_GET_ASSERTIONS(TYPE)
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_SANITY_CHECK
#define GET(TYPE) \
    local_buffer_t *local_buf = (local_buffer_t *) *buf; \
    DO_GET_ASSERTIONS(TYPE); \
    { \
        unsigned long sanity_flag_long   = 0x10ca100a;  \
        unsigned long sanity_flag_int    = 0x10ca100b;  \
        unsigned long sanity_flag_char   = 0x10ca100c;  \
        unsigned long sanity_flag_double = 0x10ca100d;  \
        unsigned long sanity_flag_short  = 0x10ca100e;  \
        unsigned long sanity_flag_float  = 0x10ca100f;  \
                                                        \
        unsigned long sanity_flag_u_long  = 0x10ca101a; \
        unsigned long sanity_flag_u_int   = 0x10ca101b; \
        unsigned long sanity_flag_u_char  = 0x10ca101c; \
                                                        \
        unsigned long sanity_flag_u_short = 0x10ca101e; \
                                                        \
        unsigned long buffer_sanity_value; \
        int buffer_sanity_count; \
        \
        memcpy( &buffer_sanity_value, local_buf->current_pointer, sizeof(unsigned long)); \
        if(  buffer_sanity_value != sanity_flag_ ## TYPE ) { \
    nexus_printf("local_get_" #TYPE ": warning: expected type of %x but found %x\n", \
		         sanity_flag_ ## TYPE , \
		         buffer_sanity_value); \
	    local_dump_buffer( local_buf ); \
	    DO_TRACEBACK; \
        } \
        local_buf->current_pointer += sizeof(unsigned long); \
        memcpy( &buffer_sanity_count, local_buf->current_pointer, sizeof(int)); \
        if( buffer_sanity_count != count ) { \
         nexus_printf("local_get_" #TYPE ": warning: expected count of %x but found %x\n", \
		         count, \
		         buffer_sanity_count); \
	    local_dump_buffer( local_buf ); \
	    DO_TRACEBACK; \
        } \
        local_buf->current_pointer += sizeof(int); \
    } \
    memcpy(dest, local_buf->current_pointer, sizeof(TYPE) * count); \
    local_buf->current_pointer += count * sizeof(TYPE);
#else  /* NEXUS_SANITY_CHECK */
#define GET(TYPE) \
    local_buffer_t *local_buf = (local_buffer_t *) *buf; \
    DO_GET_ASSERTIONS(TYPE); \
    memcpy(dest, local_buf->current_pointer, sizeof(TYPE) * count); \
    local_buf->current_pointer += count * sizeof(TYPE);
#endif /* NEXUS_SANITY_CHECK */



static void local_get_float(nexus_buffer_t *buf, float *dest, int count)
{
    GET(float);
}

static void local_get_double(nexus_buffer_t *buf, double *dest, int count)
{
    GET(double);
}

static void local_get_short(nexus_buffer_t *buf, short *dest, int count)
{
    GET(short);
}

static void local_get_u_short(nexus_buffer_t *buf, unsigned short *dest,
			    int count)
{
    GET(u_short);
}

static void local_get_int(nexus_buffer_t *buf, int *dest, int count)
{
    GET(int);
}

static void local_get_u_int(nexus_buffer_t *buf, unsigned int *dest, int count)
{
    GET(u_int);
}

static void local_get_long(nexus_buffer_t *buf, long *dest, int count)
{
    GET(long);
}

static void local_get_u_long(nexus_buffer_t *buf, unsigned long *dest,
			     int count)
{
    GET(u_long);
}

static void local_get_char(nexus_buffer_t *buf, char *dest, int count)
{
    GET(char);
}

static void local_get_u_char(nexus_buffer_t *buf, unsigned char *dest,
			     int count)
{
    GET(u_char);
}



/*
 * local_get_stashed_*()
 *
 * Get 'count' elements of the given type from the stashed 'buffer' and store
 * them into 'data'.
 */

#ifdef NEXUS_DEBUG
#define DO_GET_STASHED_ASSERTIONS(TYPE) \
    NexusBufferMagicCheck(local_get_stashed_ ## TYPE, (nexus_buffer_t *) buf); \
    NexusAssert2(local_buf->stashed, \
		 ("local_get_stashed_*(): Expected a stashed buffer\n")); \
    if (local_buf->current_pointer > (local_buf->base_pointer + local_buf->size)) \
    { \
        DO_TRACEBACK; \
	nexus_fatal("nexus_get_stashed_" #TYPE ": Buffer overrun\n"); \
    } \
    if(NexusDebug(3)) \
    { \
        nexus_printf("local_get_stashed_" #TYPE "(): base:%x size:%lu current:%x limit:%x count:%d\n", \
		     local_buf->base_pointer, \
		     local_buf->size, \
		     local_buf->current_pointer, \
		     local_buf->base_pointer + local_buf->size, \
		     count); \
    }

#else /* NEXUS_DEBUG */
#define DO_GET_STASHED_ASSERTIONS(TYPE)
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_SANITY_CHECK
#define GET_STASHED(TYPE) \
    local_buffer_t *local_buf = (local_buffer_t *) *buf; \
    DO_GET_STASHED_ASSERTIONS(TYPE); \
    { \
        unsigned long sanity_flag_long   = 0x10ca100a;  \
        unsigned long sanity_flag_int    = 0x10ca100b;  \
        unsigned long sanity_flag_char   = 0x10ca100c;  \
        unsigned long sanity_flag_double = 0x10ca100d;  \
        unsigned long sanity_flag_short  = 0x10ca100e;  \
        unsigned long sanity_flag_float  = 0x10ca100f;  \
                                                        \
        unsigned long sanity_flag_u_long  = 0x10ca101a; \
        unsigned long sanity_flag_u_int   = 0x10ca101b; \
        unsigned long sanity_flag_u_char  = 0x10ca101c; \
                                                        \
        unsigned long sanity_flag_u_short = 0x10ca101e; \
                                                        \
        unsigned long buffer_sanity_value; \
        int buffer_sanity_count; \
        \
        memcpy( &buffer_sanity_value, local_buf->current_pointer, sizeof(unsigned long)); \
        if(  buffer_sanity_value != sanity_flag_ ## TYPE ) { \
           nexus_printf("local_get_stashed_" #TYPE ": warning: expected type of %x but found %x\n", \
		         sanity_flag_ ## TYPE , \
		         buffer_sanity_value); \
	    local_dump_buffer( local_buf ); \
	    DO_TRACEBACK; \
        } \
        local_buf->current_pointer += sizeof(unsigned long); \
        memcpy( &buffer_sanity_count, local_buf->current_pointer, sizeof(int)); \
        if( buffer_sanity_count != count ) { \
         nexus_printf("local_get_stashed_" #TYPE ": warning: expected count of %x but found %x\n", \
		         count, \
		         buffer_sanity_count); \
	    local_dump_buffer( local_buf ); \
	    DO_TRACEBACK; \
        } \
        local_buf->current_pointer += sizeof(int); \
    } \
    memcpy(dest, local_buf->current_pointer, sizeof(TYPE) * count); \
    local_buf->current_pointer += count * sizeof(TYPE);
#else  /* NEXUS_SANITY_CHECK */
#define GET_STASHED(TYPE) \
    local_buffer_t *local_buf = (local_buffer_t *) *buf; \
    DO_GET_STASHED_ASSERTIONS(TYPE); \
    memcpy(dest, local_buf->current_pointer, sizeof(TYPE) * count); \
    local_buf->current_pointer += count * sizeof(TYPE);
#endif /* NEXUS_SANITY_CHECK */



static void local_get_stashed_float(nexus_stashed_buffer_t *buf,
				    float *dest, int count)
{
    GET_STASHED(float);
}

static void local_get_stashed_double(nexus_stashed_buffer_t *buf,
				     double *dest, int count)
{
    GET_STASHED(double);
}

static void local_get_stashed_short(nexus_stashed_buffer_t *buf,
				    short *dest, int count)
{
    GET_STASHED(short);
}

static void local_get_stashed_u_short(nexus_stashed_buffer_t *buf,
				      unsigned short *dest, int count)
{
    GET_STASHED(u_short);
}

static void local_get_stashed_int(nexus_stashed_buffer_t *buf,
				  int *dest, int count)
{
    GET_STASHED(int);
}

static void local_get_stashed_u_int(nexus_stashed_buffer_t *buf,
				    unsigned int *dest, int count)
{
    GET_STASHED(u_int);
}

static void local_get_stashed_long(nexus_stashed_buffer_t *buf,
				   long *dest, int count)
{
    GET_STASHED(long);
}

static void local_get_stashed_u_long(nexus_stashed_buffer_t *buf,
				     unsigned long *dest, int count)
{
    GET_STASHED(u_long);
}

static void local_get_stashed_char(nexus_stashed_buffer_t *buf,
				   char *dest, int count)
{
    GET_STASHED(char);
}

static void local_get_stashed_u_char(nexus_stashed_buffer_t *buf,
				     unsigned char *dest, int count)
{
    GET_STASHED(u_char);
}
