/*
 * Nexus
 * Authors:     Steven Tuecke and Robert Olson
 *              Argonne National Laboratory
 *
 * pr_tcp.c		- TCP/IP protocol module
 */

static char *rcsid = "$Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/pr_tcp.c,v 1.60 1995/06/15 03:17:57 tuecke Exp $";

#include "internal.h"

#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <sys/param.h>
#include <rpc/types.h>
#include <rpc/xdr.h>

#if defined(NEXUS_ARCH_AIX) 
#include <sys/select.h>
#if defined(NEXUS_ARCH_AIX_DCE_THREADS) && defined(NEXUS_THREAD_SAFE_SELECT)
#include <dce/cma_ux.h>
#endif
#endif

#ifdef NEXUS_LITE
#undef NEXUS_THREAD_SAFE_SELECT
#endif /* NEXUS_LITE */

#ifdef NEXUS_ARCH_AXPOSF
#define NX_NTOHL(m) m
#define NX_HTONL(m) m
#else
#define NX_NTOHL(m) ntohl(m)
#define NX_HTONL(m) htonl(m)
#endif

#ifdef NEXUS_SPOOF_COPY
nexus_bool_t _nx_spoof_copy = NEXUS_FALSE;
#endif

#ifndef SOMAXCONN /* for Linux */
#define SOMAXCONN 5
#endif

static int xdrrc;

#ifdef NEXUS_THREAD_SAFE_SELECT
/*
 * Thread is handler?
 *
 * Thread specific storage is used to keep track if the current
 * thread is a handler thread or not.
 */
static nexus_thread_key_t i_am_tcp_handler_thread_key;
#define _nx_set_i_am_tcp_handler_thread() \
    nexus_thread_setspecific(i_am_tcp_handler_thread_key, (void *) 1)
#define _nx_i_am_tcp_handler_thread(Result) \
    *(Result) = (nexus_bool_t)nexus_thread_getspecific(i_am_tcp_handler_thread_key)
#endif /* NEXUS_THREAD_SAFE_SELECT */


/*
 * Only one thread is allowed to be in the tcp code (and thus
 * mucking with data structures) at a time.
 */

#ifdef NEXUS_LITE
#define tcp_enter()
#define tcp_exit()
#else  /* NEXUS_LITE */
static nexus_mutex_t		tcp_mutex;
#define tcp_enter() nexus_mutex_lock(&tcp_mutex);
#define tcp_exit()  nexus_mutex_unlock(&tcp_mutex);
#endif /* NEXUS_LITE */

static nexus_bool_t		tcp_done;
static nexus_bool_t		handle_in_progress;

#ifdef NEXUS_THREAD_SAFE_SELECT
static nexus_bool_t		using_handler_thread;
static nexus_bool_t		handler_thread_done;
static int			pipe_to_self[2]; /* 0=read, 1=write */
static nexus_mutex_t		handler_thread_done_mutex;
static nexus_cond_t		handler_thread_done_cond;
#endif /* NEXUS_THREAD_SAFE_SELECT */

#define tcp_fatal tcp_exit(); nexus_fatal

    
/*
 * On an old version of HP/UX (7.something, I believe), if
 * you did a non-blocking write of a buffer that was greater
 * than 4096 bytes, the write would return a "Message too long"
 * error, instead of writing what it could and then returning.
 *
 * If you ever hit a situation like this again, set
 * MAX_WRITE_PACKET to the maximum number of bytes that you
 * want passed to write() at a time.  For example:
 *	#define MAX_WRITE_PACKET 4096
 *
 * If MAX_WRITE_PACKET==0, then it will always just write the
 * full buffer, no matter how long it is.
 *
 * This can also be useful if you want to break write operations
 * into blocks for, say, performance reasons.
 */
#define MAX_WRITE_PACKET 0

#ifdef NEXUS_ARCH_HPUX
#define NEXUS_FD_SET_CAST (int *)
#else
#define NEXUS_FD_SET_CAST (fd_set *)
#endif

	
/* If connect() fails with ECONNREFUSED, try again this many times quickly */
#define CONNECT_BACKOFF_FAST_RETRY	5

/* After connect() fails CONNECT_BACKOFF_FAST_RETRY times, sleep for
 * this many microseconds between connect attempts */
#define CONNECT_BACKOFF_USEC		100000

/* Save at least this many file descriptors for the user to use */
#define MIN_SAVE_FDS		3

/* Save this many files descriptors for the user by default */
#define DEFAULT_SAVE_FDS	10

/*
 * Other useful defines
 */
#define INCOMING_RECV_CLOSED           -1
#define INCOMING_RECV_HEADER		0
#define INCOMING_RECV_BODY		1

#define BLOCKING			NEXUS_TRUE
#define NONBLOCKING			NEXUS_FALSE

#define HANDLE_MESSAGES			1
#define ENQUEUE_MESSAGES		2

/* Argument choices for select_and_read() */
#define POLL_ONE_SELECT			1
#define POLL_UNTIL_SELECT_FAILS		2
#define POLL_BLOCKING_UNTIL_RECEIVE	3
#define POLL_FROM_HANDLER_THREAD	4
#define POLL_ONE_TIMED_SELECT		5
	
#define OPEN_FLAG		0
#define CLOSE_NORMAL_FLAG	1
#define CLOSE_ABNORMAL_FLAG	2
#define CLOSE_SHUTDOWN_FLAG	3


/*
 * ROUND_MALLOC
 *
 * When a message buffer is allocated, the size of the header information
 * at the head of the buffer is rounded up to ROUND_MALLOC bytes.  Further,
 * the user's part of that buffer is kept on this boundary as well.
 * This is done so that the user's portion of the buffer can be kept
 * word aligned, which could potentially speed copy operations.
 *
 * If ROUND_MALLOC==0, then the header size will not be rounded up.
 */
#if defined(NEXUS_ARCH_CRAY) || defined(NEXUS_ARCH_AXP)
#define ROUND_MALLOC	8
#else
#define ROUND_MALLOC	4
#endif


/*
 * TCP_MSG_HEADER_INTS
 * TCP_MSG_HEADER_SIZE
 *
 * When sending a message, we send a header which contains
 * a single byte for a flag (OPEN_FLAG, CLOSE_NORMAL_FLAG,
 * CLOSE_ABNORMAL_FLAG, CLOSE_SHUTDOWN_FLAG), followed
 * by unsigned longs for:
 *
 *   1) Message size
 *   2) Message is in XDR format
 *   3) Context in the gp
 *   4) Address in the gp
 *   5) Handler id
 *   6) Length of the handler name string
 *   7) node_id in the gp    (#ifdef NEXUS_PROFILE)
 *   8) context_id in the gp (#ifdef NEXUS_PROFILE)
 *
 * Cray note: To have the Cray interoperate with other machines, we
 * need to pack the unsigned longs into 4 bytes instead of 8.
 * But in a Cray only version it might be more efficient to
 * just send 8 byte unsigned longs.
 *
 * TCP_MSG_HEADER_INTS is the number of ints in the header.
 * TCP_MSG_HEADER_SIZE is the size of the header in bytes.
 */
#ifdef NEXUS_PROFILE
#define TCP_MSG_HEADER_INTS		8
#else
#define TCP_MSG_HEADER_INTS		6
#endif

#if defined(NEXUS_ARCH_CRAY) || defined(NEXUS_ARCH_AXP)
#define TCP_MSG_HEADER_SIZE		((TCP_MSG_HEADER_INTS * 8) + 1)
#else  /* NEXUS_ARCH_CRAY */
#define TCP_MSG_HEADER_SIZE		((TCP_MSG_HEADER_INTS * 4) + 1)
#endif /* NEXUS_ARCH_CRAY */


/*
 * Some forward typedef declarations...
 */
typedef struct _tcp_buffer_t	tcp_buffer_t;
typedef struct _tcp_proto_t	tcp_proto_t;
typedef struct _tcp_incoming_t	tcp_incoming_t;
typedef struct _tcp_listener_t	tcp_listener_t;


/*
 * tcp_buffer_t
 *
 * This is an overload of nexus_buffer_t.  It adds the
 * tcp specific information to that structure.
 */
struct _tcp_buffer_t
{
#ifdef NEXUS_DEBUG
    int				magic;
#endif
    nexus_buffer_funcs_t *	funcs;
    tcp_buffer_t *		next;
    int				buffer_type; /* SEND_BUFFER || RECV_BUFFER */
    char *			storage;
    XDR				xdrs;
    int				use_xdr;
    union
    {
	struct /* send */
	{
	    nexus_global_pointer_t *	gp;
	    int				handler_id;
	    char *			handler_name;
	    unsigned long		handler_name_length;
	    unsigned long		header_size;
	    unsigned long		size;
	    int				n_elements;
	    char *			header_pointer;
	    char *			name_pointer;
	    char *			base_pointer;
	    char *			current_pointer;
	} send;
	struct /* recv */
	{
	    int				handler_id;
	    char *			handler_name;
	    unsigned long		handler_name_length;
	    nexus_handler_type_t	handler_type;
	    nexus_handler_func_t	handler_func;
	    unsigned long		context;
	    unsigned long		address;
#ifdef NEXUS_PROFILE
	    int				node_id;
	    int				context_id;
#endif /* NEXUS_PROFILE */    
	    nexus_bool_t		stashed;
	    unsigned long		size;
	    char *			current_pointer;
	} recv;
    } u;
};

/*
 * A tcp_buffer_t free list, to avoid malloc calls on the
 * main body of a message buffer.
 */
static tcp_buffer_t *buffer_free_list = (tcp_buffer_t *) NULL;

#ifdef NEXUS_LITE
#define lock_buffer_free_list()
#define unlock_buffer_free_list()
#else  /* NEXUS_LITE */
static nexus_mutex_t	buffer_free_list_mutex;
#define lock_buffer_free_list()   nexus_mutex_lock(&buffer_free_list_mutex)
#define unlock_buffer_free_list() nexus_mutex_unlock(&buffer_free_list_mutex)
#endif /* NEXUS_LITE */

#define SEND_BUFFER	1
#define RECV_BUFFER	2

#ifdef NEXUS_DEBUG
#define MallocTcpBuffer(Routine, Buf) \
    { \
	NexusMalloc(Routine, Buf, tcp_buffer_t *, \
		    sizeof(struct _tcp_buffer_t)); \
	Buf->magic = NEXUS_BUFFER_MAGIC; \
    }
#else  /* NEXUS_DEBUG */
#define MallocTcpBuffer(Routine, Buf) \
    { \
	NexusMalloc(Routine, Buf, tcp_buffer_t *, \
		    sizeof(struct _tcp_buffer_t)); \
    }
#endif /* NEXUS_DEBUG */

#define GetTcpBuffer(Routine, Buf) \
    lock_buffer_free_list(); \
    if (buffer_free_list) \
    { \
	Buf = buffer_free_list; \
	buffer_free_list = buffer_free_list->next; \
	unlock_buffer_free_list(); \
    } \
    else \
    { \
	unlock_buffer_free_list(); \
	MallocTcpBuffer(Routine, Buf); \
    }

#define FreeTcpBuffer(Buf) \
{ \
    if ((Buf)->storage != (char *) NULL) \
	NexusFree((Buf)->storage); \
    lock_buffer_free_list(); \
    (Buf)->next = buffer_free_list; \
    buffer_free_list = (Buf); \
    unlock_buffer_free_list(); \
}

#ifdef NEXUS_ARCH_HPUX
#define HP_HACK_CAST (int)
#else
#define HP_HACK_CAST
#endif
#define SetBufferPointers(Buf) \
{ \
    (Buf)->u.send.base_pointer \
	= ((Buf)->storage + (Buf)->u.send.header_size); \
    (Buf)->u.send.name_pointer \
	= ((Buf)->u.send.base_pointer \
	   - HP_HACK_CAST (Buf)->u.send.handler_name_length); \
    (Buf)->u.send.header_pointer \
	= ((Buf)->u.send.name_pointer - TCP_MSG_HEADER_SIZE); \
}


/*
 * tcp_proto_t
 *
 * This is an overload of nexus_proto_t.  It adds the
 * tcp specific information to that structure.
 */
struct _tcp_proto_t
{
    nexus_proto_type_t	type;	/* NEXUS_PROTO_TYPE_TCP */
    nexus_proto_funcs_t *funcs;
    char *		host;
    u_short		port;
    int			fd;
    int			reference_count;
    nexus_bool_t	send_in_progress;
#ifndef NEXUS_LITE
    int			sends_waiting;
    nexus_cond_t	sends_waiting_cond;
#endif /* NEXUS_LITE */
};


/*
 * tcp_incoming_t
 *
 * One of these is allocated for each incoming file descriptor.
 * It stores partial incoming messages until they are complete
 * and can be handled or queued.
 *
 * The 'type' field should be in the same location as the 'type'
 * field in a tcp_proto_t, so that tcp_proto_t's and
 * tcp_incoming_t's can be distinguished from each other.
 *
 * The 'header_buf' will be used to temporarily store the header
 * information as it is read, before it is parsed.
 */
#define INCOMING_TYPE NEXUS_PROTO_TYPE_INTERNAL1
#define HEADER_BUF_LENGTH TCP_MSG_HEADER_SIZE

struct _tcp_incoming_t
{
    nexus_proto_type_t	type;	/* INCOMING_TYPE */
    int			fd;
    tcp_incoming_t *	next_free;

#ifdef NEXUS_DEBUG
    int			connected_pid;
#endif /* NEXUS_DEBUG */

    /*
     * The rest of this is for handling non-blocking reads and writes,
     * and the resulting partial messages that this data structure must
     * hang onto.
     */
    int   recv_state;		/* receiving header or body */
    tcp_buffer_t *recv_buf;	/* Buffer we're receiving into */
    char *recv_buf_current;	/* Place in buffer that we're receving into */
    int   recv_n_left;		/* # bytes left to receive */
    char  header_buf[HEADER_BUF_LENGTH]; /* space to recv message header */
};

static tcp_incoming_t *incoming_free_list = (tcp_incoming_t *) NULL;

#define ResetIncoming(Incoming) \
{ \
    (Incoming)->recv_state = INCOMING_RECV_HEADER; \
    (Incoming)->recv_buf = (tcp_buffer_t *) NULL; \
    (Incoming)->recv_buf_current = &(Incoming->header_buf[0]); \
    (Incoming)->recv_n_left = TCP_MSG_HEADER_SIZE; \
}


/*
 * This is a message queue containing all messages that have
 * arrived and are waiting to be handled.
 */
static tcp_buffer_t *	message_queue_head;
static tcp_buffer_t *	message_queue_tail;

#define EnqueueMessage(buf) \
{ \
    if (message_queue_head == (tcp_buffer_t *) NULL) \
    { \
	message_queue_head = message_queue_tail = buf; \
    } \
    else \
    { \
	message_queue_tail->next = buf; \
	message_queue_tail = buf; \
    } \
    buf->next = (tcp_buffer_t *) NULL; \
}
#define DequeueMessage(buf) \
{ \
    /* Assume that message_queue_head != NULL */ \
    buf = message_queue_head; \
    message_queue_head = message_queue_head->next; \
}
#define MessagesEnqueued() (message_queue_head != (tcp_buffer_t *) NULL)

#define TestAndHandleMessages() \
    if (handle_or_enqueue == HANDLE_MESSAGES && MessagesEnqueued()) \
    { \
	handle_enqueued_messages(); \
    }


/*
 * tcp_listener_t
 *
 * One of these is allocated for each file descriptor that is
 * used for listening on a port.
 */
#define LISTENER_TYPE NEXUS_PROTO_TYPE_INTERNAL2

struct _tcp_listener_t
{
    nexus_proto_type_t		type;	/* LISTENER_TYPE */
    int				fd;
    u_short			port;
    nexus_bool_t		internal;
    int (*approval_func)(char *url, nexus_global_pointer_t *gp);
    struct _tcp_listener_t *	next;
    struct _tcp_listener_t *	prev;
};

static tcp_listener_t *all_listeners = (tcp_listener_t *) NULL;

/*
 * My local TCP connection information
 */
static u_short	tcp_local_port;
static int	tcp_local_fd;
static int	tcp_local_pid;
static char *	tcp_local_host;


/*
 * Stuff for keeping track of file descriptors.
 *   fd_tablesize:	Maximum number of fds that system supports
 *   max_fds_to_use:	Maximum number of fds that we will open at once
 *   n_fds_in_use:	Number of fds that we are currently using
 *   fd_to_proto:	Array of pointers to tcp_proto_t's or
 *				tcp_incoming_t's.  This is used
 *				to map from an fd to the appropriate
 *				proto or incoming data structure.
 *   current_fd_set:	The current fd_set that should be used for select()
 *   next_fd_to_close:  If we run out of fd, start looking for an fd
 *				to close starting at this fd.
 *   modified_fd_table:	Set to NEXUS_TRUE whenever the fd table
 *				changes.  This signals the
 *				handler thread which is blocked in
 *				a select that the fd_table is
 *				different than when it was called,
 *				and therefore should not be trusted.
 */
static int fd_tablesize;
static int max_fds_to_use;
static int n_fds_in_use;
static void **fd_to_proto;
static fd_set current_fd_set;
static int next_fd_to_close;

#ifdef NEXUS_THREAD_SAFE_SELECT
static nexus_bool_t modified_fd_table;
#endif

#ifdef NEXUS_THREAD_SAFE_SELECT
/*
 * When using a blocking select from a handler thread, we do
 * not want another thread to close an fd out from under the select.
 * Under HP-UX 9.x with DCE threads, this causes the select to
 * bomb out.
 *
 * Our solution is to defer the close until after the select returns.
 *
 *   blocking_select_in_process: Set to NEXUS_TRUE while a blocking
 *				select() call is in progress.
 *   fd_needs_closing:	Set to NEXUS_TRUE whenever an fd in the table
 *				needs closing, but a select is blocked
 *				which uses this fd.
 *   close_fd_set:	Used to keep track of which fd's need to be closed.
 */
static nexus_bool_t	blocking_select_in_progress;
static nexus_bool_t	fd_needs_closing;
static fd_set		close_fd_set;
#endif /* NEXUS_THREAD_SAFE_SELECT */


#if defined(NEXUS_ARCH_SOLARIS) || defined(NEXUS_ARCH_HPUX)
#define NUM_FDS sysconf(_SC_OPEN_MAX)
#else
/* This works for BSD machines */
#define NUM_FDS getdtablesize()
#endif


/*
 * When tcp_destroy_proto() decrements proto->reference_count to 0, it
 * tries to close proto->fd by sending a close message down the fd
 * and then closing the fd.  But its possible that it cannot close
 * the proto because the socket buffer was full.
 * In this case, it bumps n_deferred_closes
 * so that we can try to close again later in make_room_for_new_fd().
 *
 * Do a deferred closed proto is one with:
 *     (proto->reference_count == 0) && (proto->fd >= 0)
 */
static int n_deferred_closes = 0;


/*
 * Protocol table stuff.
 *
 * The protocol table is hashed on host/port. The table itself is an
 * array of header structures pointing to a linked list of buckets.
 *
 * This table is used to avoid creating multiple tcp_proto_t
 * objects to the same context.  Multiple global pointers to the same
 * context share a tcp_proto_t.
 */
typedef struct _proto_table_entry_t
{
    tcp_proto_t *proto;
    struct _proto_table_entry_t *next;
} proto_table_entry_t;

#define PROTO_TABLE_SIZE 149

struct _proto_table_entry_t	proto_table[PROTO_TABLE_SIZE];

static void			proto_table_init();
static int			proto_table_hash(char *host, u_short port);
static void			proto_table_insert(tcp_proto_t *proto);
static tcp_proto_t *		proto_table_lookup(char *host, u_short port);

#ifdef NEXUS_DEBUG
static char tmpbuf1[10240];
static char tmpbuf2[10240];
#endif


/*
 * Various forward declarations of procedures
 */
static nexus_proto_type_t	tcp_proto_type();
static void		tcp_init(int *argc, char ***argv);
static void		tcp_shutdown(nexus_bool_t shutdown_others);
static void		tcp_abort();
static void		tcp_poll();
static void		tcp_blocking_poll();

static void		tcp_init_remote_service_request(
					    nexus_buffer_t *buffer,
					    nexus_global_pointer_t *gp,
					    char *handler_name,
					    int handler_id);
static void		tcp_send_remote_service_request(
						   nexus_buffer_t *buffer);
static void		tcp_send_urgent_remote_service_request(
						   nexus_buffer_t *buffer);
static void		tcp_destroy_proto(nexus_proto_t *nexus_proto);
static nexus_mi_proto_t*tcp_get_my_mi_proto();
static nexus_bool_t	tcp_construct_from_mi_proto(nexus_proto_t **proto,
						  nexus_mi_proto_t *mi_proto);
#ifndef NEXUS_LITE
static void		tcp_get_creator_proto_params(char *buf,
						     int buf_size,
						     char *node_name,
						     int node_number);
static void		tcp_construct_creator_proto(nexus_proto_t **proto);
#endif /* NEXUS_LITE */
static int		tcp_compare_protos(nexus_proto_t *proto1,
					   nexus_proto_t *proto2);


static tcp_proto_t *	construct_proto(char *host, u_short port, int fd);
static void		open_proto(tcp_proto_t *proto);
static void		close_proto(tcp_proto_t *proto);

static tcp_incoming_t *	construct_incoming(int fd);
static void		close_incoming(tcp_incoming_t *incoming);

static nexus_bool_t	close_incoming_with_message(tcp_incoming_t *incoming,
						    nexus_bool_t read_messages);
static nexus_bool_t	close_proto_with_message(tcp_proto_t *proto);
static void		check_proto_for_close(tcp_proto_t *proto);
static tcp_listener_t *	construct_listener(int fd,
					   u_short port,
					   nexus_bool_t internal);
static void		close_listener(tcp_listener_t *listener);

static void		add_proto_with_fd(void *proto, int fd);
static void		remove_proto_with_fd(void *proto, int fd);
static void		make_room_for_new_fd();

static nexus_bool_t	select_and_read(int poll_type, int handle_or_enqueue);
static void		read_incoming(tcp_incoming_t *incoming,
				      int handle_or_enqueue);
static void		handle_enqueued_messages();

static void		accept_new_connection(tcp_listener_t *listener);
static void		config_socket(int s);
static void		set_nonblocking(int fd);
static int		do_connect(char *host, u_short port);
static int		setup_listener(u_short *port);

static void		accept_attachment(tcp_listener_t *listener,
					  int new_fd);

#ifdef NEXUS_DEBUG
static void		print_fd_set(char *fmt_string, fd_set *set);
#endif /* NEXUS_DEBUG */

static void		tcp_set_buffer_size(nexus_buffer_t *buffer,
					    int size, int n_elements);
static int		tcp_check_buffer_size(nexus_buffer_t *buffer,
					      int slack, int increment);
static void		tcp_free_buffer(nexus_buffer_t *buffer);
static void		tcp_stash_buffer(nexus_buffer_t *buffer,
				       nexus_stashed_buffer_t *stashed_buffer);
static void		tcp_free_stashed_buffer(
				       nexus_stashed_buffer_t *stashed_buffer);


static int	tcp_sizeof_float	(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_double	(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_short	(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_u_short	(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_int		(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_u_int	(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_long		(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_u_long	(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_char		(nexus_buffer_t *buffer, int count);
static int	tcp_sizeof_u_char	(nexus_buffer_t *buffer, int count);


static void	tcp_put_float	(nexus_buffer_t *buffer,
				 float *data, int count);
static void	tcp_put_double	(nexus_buffer_t *buffer,
				 double *data, int count);
static void	tcp_put_short	(nexus_buffer_t *buffer,
				 short *data, int count);
static void	tcp_put_u_short	(nexus_buffer_t *buffer,
				 unsigned short *data, int count);
static void	tcp_put_int	(nexus_buffer_t *buffer,
				 int *data, int count);
static void	tcp_put_u_int	(nexus_buffer_t *buffer,
				 unsigned int *data, int count);
static void	tcp_put_long	(nexus_buffer_t *buffer,
				 long *data, int count);
static void	tcp_put_u_long	(nexus_buffer_t *buffer,
				 unsigned long *data, int count);
static void	tcp_put_char	(nexus_buffer_t *buffer,
				 char *data, int count);
static void	tcp_put_u_char	(nexus_buffer_t *buffer,
				 unsigned char *data, int count);

static void	tcp_get_float	(nexus_buffer_t *buffer,
				 float *dest, int count);
static void	tcp_get_double	(nexus_buffer_t *buffer,
				 double *dest, int count);
static void	tcp_get_short	(nexus_buffer_t *buffer,
				 short *dest, int count);
static void	tcp_get_u_short	(nexus_buffer_t *buffer,
				 unsigned short *dest, int count);
static void	tcp_get_int	(nexus_buffer_t *buffer,
				 int *dest, int count);
static void	tcp_get_u_int	(nexus_buffer_t *buffer,
				 unsigned int *dest, int count);
static void	tcp_get_long	(nexus_buffer_t *buffer,
				 long *dest, int count);
static void	tcp_get_u_long	(nexus_buffer_t *buffer,
				 unsigned long *dest, int count);
static void	tcp_get_char	(nexus_buffer_t *buffer,
				 char *dest, int count);
static void	tcp_get_u_char	(nexus_buffer_t *buffer,
				 unsigned char *dest, int count);

static void	tcp_get_stashed_float	(nexus_stashed_buffer_t *buffer,
					 float *dest, int count);
static void	tcp_get_stashed_double	(nexus_stashed_buffer_t *buffer,
					 double *dest, int count);
static void	tcp_get_stashed_short	(nexus_stashed_buffer_t *buffer,
					 short *dest, int count);
static void	tcp_get_stashed_u_short	(nexus_stashed_buffer_t *buffer,
					 unsigned short *dest, int count);
static void	tcp_get_stashed_int	(nexus_stashed_buffer_t *buffer,
					 int *dest, int count);
static void	tcp_get_stashed_u_int	(nexus_stashed_buffer_t *buffer,
					 unsigned int *dest, int count);
static void	tcp_get_stashed_long	(nexus_stashed_buffer_t *buffer,
					 long *dest, int count);
static void	tcp_get_stashed_u_long	(nexus_stashed_buffer_t *buffer,
					 unsigned long *dest, int count);
static void	tcp_get_stashed_char	(nexus_stashed_buffer_t *buffer,
					 char *dest, int count);
static void	tcp_get_stashed_u_char	(nexus_stashed_buffer_t *buffer,
					 unsigned char *dest, int count);

static void		tcp_xdr_set_buffer_size(nexus_buffer_t *buffer,
					    int size, int n_elements);
static int		tcp_xdr_check_buffer_size(nexus_buffer_t *buffer,
					      int slack, int increment);
static void		tcp_xdr_free_buffer(nexus_buffer_t *buffer);
static void		tcp_xdr_stash_buffer(nexus_buffer_t *buffer,
				       nexus_stashed_buffer_t *stashed_buffer);
static void		tcp_xdr_free_stashed_buffer(
				       nexus_stashed_buffer_t *stashed_buffer);

static int	tcp_xdr_sizeof_float	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_double	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_short	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_u_short	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_int	 (nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_u_int	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_long	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_u_long	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_char	(nexus_buffer_t *buffer, int count);
static int	tcp_xdr_sizeof_u_char	(nexus_buffer_t *buffer, int count);


static void	tcp_xdr_put_float	(nexus_buffer_t *buffer,
				 float *data, int count);
static void	tcp_xdr_put_double	(nexus_buffer_t *buffer,
				 double *data, int count);
static void	tcp_xdr_put_short	(nexus_buffer_t *buffer,
				 short *data, int count);
static void	tcp_xdr_put_u_short	(nexus_buffer_t *buffer,
				 unsigned short *data, int count);
static void	tcp_xdr_put_int	(nexus_buffer_t *buffer,
				 int *data, int count);
static void	tcp_xdr_put_u_int	(nexus_buffer_t *buffer,
				 unsigned int *data, int count);
static void	tcp_xdr_put_long	(nexus_buffer_t *buffer,
				 long *data, int count);
static void	tcp_xdr_put_u_long	(nexus_buffer_t *buffer,
				 unsigned long *data, int count);
static void	tcp_xdr_put_char	(nexus_buffer_t *buffer,
				 char *data, int count);
static void	tcp_xdr_put_u_char	(nexus_buffer_t *buffer,
				 unsigned char *data, int count);

static void	tcp_xdr_get_float	(nexus_buffer_t *buffer,
				 float *dest, int count);
static void	tcp_xdr_get_double	(nexus_buffer_t *buffer,
				 double *dest, int count);
static void	tcp_xdr_get_short	(nexus_buffer_t *buffer,
				 short *dest, int count);
static void	tcp_xdr_get_u_short	(nexus_buffer_t *buffer,
				 unsigned short *dest, int count);
static void	tcp_xdr_get_int	(nexus_buffer_t *buffer,
				 int *dest, int count);
static void	tcp_xdr_get_u_int	(nexus_buffer_t *buffer,
				 unsigned int *dest, int count);
static void	tcp_xdr_get_long	(nexus_buffer_t *buffer,
				 long *dest, int count);
static void	tcp_xdr_get_u_long	(nexus_buffer_t *buffer,
				 unsigned long *dest, int count);
static void	tcp_xdr_get_char	(nexus_buffer_t *buffer,
				 char *dest, int count);
static void	tcp_xdr_get_u_char	(nexus_buffer_t *buffer,
				 unsigned char *dest, int count);

static void	tcp_xdr_get_stashed_float	(nexus_stashed_buffer_t *buffer,
					 float *dest, int count);
static void	tcp_xdr_get_stashed_double	(nexus_stashed_buffer_t *buffer,
					 double *dest, int count);
static void	tcp_xdr_get_stashed_short	(nexus_stashed_buffer_t *buffer,
					 short *dest, int count);
static void	tcp_xdr_get_stashed_u_short	(nexus_stashed_buffer_t *buffer,
					 unsigned short *dest, int count);
static void	tcp_xdr_get_stashed_int	(nexus_stashed_buffer_t *buffer,
					 int *dest, int count);
static void	tcp_xdr_get_stashed_u_int	(nexus_stashed_buffer_t *buffer,
					 unsigned int *dest, int count);
static void	tcp_xdr_get_stashed_long	(nexus_stashed_buffer_t *buffer,
					 long *dest, int count);
static void	tcp_xdr_get_stashed_u_long	(nexus_stashed_buffer_t *buffer,
					 unsigned long *dest, int count);
static void	tcp_xdr_get_stashed_char	(nexus_stashed_buffer_t *buffer,
					 char *dest, int count);
static void	tcp_xdr_get_stashed_u_char	(nexus_stashed_buffer_t *buffer,
					 unsigned char *dest, int count);

static nexus_proto_funcs_t tcp_proto_funcs =
{
    tcp_proto_type,
    tcp_init,
    tcp_shutdown,
    tcp_abort,
    tcp_poll,
    tcp_blocking_poll,
    tcp_init_remote_service_request,
    tcp_destroy_proto,
    tcp_get_my_mi_proto,
    tcp_construct_from_mi_proto,
#ifdef NEXUS_LITE
    NULL, /* tcp_get_creator_proto_params */
    NULL, /* tcp_construct_creator_proto */
#else  /* NEXUS_LITE */
    tcp_get_creator_proto_params,
    tcp_construct_creator_proto,
#endif /* NEXUS_LITE */
    tcp_compare_protos,
};

static nexus_buffer_funcs_t tcp_buffer_funcs =
{
    tcp_set_buffer_size,
    tcp_check_buffer_size,
    tcp_send_remote_service_request,
    tcp_send_urgent_remote_service_request,
    tcp_free_buffer,
    tcp_stash_buffer,
    tcp_free_stashed_buffer,
    tcp_sizeof_float,
    tcp_sizeof_double,
    tcp_sizeof_short,
    tcp_sizeof_u_short,
    tcp_sizeof_int,
    tcp_sizeof_u_int,
    tcp_sizeof_long,
    tcp_sizeof_u_long,
    tcp_sizeof_char,
    tcp_sizeof_u_char,
    tcp_put_float,
    tcp_put_double,
    tcp_put_short,
    tcp_put_u_short,
    tcp_put_int,
    tcp_put_u_int,
    tcp_put_long,
    tcp_put_u_long,
    tcp_put_char,
    tcp_put_u_char,
    tcp_get_float,
    tcp_get_double,
    tcp_get_short,
    tcp_get_u_short,
    tcp_get_int,
    tcp_get_u_int,
    tcp_get_long,
    tcp_get_u_long,
    tcp_get_char,
    tcp_get_u_char,
    tcp_get_stashed_float,
    tcp_get_stashed_double,
    tcp_get_stashed_short,
    tcp_get_stashed_u_short,
    tcp_get_stashed_int,
    tcp_get_stashed_u_int,
    tcp_get_stashed_long,
    tcp_get_stashed_u_long,
    tcp_get_stashed_char,
    tcp_get_stashed_u_char,
};

static nexus_buffer_funcs_t tcp_xdr_buffer_funcs =
{
    tcp_xdr_set_buffer_size,
    tcp_xdr_check_buffer_size,
    tcp_send_remote_service_request,
    tcp_send_urgent_remote_service_request,
    tcp_xdr_free_buffer,
    tcp_xdr_stash_buffer,
    tcp_xdr_free_stashed_buffer,
    tcp_xdr_sizeof_float,
    tcp_xdr_sizeof_double,
    tcp_xdr_sizeof_short,
    tcp_xdr_sizeof_u_short,
    tcp_xdr_sizeof_int,
    tcp_xdr_sizeof_u_int,
    tcp_xdr_sizeof_long,
    tcp_xdr_sizeof_u_long,
    tcp_xdr_sizeof_char,
    tcp_xdr_sizeof_u_char,
    tcp_xdr_put_float,
    tcp_xdr_put_double,
    tcp_xdr_put_short,
    tcp_xdr_put_u_short,
    tcp_xdr_put_int,
    tcp_xdr_put_u_int,
    tcp_xdr_put_long,
    tcp_xdr_put_u_long,
    tcp_xdr_put_char,
    tcp_xdr_put_u_char,
    tcp_xdr_get_float,
    tcp_xdr_get_double,
    tcp_xdr_get_short,
    tcp_xdr_get_u_short,
    tcp_xdr_get_int,
    tcp_xdr_get_u_int,
    tcp_xdr_get_long,
    tcp_xdr_get_u_long,
    tcp_xdr_get_char,
    tcp_xdr_get_u_char,
    tcp_xdr_get_stashed_float,
    tcp_xdr_get_stashed_double,
    tcp_xdr_get_stashed_short,
    tcp_xdr_get_stashed_u_short,
    tcp_xdr_get_stashed_int,
    tcp_xdr_get_stashed_u_int,
    tcp_xdr_get_stashed_long,
    tcp_xdr_get_stashed_u_long,
    tcp_xdr_get_stashed_char,
    tcp_xdr_get_stashed_u_char,
};


/*
 * _nx_pr_tcp_info()
 *
 * Return the nexus_proto_funcs_t function table for this protocol module.
 *
 * This procedure is used for bootstrapping the protocol module.
 * The higher level Nexus code needs to call this routine to
 * retrieve the functions it needs to use this protocol module.
 */
void *_nx_pr_tcp_info()
{
    return((void *) (&tcp_proto_funcs));
} /* _nx_pr_tcp_info() */


/*
 * tcp_proto_type()
 *
 * Return the nexus_proto_type_t for this protocol module.
 */
static nexus_proto_type_t tcp_proto_type()
{
    return (NEXUS_PROTO_TYPE_TCP);
} /* tcp_proto_type() */


#ifdef NEXUS_THREAD_SAFE_SELECT
/*
 * tcp_handler_thread()
 *
 * In the multi-threaded version, this is the entry point
 * for the handler thread.
 */
static void *tcp_handler_thread(void *arg)
{
    _nx_set_i_am_tcp_handler_thread();
    
    tcp_enter();
    select_and_read(POLL_FROM_HANDLER_THREAD, HANDLE_MESSAGES);
    tcp_exit();

    nexus_mutex_lock(&handler_thread_done_mutex);
    handler_thread_done = NEXUS_TRUE;
    nexus_cond_signal(&handler_thread_done_cond);
    nexus_mutex_unlock(&handler_thread_done_mutex);
    
    return ((void *) NULL);
} /* tcp_handler_thread() */
#endif /* NEXUS_THREAD_SAFE_SELECT */


/*
 * tcp_init()
 *
 * Initialize the TCP protocol.
 */
static void tcp_init(int *argc, char ***argv)
{
    int i;
    int save_fds;
    int arg_num;
    tcp_listener_t *listener;
    
    /*
     * Set up the listener for this process. The nexus_proto_t returned
     * refers to this tcp port.
     */
    tcp_local_port = 0;
    tcp_local_fd = setup_listener(&tcp_local_port);
    if (tcp_local_fd < 0)
	nexus_fatal("tcp_init(): setup_listener() failed\n");
    _nx_md_get_host_id(&tcp_local_host, &tcp_local_pid);
    listener = construct_listener(tcp_local_fd, tcp_local_port, NEXUS_TRUE);

    /*
     * Set up file descriptor tables
     */
    fd_tablesize = NUM_FDS;
    if ((arg_num = nexus_find_argument(argc, argv, "save_fds", 2)) >= 0)
    {
	save_fds = atoi((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	save_fds = DEFAULT_SAVE_FDS;
    }
    save_fds = NEXUS_MIN(NEXUS_MAX(save_fds, MIN_SAVE_FDS), fd_tablesize);
    max_fds_to_use = fd_tablesize - save_fds - 1; /* -1 for bootstrap_fd */
    n_fds_in_use = 0;
    NexusMalloc(tcp_init(), fd_to_proto, void **,
		sizeof(void *) * fd_tablesize);
    for (i = 0; i < fd_tablesize; i++)
	fd_to_proto[i] = (void *) NULL;
    FD_ZERO(&current_fd_set);
    add_proto_with_fd((void *) listener, tcp_local_fd);

    next_fd_to_close = 0;
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    blocking_select_in_progress = NEXUS_FALSE;
    fd_needs_closing = NEXUS_FALSE;
    FD_ZERO(&close_fd_set);
#endif /* NEXUS_THREAD_SAFE_SELECT */
    
    nexus_debug_printf(3, ("tcp_init(): Listening on %d\n", tcp_local_port));

    message_queue_head = message_queue_tail = (tcp_buffer_t *) NULL;
    proto_table_init();
    tcp_done = NEXUS_FALSE;
    handle_in_progress = NEXUS_FALSE;

#ifndef NEXUS_LITE
    nexus_mutex_init(&tcp_mutex, (nexus_mutexattr_t *) NULL);
    nexus_mutex_init(&buffer_free_list_mutex, (nexus_mutexattr_t *) NULL);
#endif /* NEXUS_LITE */
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    nexus_thread_key_create(&i_am_tcp_handler_thread_key, NULL);
    
    if (nexus_preemptive_threads())
    {
	nexus_thread_t thread;
	using_handler_thread = NEXUS_TRUE;

	/*
	 * Create a pipe to myself, so that I can wake up the handler
	 * thread that is blocked on a select
	 */
	if (pipe(pipe_to_self) != 0)
	{
	    nexus_fatal("tcp_init(): pipe() failed with errno=%d\n", errno);
	}
	n_fds_in_use += 2;
	FD_SET(pipe_to_self[0], &current_fd_set);

	/* NULL out the poll and blocking_poll entries in my function table */
	tcp_proto_funcs.poll = NULL;
	tcp_proto_funcs.blocking_poll = NULL;
	
	/* Create the handler thread */
	handler_thread_done = NEXUS_FALSE;
	nexus_mutex_init(&handler_thread_done_mutex,
			 (nexus_mutexattr_t *) NULL);
	nexus_cond_init(&handler_thread_done_cond,
			(nexus_condattr_t *) NULL);
	nexus_thread_create(&thread,
			    (nexus_thread_attr_t *) NULL,
			    tcp_handler_thread,
			    (void *) NULL);
    }
    else
    {
	using_handler_thread = NEXUS_FALSE;
    }
#endif /* NEXUS_THREAD_SAFE_SELECT */
} /* tcp_init() */


/*
 * tcp_shutdown()
 *
 * This routine is called during normal shutdown of a process.
 *
 * Shutdown the tcp protocol module by closing all off of the
 * proto's that have open fd's.  This will result in an orderly
 * shutdown of fd's, so that other nodes will not detect an
 * abnormal eof on the other ends of these fd's.
 *
 * It is ok if this procedure takes a little while to complete -- for
 * example to wait for another process to read off its fd and
 * free up some socket buffer space.
 *
 * An orderly closing of a connection is done as follows:
 *   1) Send the CLOSE_NORMAL_FLAG message down the fd (1 byte,
 *      signaling a close due to a normal termination) down the fd,
 *	if shutdown_others==NEXUS_FALSE.  Otherwise send the
 *	CLOSE_SHUTDOWN_FLAG.
 *	If the write would block, then sleep for a bit and try again.
 *   2) Then close the fd and its associated proto.
 */
static void tcp_shutdown(nexus_bool_t shutdown_others)
{
    int i;
    nexus_bool_t closed;
    char close_flag;
    int fd;
    int n_sent;
    tcp_incoming_t *incoming;
    tcp_proto_t *proto;
    tcp_listener_t *listener;
    nexus_bool_t is_proto;

    tcp_enter();
    tcp_done = NEXUS_TRUE;

#ifdef NEXUS_THREAD_SAFE_SELECT
    if (using_handler_thread)
    {
	nexus_bool_t i_am_tcp_handler_thread;
	_nx_i_am_tcp_handler_thread(&i_am_tcp_handler_thread);
	if (!i_am_tcp_handler_thread)
	{
	    /*
	     * Close the pipe_to_self to wakeup the handler thread
	     * from the select, and wait for the handler thread
	     * to shutdown.
	     */
	    close_flag = '\0';
	    while (   (write(pipe_to_self[1], &close_flag, 1) == -1)
		   && (errno == EINTR) )
		;
	    while (   (close(pipe_to_self[1]) == -1)
		   && (errno == EINTR) )
		;
	    tcp_exit();
	    nexus_mutex_lock(&handler_thread_done_mutex);
	    while (!handler_thread_done)
	    {
		nexus_cond_wait(&handler_thread_done_cond,
				&handler_thread_done_mutex);
	    }
	    nexus_mutex_unlock(&handler_thread_done_mutex);
	    tcp_enter();
	}
	nexus_mutex_destroy(&handler_thread_done_mutex);
	nexus_cond_destroy(&handler_thread_done_cond);
    }
#endif /* NEXUS_THREAD_SAFE_SELECT */

    if (shutdown_others)
    {
	close_flag = CLOSE_SHUTDOWN_FLAG;
    }
    else
    {
	close_flag = CLOSE_NORMAL_FLAG;
    }
    
    for (i = 0; i < fd_tablesize; i++)
    {
	nexus_debug_printf(8, ("tcp_shutdown(): fd:%d proto:%x\n",
			       i, fd_to_proto[i]) );
	if (fd_to_proto[i] != (tcp_proto_t *) NULL)
	{
	    closed = NEXUS_FALSE;
	    
	    if (((tcp_proto_t *)(fd_to_proto[i]))->type == LISTENER_TYPE)
	    {
		listener = (tcp_listener_t *) (fd_to_proto[i]);
		close_listener(listener);
		closed = NEXUS_TRUE;
	    }
	    else if (((tcp_proto_t *)(fd_to_proto[i]))->type == INCOMING_TYPE)
	    {
		incoming = (tcp_incoming_t *) (fd_to_proto[i]);
		fd = incoming->fd;
		is_proto = NEXUS_FALSE;
	    }
	    else
	    {
		proto = (tcp_proto_t *) (fd_to_proto[i]);
		fd = proto->fd;
		is_proto = NEXUS_TRUE;
	    }
	    nexus_debug_printf(5, ("tcp_shutdown(): checking fd:%d\n",fd));
	    
	    while (!closed)
	    {
		n_sent = write(fd, &close_flag, 1);
		nexus_debug_printf(5,("tcp_shutdown(); on shut down fd:%d write returned: %d\n",fd,n_sent));
		if (n_sent > 0)
		{
		    /* Successfully sent close message */
		    if (is_proto)
		    {
			close_proto(proto);
		    }
		    else
		    {
			close_incoming(incoming);
		    }
		    closed = NEXUS_TRUE;
		}
		else if (n_sent < 0 && errno != EWOULDBLOCK && errno != EINTR)
		{
		    /* Something wierd happened on the socket, so just go on */
		    nexus_debug_printf(3, ("tcp_shutdown(): trouble shutting down fd:%d, received errno:%d\n",fd,errno));
		    closed = NEXUS_TRUE;
		}
		
		if (!closed)
		{
		    /*
		     * hitting this in the debugger caused the program
		     * to exit.  possibly indicating a race condition on
		     * shutting down the program
		     */
		    nexus_usleep(100000);  
		}
	    }
	}
    }
    
    tcp_exit();

} /* tcp_shutdown() */


/*
 * tcp_abort()
 *
 * This routine is called during the _abnormal_ shutdown of a process.
 *
 * Shutdown the tcp protocol module by closing all off of the
 * proto's that have open fd's.  This will hopefully result
 * in a reasonably  orderly shutdown of fd's, so that
 * other nodes do not spew error messages when they detect
 * closed sockets.
 *
 * However, this procedure should not wait to do things cleanly,
 * like tcp_shutdown().  If it cannot close a fd cleanly (first
 * write a CLOSE_ABNORMAL_FLAG to it), then it should close the
 * fd anyway and get on with things.  The highest priority is
 * to just get shut down.
 *
 * The closing of a connection is done as follows:
 *   1) Send the CLOSE_ABNORMAL_FLAG message down the fd (1 byte,
 *      signaling a close due to a normal termination) down the fd.
 *	If the write would block, give up and go on to next step.
 *   2) Then close the fd.
 */
static void tcp_abort()
{
    int i;
    char close_flag = CLOSE_ABNORMAL_FLAG;

    for (i = 0; i < fd_tablesize; i++)
    {
	if (fd_to_proto[i] != (tcp_proto_t *) NULL)
	{
	    if (((tcp_proto_t *)(fd_to_proto[i]))->type != LISTENER_TYPE)
	    {
		write(i, &close_flag, 1);
	    }
	    close(i);
	}
    }

} /* tcp_abort() */


/*
 * tcp_poll()
 *
 * In a version of the tcp protocol module that does not
 * use a handler thread (preemptive thread module & thread safe select)
 * this routine should check to see if there are
 * any messages to receive, and if so then receive them and invoke them.
 */
static void tcp_poll()
{
    nexus_bool_t message_handled = NEXUS_FALSE;

#ifdef NEXUS_THREAD_SAFE_SELECT
    /*
     * This should not be called if a separate handler thread is in use.
     */
    NexusAssert2((!using_handler_thread),
		 ("tcp_poll(): Internal error: Should never be called when using a handler thread\n") );
#endif /* NEXUS_THREAD_SAFE_SELECT */
    
    nexus_debug_printf(6, ("tcp_poll(): entering\n"));

    tcp_enter();
    if (!handle_in_progress)
    {
	message_handled = select_and_read(POLL_UNTIL_SELECT_FAILS,
					  HANDLE_MESSAGES);
    }
    tcp_exit();

#ifndef NEXUS_LITE
    /*
     * Only yield the processor if there was a message handled.
     * That handler may have enabled another thread for execution.
     */
    if (message_handled)
    {
	nexus_thread_yield();
    }
#endif /* NEXUS_LITE */
    
    nexus_debug_printf(6, ("tcp_poll(): exiting\n"));

} /* tcp_poll() */


/*
 * tcp_blocking_poll()
 *
 * In a version of the tcp protocol module that does not
 * use a handler thread, this routine should do a blocking
 * select and invoke the handler for any received messages.
 * This is only called if this is the only protocol module
 * requiring polling, and all other threads are suspended.
 *
 * It is ok for this to return without actually handling
 * a message.
 *
 * Note: There is no need to call nexus_thread_yield() from this
 * routine like there is in tcp_poll(), because this routine will only
 * be used from an idle or handler thread which will do a subsequent
 * nexus_thread_yield() anyway.
 */
static void tcp_blocking_poll()
{
#ifdef NEXUS_THREAD_SAFE_SELECT
    /*
     * This should not be called if a separate handler thread is in use.
     */
    NexusAssert2((!using_handler_thread),
		 ("tcp_blocking_poll(): Internal error: Should never be called when using a handler thread\n") );
#endif /* NEXUS_THREAD_SAFE_SELECT */
    
    nexus_debug_printf(2, ("tcp_blocking_poll(): entering\n"));
    tcp_enter();
    if (!handle_in_progress)
    {
	select_and_read(POLL_ONE_TIMED_SELECT,
			HANDLE_MESSAGES);
	/*
	select_and_read(POLL_BLOCKING_UNTIL_RECEIVE,
			HANDLE_MESSAGES);
	*/
    }
    tcp_exit();
    nexus_debug_printf(2, ("tcp_blocking_poll(): exiting\n"));
    
} /* tcp_blocking_poll() */


/*
 * tcp_init_remote_service_request()
 *
 * Initiate a remote service request to the node and context specified
 * in the gp, to the handler specified by handler_id and handler_name.
 *
 * Note: The things pointed to by 'gp' and 'handler_name' should
 * NOT change between this init_rsr call and the subsequent send_rsr.
 * Only the pointers are stashed away, and not the full data.
 *
 * Return: Fill in 'buffer' with a nexus_buffer_t.
 */
static void tcp_init_remote_service_request(nexus_buffer_t *buffer,
					    nexus_global_pointer_t *gp,
					    char *handler_name,
					    int handler_id)
{
    tcp_buffer_t *tcp_buffer;
    int header_size;
#if ROUND_MALLOC > 1
    int header_mod;
#endif    

    NexusAssert2((gp->proto->type == NEXUS_PROTO_TYPE_TCP),
		 ("tcp_init_remote_service_request(): Internal error: proto_type is not NEXUS_PROTO_TYPE_TCP\n"));

    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "tcp_init_remote_service_request");

    GetTcpBuffer(tcp_init_remote_service_request(), tcp_buffer);

    nexus_debug_printf(2, ("tcp_init_remote_service_request(): to: (%s/%d) (%lu,%lu)  %s-%d\n",((tcp_proto_t *)(gp->proto))->host, ((tcp_proto_t *)(gp->proto))->port, gp->context, gp->address, handler_name, handler_id));

#ifndef NEXUS_FORCE_XDR
    if (gp->class_type == NEXUS_CLASS_TYPE)
    {
	tcp_buffer->funcs = &tcp_buffer_funcs;
	tcp_buffer->use_xdr = NEXUS_FALSE;
    }
    else
#endif
    {

	tcp_buffer->funcs = &tcp_xdr_buffer_funcs;
	tcp_buffer->use_xdr = NEXUS_TRUE;
	/* 
	 * set ops to NULL to show that this xdr stream hasn't been
	 * used and therefore shouldn't be destroyed later
	 */
	tcp_buffer->xdrs.x_ops = NULL;
    }
    tcp_buffer->next = (tcp_buffer_t *) NULL;
    tcp_buffer->buffer_type = SEND_BUFFER;
    tcp_buffer->storage = (char *) NULL;
    tcp_buffer->u.send.gp = gp;
    tcp_buffer->u.send.handler_id = handler_id;
    tcp_buffer->u.send.handler_name = handler_name;
    
    tcp_buffer->u.send.handler_name_length = strlen(handler_name)+1;
    {
	int i;

	i=tcp_buffer->u.send.handler_name_length % 4;
	if (i) tcp_buffer->u.send.handler_name_length += 4-i;
    }

    tcp_buffer->u.send.size = 0;
    tcp_buffer->u.send.n_elements = -1;
    tcp_buffer->u.send.base_pointer = (char *) NULL;
    tcp_buffer->u.send.current_pointer = (char *) NULL;
    tcp_buffer->u.send.header_pointer = (char *) NULL;

    NexusAssert2((tcp_buffer->u.send.handler_name_length < NEXUS_MAX_HANDLER_NAME_LENGTH),
		 ("tcp_init_remote_service_request(): Handler name exceeds maximum length (%d): %s\n", NEXUS_MAX_HANDLER_NAME_LENGTH, handler_name));
    
    header_size = TCP_MSG_HEADER_SIZE + tcp_buffer->u.send.handler_name_length;
#if ROUND_MALLOC > 1
    header_mod = header_size % ROUND_MALLOC;
    if (header_mod != 0)
    {
	header_size += (ROUND_MALLOC - header_mod);
    }
#endif    
    tcp_buffer->u.send.header_size = header_size;

    *buffer = (nexus_buffer_t) tcp_buffer;
} /* tcp_init_remote_service_request() */


/*
 * tcp_send_remote_service_request()
 *
 * Generate a remote service request message to the node and context
 * saved in the 'nexus_buffer'.
 */
static void tcp_send_remote_service_request(nexus_buffer_t *nexus_buffer)
{
    tcp_buffer_t *buf;
    tcp_proto_t *proto;
    unsigned long tmp_longs[TCP_MSG_HEADER_INTS];
    int msg_size, total_size, n_left, n_sent, write_len;
    char *send_ptr;
    int fd;
    int write_errno;
#ifdef NEXUS_THREAD_SAFE_SELECT
    int i_am_tcp_handler_thread = -1;
#endif /* NEXUS_THREAD_SAFE_SELECT */
#ifdef NEXUS_PROFILE
    int node_id, context_id;
#endif    

    NexusBufferMagicCheck(tcp_send_remote_service_request, nexus_buffer);

    buf = (tcp_buffer_t *) *nexus_buffer;
    NexusAssert2((buf->buffer_type == SEND_BUFFER),
		 ("tcp_send_remote_service_request(): Internal error: Expected a send buffer\n"));
    
    NEXUS_INTERROGATE((buf->u.send.gp), _NX_GLOBAL_POINTER_T, "tcp_init_remote_service_request");

    proto = (tcp_proto_t *) buf->u.send.gp->proto;
    NexusAssert2((proto->type == NEXUS_PROTO_TYPE_TCP),
		 ("tcp_send_remote_service_request(): Internal error: proto_type is not NEXUS_PROTO_TYPE_TCP\n"));

    nexus_debug_printf(2, ("tcp_send_remote_service_request(): invoked with buffer:%x\n", nexus_buffer));

    if (buf->storage == (char *) NULL)
    {
	/*
	 * This is a zero length message.
	 * Neither tcp_set_buffer_size() nor tcp_check_buffer_size()
	 * was called for this buffer.
	 * So use a dummy buffer for the header information.
	 */
	msg_size = 0;
	NexusMalloc(tcp_send_remote_service_request(),
		    buf->storage,
		    char *,
		    buf->u.send.header_size);
	SetBufferPointers(buf);
    }
    else
    {
	if(buf->use_xdr) msg_size = xdr_getpos(&(buf->xdrs));
	else
	msg_size = buf->u.send.current_pointer - buf->u.send.base_pointer;
    }
	
    /*
     * Setup message header:
     *
     * Cray note: 8 bytes are used for unsigned longs instead of 4.
     * So the following code will likely have to change for the Cray to
     * interoperate with other (32 bit) machines.
     *
     * After the header comes the handler name string, _not_ NULL terminated.
     */
    *(buf->u.send.header_pointer) = OPEN_FLAG;
    tmp_longs[0] = NX_HTONL(msg_size);
    tmp_longs[1] = NX_HTONL(buf->use_xdr);
    tmp_longs[2] = NX_HTONL(buf->u.send.gp->context);
    tmp_longs[3] = NX_HTONL(buf->u.send.gp->address);
    tmp_longs[4] = NX_HTONL((unsigned long) (buf->u.send.handler_id));
    tmp_longs[5] = NX_HTONL((unsigned long) (buf->u.send.handler_name_length));
#ifdef NEXUS_PROFILE
    _nx_node_id(&node_id);
    _nx_context_id(&context_id);
    tmp_longs[6] = NX_HTONL((long) node_id);
    tmp_longs[7] = NX_HTONL((long) context_id);
#endif    
    memcpy((buf->u.send.header_pointer + 1), tmp_longs,
	   TCP_MSG_HEADER_INTS * sizeof(unsigned long));

    /*
     * Note: the performance of this is probably crappy and should be
     * changed but right now it's a test to make XDR work right.
     * i.e. not crash because of alignment errors
     * so the transmitted handler name length should be an even word size
     * but we only want to copy in strlen(handler_name) characters
     * so that we don't read past the end of the user string
     */

    memset(buf->u.send.name_pointer, 0, buf->u.send.handler_name_length);
    memcpy(buf->u.send.name_pointer, buf->u.send.handler_name,
	   strlen(buf->u.send.handler_name));

    
    total_size = (msg_size + TCP_MSG_HEADER_SIZE
		  + buf->u.send.handler_name_length);

#ifdef NEXUS_PROFILE
    _nx_pablo_log_remote_service_request_send(buf->u.send.gp->node_id,
					      buf->u.send.gp->context_id,
					      buf->u.send.handler_name,
					      buf->u.send.handler_id,
					      msg_size);
#endif
	
    tcp_enter();

#ifndef NEXUS_LITE
    while (proto->send_in_progress)
    {
	/*
	 * There is already another thread sending on this proto.
	 * That thread probably would have blocked on a write,
	 * so it yielded in the hope that some else could do
	 * useful work in the mean time.  But since this thread's work
	 * is on the same proto, it cannot do anything.
	 */
	proto->sends_waiting++;
	nexus_cond_wait(&(proto->sends_waiting_cond), &tcp_mutex);
	proto->sends_waiting--;
    }
#endif /* NEXUS_LITE */

    proto->send_in_progress = NEXUS_TRUE;

    /*
     * Open up the proto if it is not already open.
     */
    if (proto->fd < 0)
    {
	open_proto(proto); /* May cause messages to be enqueued */
    }
    fd = proto->fd;
    NexusAssert2((fd >= 0),
		 ("tcp_send_remote_servie_request(): Internal error: Failed to open proto\n"));
    
    /*
     * Write 'total_size' bytes, starting at 'buf->u.send.header_pointer',
     * to the file discriptor of the proto.
     * Do not return until all bytes have been
     * written.  It is assumed that the thing being written is a
     * complete, atomic unit -- a message.
     *
     * If write() would block, then back off and read before
     * attempting to write again.
     *
     * It is possible that the process on the other end of this fd could
     * close this fd in the middle of writing.  (It is caching fds, and
     * needs some free ones.)  In this case, we need to reopen the
     * connection and restart the write from the beginning.
     */

    send_ptr = buf->u.send.header_pointer;
    n_left = total_size;
    
    while (n_left > 0)
    {
#if MAX_WRITE_PACKET > 0
	write_len = n_left > MAX_WRITE_PACKET ? MAX_WRITE_PACKET : n_left;
#else
	write_len = n_left;
#endif

	n_sent = write(fd, send_ptr, write_len);
	write_errno = errno;
	
	/*
	 * n_sent: is > 0 on success -- number of bytes written
	 *         is < 0 on error -- need to check errno
	 *         is 0 (SysV) or (-1 && errno==EWOULDBLOCK) (BSD)
	 *             if the write would block without writing anything
	 */

	
#ifdef NEXUS_DEBUG
	if (NexusDebug(3))
	{
            struct sockaddr_in local, remote;
            int len;

            len = sizeof(local);
            getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
            getpeername(proto->fd, (struct sockaddr *) &remote, &len);
	    
	    nexus_printf("tcp_send_remote_service_request(): Write attempt of %d bytes on fd=%d n_left=%d local=%d remote=%d returns %d\n",
			 write_len,
			 fd,
			 n_left,
			 ntohs(local.sin_port),
			 ntohs(remote.sin_port),
			 n_sent);
	    if (n_sent > 0)
	    {
		int i;
		char *buf = tmpbuf1, *tbuf = tmpbuf2;
		buf[0] = '\0';
		for (i = 0; /* i < 200 && */ i < n_sent; i++)
		{
		    sprintf(tbuf, "%u ", (unsigned int) send_ptr[i]);
		    strcat(buf, tbuf);
		}

		nexus_printf("Beginning of message: %s\n", buf);
	    }
	}
#endif
	
	if (n_sent > 0)
	{
	    n_left -= n_sent;
	    send_ptr += n_sent;
	}
	else if ((n_sent == 0) || (n_sent < 0 && write_errno == EWOULDBLOCK))
	{
	    /*
	     * We couldn't write (socket buffer is full).
	     *
	     * To avoid deadlock, we must try a read pass
	     * on all fds before retrying the write.
	     * (If two processes are sending big messages to
	     * each other simultaneously, they must each write some,
	     * then read some, then write some, then read some, etc.,
	     * so the the messages can slide past each other.)
	     */
#ifdef NEXUS_THREAD_SAFE_SELECT
	    if (using_handler_thread)
	    {
		if (i_am_tcp_handler_thread < 0)
		{
		    _nx_i_am_tcp_handler_thread(&i_am_tcp_handler_thread);
		    NexusAssert2((i_am_tcp_handler_thread >= 0),
				 ("tcp_send_remote_service_request(): Internal error: _nx_i_am_tcp_handler_thread() failed\n"));
		}
		if (i_am_tcp_handler_thread)
		{
		    /*
		     * This send must be from a handler.
		     * So just receive messages and enqueue them.
		     * We'll handle them later when the handler that
		     * called this send returns.
		     */
		    select_and_read(POLL_ONE_SELECT, ENQUEUE_MESSAGES);
		}
		else
		{
		    /*
		     * Else if I am not the handler thread, then yield.
		     * The handler thread will wake up automatically to
		     * handle messages if something arrives, but we must
		     * release the tcp_mutex to allow it to do its thing.
		     * The yield will also let other threads get useful
		     * work done.
		     *
		     * Note: Since a handler thread is only used with
		     * preemptive threads, yielding here should not cause
		     * a problem with starving this thread.
		     */
		    tcp_exit();
		    nexus_thread_yield();
		    tcp_enter();
		}
	    }
	    else
#endif /* NEXUS_THREAD_SAFE_SELECT */
	    {
		/*
		 * Messages enqueued here will get handled below by nexus_poll.
		 */
		select_and_read(POLL_ONE_SELECT, ENQUEUE_MESSAGES);
	    }
	}
	else /* n_sent < 0 */
	{
	    if (write_errno == EINTR)
	    {
		/* Do nothing.  Just try again. */
	    }
	    else if (write_errno == EBADF)
	    {
		/*
		 * The fd was closed unexpectedly.
		 * So call check_proto_for_close() to see what's going on.
		 * The choices are:
		 *   1) The process at the other end died (or
		 *      erroneously closed the fd).  In this case
		 *      check_proto_for_close() will fatal error out.
		 *   2) It closed the fd, so as to make room
		 *      to open another fd.  In this case we need
		 *      to reopen the fd and restart the write
		 *	from the beginning of the buffer.
		 */
		nexus_debug_printf(2, ("tcp_send_remote_service_request(): checking for proto close\n"));
		check_proto_for_close(proto);
		NexusAssert2((proto->fd < 0),
			     ("tcp_send_remote_service_request(): Internal error: expected fd<0\n"));
		open_proto(proto); /* May cause messages to be enqueued */
		fd = proto->fd;
		send_ptr = buf->u.send.header_pointer;
		n_left = total_size;
	    }
	    else
	    {
		tcp_fatal("tcp_send_remote_service_request(): Write failed: %s\n",
			  _nx_md_system_error_string());
	    }
	}
    }

    proto->send_in_progress = NEXUS_FALSE;
    
#ifndef NEXUS_LITE
    if (proto->sends_waiting > 0)
    {
	nexus_cond_signal(&(proto->sends_waiting_cond));
    }
#endif /* NEXUS_LITE */
    
    tcp_exit();
    if(buf->use_xdr && buf->xdrs.x_ops)
    {
	xdr_destroy(&(buf->xdrs));
    }
    FreeTcpBuffer(buf);

    nexus_poll();

} /* tcp_send_remote_service_request() */


/*
 * tcp_send_urgent_remote_service_request()
 */
static void tcp_send_urgent_remote_service_request(nexus_buffer_t *buffer)
{
    tcp_send_remote_service_request(buffer);
} /* tcp_send_urgent_remote_service_request() */


/*
 * tcp_destroy_proto()
 *
 * Decrement the reference count for this proto.  If it goes to 0
 * then close the fd used by this proto.
 */
static void tcp_destroy_proto(nexus_proto_t *nexus_proto)
{
    tcp_proto_t *proto = (tcp_proto_t *) nexus_proto;

    tcp_enter();
    
#ifdef NEXUS_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;

	if (proto->fd >= 0)
	{
            len = sizeof(local);
	    getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
	    getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	    nexus_printf("tcp_destroy(): proto=%x %s/%hu fd=%d local=%hu remote=%hu refcount=%d\n",
			 proto,
			 proto->host, proto->port,
			 proto->fd,
			 htons(local.sin_port),
			 htons(remote.sin_port),
			 proto->reference_count);
	}
	else
	{
	    nexus_printf("tcp_destroy(): proto=%x %s/%hu fd=%d refcount=%d\n",
			 proto,
			 proto->host, proto->port,
			 proto->fd,
			 proto->reference_count);
	}
    }
#endif
    
    proto->reference_count--;
    
    NexusAssert2((proto->reference_count >= 0),
		 ("tcp_destroy(): Internal error: Reference count < 0\n"));
	       
    if (proto->reference_count == 0 && proto->fd >= 0)
    {
#ifndef NEXUS_LITE
	NexusAssert2((proto->sends_waiting == 0),
		     ("tcp_destroy(): Internal error: sends_waiting != 0\n"));
#endif /* NEXUS_LITE */
	if (!close_proto_with_message(proto))
	{
	    /*
	     * The socket buffer for this fd is apparently full,
	     * so we cannot close this fd.
	     *
	     * So we need to defer the close.  This is done by
	     * incrementing n_deferred_closes.  At some later point,
	     * we can quickly check for the presence of deferred
	     * closes by checking this variable, and if there are
	     * then search the fd_to_proto table for protos with
	     * a reference_count of 0 and an open fd.
	     */
	    n_deferred_closes++;
	}
    }

    tcp_exit();
    
} /* tcp_destroy_proto() */


/*
 * tcp_get_my_mi_proto()
 *
 * Return the machine independent tcp protocol information
 * for this protocol.
 *
 * The tcp proto requires one integer (the tcp port) and one
 * string (the hostname) to encode its information.
 */
static nexus_mi_proto_t *tcp_get_my_mi_proto()
{
    nexus_mi_proto_t *mi_proto;
    int host_length;

    host_length = strlen(tcp_local_host);
    NexusMalloc(tcp_get_my_mi_proto(), mi_proto, nexus_mi_proto_t *,
		(sizeof(nexus_mi_proto_t)	/* base */
		 + sizeof(int)			/* space for ints */
		 + sizeof(int)			/* space for string_lengths */
		 + sizeof(char *)		/* space for string pointers */
		 + host_length + 1		/* space for strings */
		 ));
    mi_proto->proto_type = NEXUS_PROTO_TYPE_TCP;
    mi_proto->n_ints = 1;
    mi_proto->n_strings = 1;
    mi_proto->ints = (int *) (((char *) mi_proto)
			      + sizeof(nexus_mi_proto_t));
    mi_proto->ints[0] = (int) tcp_local_port;
    mi_proto->string_lengths = (int *) (((char *) mi_proto)
					+ sizeof(nexus_mi_proto_t)
					+ sizeof(int));
    mi_proto->string_lengths[0] = host_length;
    mi_proto->strings = (char **) (((char *) mi_proto)
				   + sizeof(nexus_mi_proto_t)
				   + sizeof(int)
				   + sizeof(int));
    mi_proto->strings[0] = (char *) (((char *) mi_proto)
				     + sizeof(nexus_mi_proto_t)
				     + sizeof(int)
				     + sizeof(int)
				     + sizeof(char *));
    strcpy(mi_proto->strings[0], tcp_local_host);

    return (mi_proto);
} /* tcp_get_my_mi_proto() */


/*
 * tcp_construct_from_mi_proto()
 *
 * The passed machine independent protocol, 'mi_proto', should
 * be a tcp protocol that I can use to connect to that node:
 *  - If it is not a tcp protocol, then fatal out.
 *  - If it is a tcp protocol:
 *	- If I cannot use this protocol to attach to the node, then
 *		return NEXUS_FALSE.  (This option is useful if two nodes
 *		both speak a particular protocol, but they cannot
 *		talk to each other via that protocol.  For example,
 *		on two MPP, the nodes within a single MPP can
 *		talk to each other via the native messaging protocol,
 *		but cannot talk to the nodes on the other MPP
 *		using that native protocol.)
 *	- If this tcp protocol points to myself, then set
 *		*proto=_nx_local_proto, and return NEXUS_TRUE.
 *	- Otherwise, construct a tcp protocol object for this mi_proto
 *		and put it in *proto.  Then return NEXUS_TRUE.  
 */
static nexus_bool_t tcp_construct_from_mi_proto(nexus_proto_t **proto,
					  nexus_mi_proto_t *mi_proto)
{
    char *host;
    u_short port;
    
    NexusAssert2((mi_proto->proto_type == NEXUS_PROTO_TYPE_TCP),
		 ("tcp_construct_from_mi_proto(): Internal error: Was given a non-tcp mi_proto\n"));

    /*
     * Test to see if this proto points to myself.
     * If it does, then return the _nx_local_proto.
     */
    host = mi_proto->strings[0];
    port = (short) mi_proto->ints[0];
    if (   (port == tcp_local_port)
	&& (strcmp(host, tcp_local_host) == 0) )
    {
	*proto = _nx_local_proto;
    }
    else
    {
	tcp_enter();
	*proto = (nexus_proto_t *) construct_proto(host, port, -1);
	tcp_exit();
    }
    return (NEXUS_TRUE);
} /* tcp_construct_from_mi_proto() */


#ifndef NEXUS_LITE
/*
 * tcp_get_creator_proto_params()
 *
 * Encapsulate enough protocol information into a retained command
 * line argument so that the construct_creator_proto() routine
 * can reconstruct a proto to connect back to me.
 */
static void tcp_get_creator_proto_params(char *buf,
					 int buf_size,
					 char *node_name,
					 int node_number)
{
    char tmp_buf[MAXHOSTNAMELEN + 64];
    
    nexus_stdio_lock();
    sprintf(tmp_buf, "-nf_nx_tcp \"%s %d %hu\"",
            tcp_local_host,
	    tcp_local_pid,
	    tcp_local_port);
    nexus_stdio_unlock();
    if (strlen(tmp_buf) > (size_t)buf_size)
    {
	nexus_fatal("tcp_get_creator_proto_params(): Internal error: Buffer not big enough for arguments\n");
    }
    strcpy(buf, tmp_buf);
} /* tcp_get_creator_proto_params() */


/*
 * tcp_construct_creator_proto()
 *
 * Use the stored command line arguments to construct a proto
 * to my creator.
 */
static void tcp_construct_creator_proto(nexus_proto_t **proto)
{
    char *creator_string;
    char host[MAXHOSTNAMELEN];
    int pid;
    int port;
    int rc;
    int arg_num;
    int *argc;
    char ***argv;

    nexus_get_retained_arguments(&argc, &argv);
    
    /* Get the -nf_nx_tcp argument and parse it */
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_tcp", 2)) >= 0)
    {
	creator_string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	nexus_fatal("tcp_construct_creator_proto(): Expected a -nf_nx_tcp argument\n");
    }
    nexus_stdio_lock();
    rc = sscanf(creator_string, "%s %d %d", host, &pid, &port);
    nexus_stdio_unlock();
    if (rc != 3)
    {
	nexus_fatal("tcp_construct_creator_proto(): Invalid -nf_nx_tcp argument\n");
    }
    
    tcp_enter();
    *proto = (nexus_proto_t *) construct_proto(host, port, -1);
    tcp_exit();
} /* tcp_construct_creator_proto() */
#endif /* NEXUS_LITE */


/*
 * tcp_compare_protos()
 */
static int tcp_compare_protos(nexus_proto_t *proto1,
			      nexus_proto_t *proto2)
{
    tcp_proto_t *p1 = (tcp_proto_t *) proto1;
    tcp_proto_t *p2 = (tcp_proto_t *) proto2;
    
    if (   (p1->port == p2->port)
	&& (strcmp(p1->host, p2->host) == 0) )
    {
	return (NEXUS_TRUE);
    }
    else
    {
	return (NEXUS_FALSE);
    }
} /* tcp_compare_protos() */


/*
 * construct_proto()
 *
 * Construct a tcp_proto_t for the given host and port. Look up in the
 * proto table to see if one already exists. If it does, bump its reference
 * count and return that one. Otherwise create one, insert into the
 * table with a reference count of 1 and return it.
 */
static tcp_proto_t *construct_proto(char *host, u_short port, int fd)
{
    tcp_proto_t *proto;

    proto = proto_table_lookup(host, port);
    nexus_debug_printf(3, ("construct_proto(): Table lookup returns proto=%x\n", proto));
    if (proto == (tcp_proto_t *) NULL)
    {
	NexusMalloc(construct_proto(), proto, tcp_proto_t *,
		    sizeof(tcp_proto_t));

	proto->type = NEXUS_PROTO_TYPE_TCP;
	proto->funcs = &tcp_proto_funcs;
	proto->host = _nx_copy_string(host);
	proto->port = port;
	proto->fd = fd;
	proto->reference_count = 1;
	proto->send_in_progress = NEXUS_FALSE;
#ifndef NEXUS_LITE
	proto->sends_waiting = 0;
	nexus_cond_init(&(proto->sends_waiting_cond),
			(nexus_condattr_t *) NULL);
#endif /* NEXUS_LITE */
	
	proto_table_insert(proto);
    }
    else
    {
	if (fd >= 0 && proto->fd >= 0 && fd != proto->fd)
	{
	    tcp_fatal("construct_proto(): FD mismatch between proto table entry and new entry (%d != %d)\n",
		      proto->fd, fd);
	}

	if (proto->reference_count == 0 && proto->fd >= 0)
	{
	    /*
	     * We have a deferred closed proto which we need
	     * again.  So take it out of the deferred closed
	     * count...
	     */
	    n_deferred_closes--;
	}
	
	proto->reference_count++;
    }
	
    return (proto);
} /* construct_proto() */


#ifdef DONT_INCLUDE
/*
 * free_proto()
 *
 * Free the passed 'proto'.
 */
static void free_proto(tcp_proto_t *proto)
{
    if (proto->host != (char *) NULL)
    {
	NexusFree(proto->host);
    }
    nexus_cond_destroy(&(proto->sends_waiting_cond));
    NexusFree(proto);
} /* free_proto() */
#endif /* DONT_INCLUDE */


/*
 * open_proto()
 *
 * Open the connection for the passed proto.  If it is already
 * connected, then just return.
 *
 * Note_enqueue: This routine could cause messages to be enqueued.
 */
static void open_proto(tcp_proto_t *proto)
{
    if (proto->fd >= 0)
    {
	nexus_debug_printf(2, ("open_proto(): initial return fd=%d\n", proto->fd));
	return;
    }
    make_room_for_new_fd();
    proto->fd = do_connect(proto->host, proto->port);
    nexus_debug_printf(2, ("open_proto(): do_connect(%s/%hu) returns %d\n",
			   proto->host, proto->port, proto->fd));
    NexusAssert2((proto->fd >= 0),
		 ("open_proto(): Failed to make connection\n"));
    add_proto_with_fd(proto, proto->fd);

} /* open_proto() */


/*
 * close_proto()
 *
 * Close a proto:
 *   1) Remove the proto from the fd_to_proto table
 *   2) Close the fd
 *   3) Modify proto data structure
 */
static void close_proto(tcp_proto_t *proto)
{
#ifdef NEXUS_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(proto->fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	nexus_printf("close_proto(): closing proto %x %s/%hu fd=%d local=%hu remote=%hu\n",
		     proto,
		     proto->host, proto->port,
		     proto->fd,
		     htons(local.sin_port),
		     htons(remote.sin_port));
    }
#endif
    
    remove_proto_with_fd(proto, proto->fd);
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    if (blocking_select_in_progress)
    {
	FD_SET(proto->fd, &close_fd_set);
	fd_needs_closing = NEXUS_TRUE;
    }
    else
#endif /* NEXUS_THREAD_SAFE_SELECT */
    {
	close(proto->fd);
    }
    
    proto->fd = -1;
} /* close_proto() */


/*
 * construct_incoming()
 *
 * Construct a tcp_incoming_t for the given file descriptor, 'fd'.
 */
static tcp_incoming_t *construct_incoming(int fd)
{
    tcp_incoming_t *incoming;

    if (incoming_free_list == (tcp_incoming_t *) NULL)
    {
	NexusMalloc(construct_incoming(), incoming, tcp_incoming_t *,
		    sizeof(tcp_incoming_t));
    }
    else
    {
	incoming = incoming_free_list;
	incoming_free_list = incoming_free_list->next_free;
    }

    incoming->type = INCOMING_TYPE;
    incoming->fd = fd;
    incoming->next_free = (tcp_incoming_t *) NULL;

    ResetIncoming(incoming);

    nexus_debug_printf(2, ("construct_incoming(): state: INCOMING_RECV_HEADER size: %d\n", TCP_MSG_HEADER_SIZE));

    return (incoming);
} /* construct_incoming() */


/*
 * close_incoming()
 *
 * Close an incoming connection:
 *   1) Remove the tcp_incoming_t from the fd_to_proto table
 *   2) Close the fd
 *   3) Put the tcp_incoming_t back on the free list.
 */
static void close_incoming(tcp_incoming_t *incoming)
{
#ifdef NEXUS_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(incoming->fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(incoming->fd, (struct sockaddr *) &remote, &len);

	nexus_printf("close_incoming(): closing %x fd=%d local=%hu remote=%hu\n",
		    incoming,
		    incoming->fd,
                    htons(local.sin_port),
                    htons(remote.sin_port));
    }
#endif
    remove_proto_with_fd((void *) incoming, incoming->fd);
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    if (blocking_select_in_progress)
    {
	FD_SET(incoming->fd, &close_fd_set);
	fd_needs_closing = NEXUS_TRUE;
    }
    else
#endif /* NEXUS_THREAD_SAFE_SELECT */
    {
	close(incoming->fd);
    }

    incoming->recv_state = INCOMING_RECV_CLOSED;
    incoming->next_free = incoming_free_list;
    incoming_free_list = incoming;
} /* close_incoming() */


/*
 * close_incoming_with_message()
 *
 * Send a normal close message down the file descriptor for
 * the given incoming connection.  If that works, then close it.
 *
 * If 'read_messages' is NEXUS_TRUE, then read any pending messages between
 * sending the close message and actually closing the fd.
 *
 * Check for a failed write due to a closed fd, and handle it
 * accordingly.  And check for any other errors.
 *
 * Note_enqueue: This routine could cause messages to be enqueued.
 *
 * Return: NEXUS_TRUE on success (incoming was closed)
 *         NEXUS_FALSE if the write would have blocked (the socket buffer
 *		is full), and thus the incoming was not closed
 */
static nexus_bool_t close_incoming_with_message(tcp_incoming_t *incoming,
					  nexus_bool_t read_messages)
{
    int n_sent;
    nexus_bool_t rc;
    int fd;
    char close_flag = CLOSE_NORMAL_FLAG;
    int write_errno;

    fd = incoming->fd;
    
#ifdef NEXUS_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(fd, (struct sockaddr *) &remote, &len);

	nexus_printf("close_incoming_with_message(): sending on fd=%d (local=%hu remote=%hu)\n",
		    fd,
                    htons(local.sin_port),
                    htons(remote.sin_port));
    }
#endif

 retry:
    n_sent = write(fd, &close_flag, 1);
    write_errno = errno;

#ifdef NEXUS_DEBUG
    if (NexusDebug(3))
    {
	nexus_printf("close_incoming_with_message(): n_sent=%d err=%s\n",
		    n_sent,
		    n_sent < 0 ? _nx_md_system_error_string(write_errno) : "<none>");
    }
#endif
    
    if (n_sent > 0)
    {
	/*
	 * We successfully closed the incoming. So:
	 *   1) Shutdown the socket
	 *   2) Read any pending data on incoming
	 *   3) Close the proto
	 */
	if (read_messages)
	{
	    shutdown(fd, 2);
	    read_incoming(incoming, ENQUEUE_MESSAGES);
	}
	close_incoming(incoming);
	rc = NEXUS_TRUE;
    }
    else if ((n_sent == 0) || (n_sent < 0 && write_errno == EWOULDBLOCK))
    {
	rc = NEXUS_FALSE;
    }
    else
    {
	if (write_errno == EINTR)
	{
	    goto retry;
	}
	else if (write_errno == EBADF)
	{
	
	    /*
	     * The fd for this incoming was closed unexpectedly.
	     *
	     * So call read_incoming() to see what's going on. The choices are:
	     *   1) The process at the other end died (or
	     *      erroneously closed the fd).  In this case
	     *      read_incoming() will fatal error out.
	     *   2) The process at the other end closed
	     *	the fd, so as to make room to
	     *      open another fd.  In this case, we
	     *      just go on our merry way, since the
	     *      other end beat us to closing the fd.
	     */

	    read_incoming(incoming, ENQUEUE_MESSAGES);

	    /*
	     * read_incoming() should either close the proto
	     * cleanly, or it should fatal out.
	     */
	    NexusAssert2((fd_to_proto[fd] == (void *) NULL),
			 ("close_uincoming_with_message(): Internal error: Expected incoming to be closed\n")); 

	    rc = NEXUS_TRUE;
	}
	else
	{
	    tcp_fatal("close_incoming_with_message(): Write failed: %s\n",
		      _nx_md_system_error_string());
	}
    }
    return (rc);
} /* close_incoming_with_message() */

				       
/*
 * close_proto_with_message()
 *
 * Send a normal close message down the file descriptor for
 * the given proto.  If that works, then close the proto.
 *
 * Check for a failed write due to a closed fd, and handle it
 * accordingly.  And check for any other errors.
 *
 * Return: NEXUS_TRUE on success (proto was closed)
 *         NEXUS_FALSE if the write would have blocked (the socket buffer
 *		is full), and thus the proto was not closed
 */
static nexus_bool_t close_proto_with_message(tcp_proto_t *proto)
{
    int n_sent;
    nexus_bool_t rc;
    int fd;
    char close_flag = CLOSE_NORMAL_FLAG;
    int write_errno;

    fd = proto->fd;

#ifdef NEXUS_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(fd, (struct sockaddr *) &remote, &len);

	nexus_printf("close_proto_with_message(): sending on fd=%d (local=%hu remote=%hu)\n",
		    fd,
                    htons(local.sin_port),
                    htons(remote.sin_port));
    }
#endif

 retry:
    n_sent = write(fd, &close_flag, 1);
    write_errno = errno;

#ifdef NEXUS_DEBUG
    if (NexusDebug(3))
    {
	nexus_printf("close_proto_with_message(): n_sent=%d err=%s\n",
		    n_sent,
		    n_sent < 0 ? _nx_md_system_error_string(write_errno) : "<none>");
    }
#endif
    
    if (n_sent > 0)
    {
	close_proto(proto);
	rc = NEXUS_TRUE;
    }
    else if ((n_sent == 0) || (n_sent < 0 && write_errno == EWOULDBLOCK))
    {
	rc = NEXUS_FALSE;
    }
    else
    {
	if (write_errno == EINTR)
	{
	    goto retry;
	}
	else if (write_errno == EBADF)
	{
	
	    /*
	     * The fd for this proto was closed unexpectedly.
	     *
	     * So read one byte from the proto.
	     * It should be a CLOSE_NORMAL_FLAG.  If so, then close
	     * the fd and go on our merry way, since the other
	     * end beat us to closing the fd.  If not, then fatal out.
	     * check_proto_for_close() will do this for us.
	     */
	    nexus_debug_printf(3, ("close_proto_with_message(): Checking for proto close\n"));
	    check_proto_for_close(proto);

	    /*
	     * check_proto_for_close() should either close the proto
	     * cleanly, or it should fatal out.
	     */
	    NexusAssert2((fd_to_proto[fd] == (void *) NULL),
			 ("close_proto_with_message(): Internal error: Expected proto to be closed\n")); 
	    
	    rc = NEXUS_TRUE;
	}
	else
	{
	    tcp_fatal("close_proto_with_message(): Write failed: %s\n",
		      _nx_md_system_error_string());
	}
    }
    return (rc);
} /* close_proto_with_message() */

				       
/*
 * check_proto_for_close()
 *
 * Do a non-blocking read from the passed proto.  The options are:
 *	1) If we get a close flag (single byte == CLOSE_NORMAL_FLAG),
 *		then close the proto.
 *	2) If we get EOF, then fatal out (the process on the other end
 *		probably died)
 *	3) If we get an error (including EWOULDBLOCK), then fatal out.
 *
 * Or in other words, this routine will either fatal out, or it will
 * return after closing the proto.
 */
static void check_proto_for_close(tcp_proto_t *proto)
{
    char buf[1];
    int n_read;

 retry:
    n_read = read(proto->fd, buf, 1);

    /*
     * n_read: is > 0 if it successfully read some bytes
     *         is < 0 on error -- need to check errno
     *         is 0 on EOF
     */
    
    nexus_debug_printf(5, ("check_proto_for_close(): read returned:%d\n",n_read));
    if (n_read > 0)
      {
	if (buf[0] == CLOSE_NORMAL_FLAG)
	{
	    /*
	     * We got a normal close of a proto.  So close our proto.
	     */
	    close_proto(proto);
	}
	else if (buf[0] == CLOSE_ABNORMAL_FLAG)
	{
	    /*
	     * We got an abnormal close of a proto.  So silent fatal out.
	     */
	    close_proto(proto);
	    tcp_exit();
	    nexus_silent_fatal();
	}
	else if (buf[0] == CLOSE_SHUTDOWN_FLAG)
	{
	    close_proto(proto);
	    tcp_exit();
	    nexus_exit(0, NEXUS_TRUE);
	}
	else
	{
	    tcp_fatal("check_proto_for_close(): Internal error: Read unexpected data from a proto\n");
	}
    }
    else if (n_read == 0)
    {
	tcp_fatal("check_proto_for_close(): Got EOF on a proto.  Another process probably died unexpectedly.\n");
    }
    else /* n_read < 0 */
    {
	if (errno == EINTR)
	{
	    /*
	     * The read() got interrupted (for example, by a signal),
	     * so retry the read.
	     */
	    goto retry;
	}
	else
	{
	    tcp_fatal("check_proto_for_close(): Read failed (errno=%d): %s\n",
		      errno, _nx_md_system_error_string());
	}
    }
} /* check_proto_for_close() */


/*
 * construct_listener()
 */
static tcp_listener_t *construct_listener(int fd,
					  u_short port,
					  nexus_bool_t internal)
{
    tcp_listener_t *listener;
    NexusMalloc(construct_listener(),
		listener,
		tcp_listener_t *,
		sizeof(tcp_listener_t));
    listener->type = LISTENER_TYPE;
    listener->fd = fd;
    listener->port = port;
    listener->internal = internal;
    if (all_listeners == (tcp_listener_t *) NULL)
    {
	listener->next = (tcp_listener_t *) NULL;
	listener->prev = (tcp_listener_t *) NULL;
	all_listeners = listener;
    }
    else
    {
	listener->next = all_listeners;
	listener->prev = (tcp_listener_t *) NULL;
	all_listeners->prev = listener;
	all_listeners = listener;
    }
    return (listener);
} /* construct_listener() */


/*
 * close_listener()
 *
 * Close a listener:
 *   1) Remove listeners from the all_listeners list
 *   2) Remove the tcp_listener_t from the fd_to_proto table
 *   3) Close the fd
 *   4) Free the tcp_listener_t structure
 */
static void close_listener(tcp_listener_t *listener)
{
    if (   (listener->next == (tcp_listener_t *) NULL)
	&& (listener->prev == (tcp_listener_t *) NULL) )
    {
	all_listeners = (tcp_listener_t *) NULL;
    }
    else if (listener->next == (tcp_listener_t *) NULL)
    {
	listener->prev->next = (tcp_listener_t *) NULL;
    }
    else if (listener->prev == (tcp_listener_t *) NULL)
    {
	listener->next->prev = (tcp_listener_t *) NULL;
	all_listeners = listener->next;
    }
    else
    {
	listener->next->prev = listener->prev;
	listener->prev->next = listener->next;
    }
    remove_proto_with_fd((void *) listener, listener->fd);
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    if (blocking_select_in_progress)
    {
	FD_SET(listener->fd, &close_fd_set);
	fd_needs_closing = NEXUS_TRUE;
    }
    else
#endif /* NEXUS_THREAD_SAFE_SELECT */
    {
	close(listener->fd);
    }

    NexusFree(listener);
} /* close_listener() */


/*
 * add_proto_with_fd()
 */
static void add_proto_with_fd(void *proto, int fd)
{
    n_fds_in_use++;
    NexusAssert(n_fds_in_use <= max_fds_to_use);
    NexusAssert(fd_to_proto[fd] == (void *) NULL);
    NexusAssert(fd >= 0 && fd < fd_tablesize);
    fd_to_proto[fd] = proto;
    FD_SET(fd, &current_fd_set);
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    modified_fd_table = NEXUS_TRUE;
#endif
} /* add_proto_with_fd() */


/*
 * remove_proto_with_fd()
 */
static void remove_proto_with_fd(void *proto, int fd)
{
    n_fds_in_use--;
    NexusAssert(n_fds_in_use >= 0);
    NexusAssert(fd_to_proto[fd] != (void *) NULL);
    NexusAssert(fd >= 0 && fd < fd_tablesize);
    fd_to_proto[fd] = (void *) NULL;
    FD_CLR(fd, &current_fd_set);
    
#ifdef NEXUS_THREAD_SAFE_SELECT
    modified_fd_table = NEXUS_TRUE;
#endif
} /* remove_proto_with_fd() */


/*
 * make_room_for_new_fd()
 *
 * This is called immediately before opening a file descriptor
 * for a proto or incoming.  It makes sure that there are
 * enough file descriptors available (after subtracting off
 * the number reserved by the user) to open a new one.  If so,
 * then it does nothing.  If not, then it closes a proto or
 * incoming to make room for the new fd.
 *
 * Note_enqueue: This routine could cause messages to be enqueued.
 */
static void make_room_for_new_fd()
{
    int i;
    int n_tries;
    tcp_proto_t *proto;
    tcp_incoming_t *incoming;
#ifdef NEXUS_THREAD_SAFE_SELECT
    int i_am_tcp_handler_thread = -1;
#endif /* NEXUS_THREAD_SAFE_SELECT */
    
    if (n_deferred_closes > 0)
    {
	/*
	 * Some previous proto close attempts
	 * were deferred (because the close message couldn't
	 * be written to the fd).  Try these closes again...
	 */
	n_tries = n_deferred_closes;
	for (i = 0; i < fd_tablesize && n_tries > 0; i++)
	{
	    proto = (tcp_proto_t *) fd_to_proto[i];
	    if (   (proto != (tcp_proto_t *) NULL)
		&& (proto->type == NEXUS_PROTO_TYPE_TCP)
		&& (proto->reference_count == 0) )
	    {
		if (close_proto_with_message(proto))
		{
		    n_deferred_closes--;
		}
		n_tries--;
	    }
	}
    }
    
    while (n_fds_in_use >= max_fds_to_use)
    {
	/*
	 * We've run out of file descriptors, so we must close one.
	 *
	 * First, look for a proto (outgoing connection) that is
	 * not in the middle of sending something.  Close one
	 * if one can be found.
	 *
	 * Second, look for an incoming connection that is
	 * between message receives.  Close one if one can be found.
	 * If one cannot be found, then do a select_and_read()
	 * pass and keep retrying to find an incoming connection
	 * to free.  This should not loop infinitely, since
	 * if an incoming is in the middle of a message receipt,
	 * the rest of the message should be coming shortly.  Once
	 * that message receipt has completed, that incoming is now
	 * a candidate for closure.
	 */

	/*
	 * First, check for any outgoing protos that are not in
	 * the middle of sending a message.
	 */
	i = next_fd_to_close;
	do
	{
	    proto = (tcp_proto_t *) fd_to_proto[i];
	    if (   (proto != (tcp_proto_t *) NULL)
		&& (proto->type == NEXUS_PROTO_TYPE_TCP)
		&& (!proto->send_in_progress) )
	    {
		/*
		 * Found one, so try to close it.
		 *
		 * The following call may or may not actually close the
		 * incoming.  But if it does close it, then n_fds_in_use
		 * will be decremented, which is what this loop is using.
		 */
		close_proto_with_message(proto);
	    }
	    i = (i + 1) % fd_tablesize;
	} while (   (n_fds_in_use >= max_fds_to_use)
		 && (i != next_fd_to_close) );
	next_fd_to_close = i;

	if (n_fds_in_use < max_fds_to_use)
	    break;
	
	/*
	 * Second, check for any incoming connections that
	 * are not in the middle of a message receive.
	 */
	do
	{
	    incoming = (tcp_incoming_t *) fd_to_proto[i];
	    if (   (incoming != (tcp_incoming_t *) NULL)
		&& (incoming->type == INCOMING_TYPE)
		&& (incoming->recv_state == INCOMING_RECV_HEADER)
		&& (incoming->recv_n_left == TCP_MSG_HEADER_SIZE) )
	    {
		/*
		 * Found one, so try to close it.
		 *
		 * The following call may or may not actually close the
		 * incoming.  But if it does close it, then n_fds_in_use
		 * will be decremented, which is what this loop is using.
		 */
		close_incoming_with_message(incoming, NEXUS_TRUE);
	    }
	    i = (i + 1) % fd_tablesize;
	} while (   (n_fds_in_use >= max_fds_to_use)
		 && (i != next_fd_to_close) );
	next_fd_to_close = i;

	if (n_fds_in_use >= max_fds_to_use)
	{
	    /*
	     * We've tried all protos and all incomings, but
	     * couldn't find any (enough) suitable candidates
	     * to free some fds.
	     * So try to read some stuff and/or yield
	     * the thread before trying again.  This may allow
	     * a proto or incoming to become a suitable
	     * candidate for closing.
	     */
#ifdef NEXUS_THREAD_SAFE_SELECT
	    if (using_handler_thread)
	    {
		if (i_am_tcp_handler_thread < 0)
		{
		    _nx_i_am_tcp_handler_thread(&i_am_tcp_handler_thread);
		    NexusAssert2((i_am_tcp_handler_thread >= 0),
				 ("make_room_for_new_fd(): Internal error: _nx_i_am_tcp_handler_thread() failed\n"));
		}
		if (i_am_tcp_handler_thread)
		{
		    /*
		     * This is the handler thread.
		     * So just read some message and enqueue them.
		     * We'll handle them later when this returns.
		     */
		    select_and_read(POLL_UNTIL_SELECT_FAILS, ENQUEUE_MESSAGES);
		}
		else
		{
		    /*
		     * Else if I am not the handler thread, then yield.
		     * The handler thread will wake up automatically to
		     * handle messages if something arrives, but we must
		     * release the tcp_mutex to allow it to do its thing.
		     * The yield will also let other threads get useful
		     * work done.
		     */
		    tcp_exit();
		    nexus_thread_yield();
		    tcp_enter();
		}
	    }
	    else
#endif /* NEXUS_THREAD_SAFE_SELECT */
	    {
		/*
		 * So read some stuff which will be handled later.
		 */
		select_and_read(POLL_UNTIL_SELECT_FAILS, ENQUEUE_MESSAGES);
	    }
	}
    }

    NexusAssert2((n_fds_in_use < max_fds_to_use),
		 ("make_room_for_new_fd(): Internal error: n_fds_in_use(%d) >= max_fds_to_use(%d)\n", n_fds_in_use, max_fds_to_use) );
    
} /* make_room_for_new_fd() */


/*
 * select_and_read();
 *
 * Do a select on all valid file descriptors, and read from all
 * that have stuff ready to read.  Once a complete message is
 * read, invoke the handler with the message.
 *
 * If poll_type==POLL_ONE_SELECT, then only do one select and read
 * pass on the file desciptors.
 *
 * If poll_type==POLL_UNTIL_SELECT_FAILS, then do not return until
 * select returns 0.  This means that we may do several select
 * and read passes on the file descritors.
 *
 * If poll_type==POLL_BLOCKING_UNTIL_RECEIVE, then using a blocking select.
 * This will only be called when there is not a tcp handler thread,
 * and when all threads have blocked.  Actually, this can cause deadlock
 * if the idle thread gets accidentally scheduled, and then another thread
 * tries to call a tcp function -- it will block on the tcp_enter().
 * You should probably use POLL_ONE_TIMED_SELECT instead.
 *
 * If poll_type==POLL_ONE_TIMED_SELECT, then using a blocking select with a
 * timeout.  This will be called when there is not a tcp handler thread,
 * and when all threads have blocked.
 *
 * If poll_type==POLL_FROM_HANDLER_THREAD, then do not ever return.
 * This should be used when calling from a separate handler thread.
 *
 * If handle_or_enqueue==HANDLE_MESSAGES, then invoke the
 * message handler immediately upon receipt of a message.
 *
 * If handle_or_enqueue==ENQUEUE_MESSAGES, then enqueue
 * any messages that are received onto the message_queue, to
 * be handled later.
 *
 * Return: NEXUS_TRUE if a message is handled, otherwise NEXUS_FALSE
 */
static nexus_bool_t select_and_read(int poll_type, int handle_or_enqueue)
{
    fd_set select_fd_set;
    int select_n_ready, n_checked;
    struct timeval timeout;
    nexus_bool_t use_timeout;
    int fd;
    nexus_bool_t done = NEXUS_FALSE;
    nexus_bool_t message_handled = NEXUS_FALSE;
    int select_errno;

    /*
     * Handle any enqueued messages if need be.
     */
    TestAndHandleMessages();

    while (!done)
    {
	
#ifdef NEXUS_DEBUG
	if (NexusDebug(5))
	{
	    print_fd_set("Selecting on fd_set=<%s>\n", &current_fd_set);
	}
#endif

	while (1)
	{
	    select_fd_set = current_fd_set;
	    
	    if (   poll_type == POLL_ONE_SELECT
		|| poll_type == POLL_UNTIL_SELECT_FAILS)
	    {
		use_timeout = NEXUS_TRUE;
		timeout.tv_sec = timeout.tv_usec = 0L;
	    }
            else if (poll_type == POLL_BLOCKING_UNTIL_RECEIVE)
	    {
		use_timeout = NEXUS_FALSE;
	    }
            else if (poll_type == POLL_ONE_TIMED_SELECT)
	    {
		use_timeout = NEXUS_TRUE;
		timeout.tv_sec = 1L;
		timeout.tv_usec = 0L;
	    }
#ifdef NEXUS_THREAD_SAFE_SELECT
	    else if (poll_type == POLL_FROM_HANDLER_THREAD)
	    {
		/*
		 * This is the handler thread, and select() is
		 * a thread safe operation.  So allow other code
		 * to execute in tcp while in a blocking select.
		 * But use modified_fd_table as a flag
		 * to tell if the fd_table was modified
		 * while the select was blocked.
		 *
		 * Don't block indefinitely. We want to be able
		 * to check for shutdown periodically.
		 */
		blocking_select_in_progress = NEXUS_TRUE;
		modified_fd_table = NEXUS_FALSE;
		/*
		use_timeout = NEXUS_FALSE;
		*/
		use_timeout = NEXUS_TRUE;
		timeout.tv_sec = 1L;
		timeout.tv_usec = 0L;
		tcp_exit();
	    }
#endif /* NEXUS_THREAD_SAFE_SELECT */
	    
	    select_n_ready = select(fd_tablesize,
				    NEXUS_FD_SET_CAST &select_fd_set,
				    NEXUS_FD_SET_CAST NULL,
				    NEXUS_FD_SET_CAST NULL,
				    (use_timeout ? &timeout : NULL) );
	    select_errno = errno;
	    
            if (poll_type == POLL_BLOCKING_UNTIL_RECEIVE)
	    {
		poll_type = POLL_UNTIL_SELECT_FAILS;
	    }
#ifdef NEXUS_THREAD_SAFE_SELECT
	    else if (poll_type == POLL_FROM_HANDLER_THREAD)
	    {
		/*
		 * This is the handler thread, with a threadsafe select().
		 * Check to make sure the fd_table wasn't modified
		 * while the select was blocked.
		 */
		tcp_enter();
		blocking_select_in_progress = NEXUS_FALSE;
		if (fd_needs_closing)
		{
		    for (fd = 0; fd < NUM_FDS; fd++)
		    {
			if (FD_ISSET(fd, &close_fd_set))
			{
			    close(fd);
			}
		    }
		    fd_needs_closing = NEXUS_FALSE;
		    FD_ZERO(&close_fd_set);
		}
		if (tcp_done)
		    return (message_handled);
		if (modified_fd_table)
		    continue;
	    }
#endif /* NEXUS_THREAD_SAFE_SELECT */

	    if (select_n_ready >= 0)
	    {
		break;
	    }
	    else if (select_n_ready < 0 && select_errno != EINTR)
	    {
		/* JGG how do I fix this? JGG */
		/* JGG
		errno = select_errno;
		JGG */
		tcp_fatal("select_and_read(): select failed: %s\n",
			  _nx_md_system_error_string());
	    }
	}

#ifdef NEXUS_DEBUG
	if (NexusDebug(4))
	{
	    nexus_printf("Select returns %d\n", select_n_ready);
	    print_fd_set("Select returned fd_set=<%s>\n", &select_fd_set);
	}
#endif

	for (n_checked = 0, fd = 0;
	     n_checked < select_n_ready;
	     fd++)
	{
	    NexusAssert2(fd < fd_tablesize,
			 ("select_and_read(): Internal error: fd >= fd_tablesize\n"));
	    
	    if (FD_ISSET(fd, &select_fd_set))
	    {
#ifdef NEXUS_DEBUG
		if (NexusDebug(5.5))
		{
		    struct sockaddr_in addr;
		    int len = sizeof(addr);
		    if (getpeername(fd, (struct sockaddr *) &addr, &len) == 0)
		    {
			nexus_printf("select_and_read(): selected fd=%d, connection is from %s/%d\n",
				    fd, inet_ntoa(addr.sin_addr),
				    htons(addr.sin_port));
		    }
		}
#endif
		n_checked++;

#ifdef NEXUS_DEBUG		    
                if (((tcp_proto_t *)(fd_to_proto[fd]))
		    == (tcp_proto_t *) NULL)
		{
		    tcp_fatal("select_and_read(): Data is ready on fd %d, but there is no proto or incoming for it\n", fd);
		}
#endif /* NEXUS_DEBUG */			
		
		if (((tcp_proto_t *)(fd_to_proto[fd]))->type
		    == LISTENER_TYPE)
		{
		    /* Somebody wants to connect to me */
		    tcp_listener_t *listener
			= (tcp_listener_t *) fd_to_proto[fd];
		    accept_new_connection(listener);
		    TestAndHandleMessages();
		}
		else if (((tcp_proto_t *)(fd_to_proto[fd]))->type
			 == INCOMING_TYPE)
		{
		    read_incoming((tcp_incoming_t *) (fd_to_proto[fd]),
				  handle_or_enqueue);
		}
		else /* type == NEXUS_PROTO_TYPE_TCP */
		{
		    nexus_debug_printf(5, ("select_and_read(): Checking for proto close\n"));
		    check_proto_for_close((tcp_proto_t *)
					  (fd_to_proto[fd]));
		}
	    }
	}

	if (   (poll_type == POLL_ONE_SELECT)
	    || (poll_type == POLL_ONE_TIMED_SELECT)
	    || (   (poll_type == POLL_UNTIL_SELECT_FAILS)
		&& (select_n_ready == 0) )
	    )
	{
	    done = NEXUS_TRUE;
	}
    }

    nexus_debug_printf(5, ("select_and_read(): returning, message_handled=%d\n", message_handled));
    return (message_handled);
    
} /* select_and_read() */


/*
 * read_incoming()
 *
 * Read as much of a complete message as we can for this incoming connection.
 * The various stages of reading a complete message are:
 *   1) Non-blocking read of message header.
 *   2) Allocate a message buffer
 *   3) Non-blocking read of the handler name.
 *	If zero length message then skip to #6.
 *   4) Non-blocking read of message body.
 *   5) Enqueue or handle complete message.
 *   6) Reset incoming for reading a new message.
 */
static void read_incoming(tcp_incoming_t *incoming, int handle_or_enqueue)
{
    int n_read;
    int read_errno;
    
    if (incoming->recv_state == INCOMING_RECV_CLOSED)
    {
	return;
    }

    NexusAssert2((   (incoming->recv_state == INCOMING_RECV_HEADER)
		  || (incoming->recv_state == INCOMING_RECV_BODY) ),
		 ("read_incoming(): Internal error: Invalid incoming->recv_state = %d\n", incoming->recv_state) );

    do
    {
	n_read = read(incoming->fd,
		      incoming->recv_buf_current,
		      incoming->recv_n_left);
	read_errno = errno;

	/*
	 * n_read: is > 0 if it successfully read some bytes
	 *         is < 0 on error -- need to check errno
	 *         is 0 on EOF
	 */

#ifdef NEXUS_DEBUG
	if (NexusDebug(2))
	{
            struct sockaddr_in local, remote;
            int len;
	    
            len = sizeof(local);
            getsockname(incoming->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
            getpeername(incoming->fd, (struct sockaddr *) &remote, &len);

	    nexus_printf("read_incoming(): Read on fd=%d n_left=%d local=%d remote=%d returns %d\n",
			incoming->fd,
			incoming->recv_n_left,
			ntohs(local.sin_port),
			ntohs(remote.sin_port),
			n_read);
	    if (n_read > 0)
	    {
		int i;

		tmpbuf1[0] = '\0';
		for (i = 0; /* i < 200 && */ i < n_read; i++)
		{
		    sprintf(tmpbuf2, "%u ",
			    (unsigned int) incoming->recv_buf_current[i]);
		    strcat(tmpbuf1, tmpbuf2);
		}

		nexus_printf("read_incoming(): Beginning of message: %s\n",
			    tmpbuf1);
	      }
	}
#endif /* NEXUS_DEBUG */
	
	if (n_read > 0)
	{
	    NexusAssert2((n_read <= incoming->recv_n_left),
			 ("read_incoming(): Internal error: n_read(%d) <= incoming->recv_n_left(%d)\n",
			  n_read, incoming->recv_n_left) );

	    incoming->recv_n_left -= n_read;

	    /* Check for a closed stream */
	    if (   (n_read == 1)
		&& (incoming->recv_state == INCOMING_RECV_HEADER)
		&& (incoming->recv_n_left == TCP_MSG_HEADER_SIZE - 1)
		&& (incoming->header_buf[0] != OPEN_FLAG) )
	    {
		incoming->recv_n_left += n_read;
		n_read = 0;	/* exit loop */
		if (incoming->header_buf[0] == CLOSE_NORMAL_FLAG)
		{
		    /*
		     * We got a close message on this fd due to either:
		     *   1) the node at the other end exited normally
		     *   2) the node at the other end closed this
		     *      socket in order to open another one
		     * So close the incoming _without_ first sending
		     * the close message down the fd.  (i.e., Just
		     * close down my end of the fd that the other
		     * end already closed.)
		     */
		    close_incoming(incoming);
		}
		else if (incoming->header_buf[0] == CLOSE_ABNORMAL_FLAG)
		{
		    /*
		     * We got an abnormal close message on this fd.
		     * This means some other node fataled out,
		     * and sent this down the pipe just before dieing.
		     * So fatal out silently, since the node that
		     * generated this message will have complained
		     * already.  (No need for cascading fatal messages.)
		     */
		    close_incoming(incoming);
		    tcp_exit();
		    nexus_silent_fatal();
		}
		else if (incoming->header_buf[0] == CLOSE_SHUTDOWN_FLAG)
		{
		    /*
		     * We got an shutdown close message on this fd.
		     * This means some other node is shutting down
		     * and is telling us to do the same.
		     * So shutdown and propogate it to other nodes.
		     */
		    close_incoming(incoming);
		    tcp_exit();
		    nexus_exit(0, NEXUS_TRUE);
		}
		else
		{
		    tcp_fatal("read_incoming(): Internal error: Got an illegal first byte of message header: %d\n", (int) incoming->header_buf[0]);
		}
	    }
		    
	    else if (incoming->recv_n_left == 0)
	    {
		/* We have completed reading something... */

		if (incoming->recv_state == INCOMING_RECV_HEADER)
		{
		    /* We have completed reading the header. */
		    unsigned long tmp_longs[TCP_MSG_HEADER_INTS];
		    tcp_buffer_t *buf;

		    /* Get a buffer and initialize it */
		    nexus_debug_printf(3, ("read_incoming(): Allocating buffer for incoming message:%x\n", buf));
		    GetTcpBuffer(read_incoming(), buf);
/*		    buf->funcs = &tcp_buffer_funcs; */
		    buf->buffer_type = RECV_BUFFER;
		    buf->u.recv.stashed = NEXUS_FALSE;

		    /* Extract the message header information */
		    memcpy(tmp_longs, &(incoming->header_buf[1]),
			   TCP_MSG_HEADER_INTS * sizeof(unsigned long));
		    buf->u.recv.size = NX_NTOHL(tmp_longs[0]);
		    buf->use_xdr = NX_NTOHL(tmp_longs[1]);
		    buf->u.recv.context = NX_NTOHL(tmp_longs[2]);
		    buf->u.recv.address = NX_NTOHL(tmp_longs[3]);
		    buf->u.recv.handler_id = (int) NX_NTOHL(tmp_longs[4]);
		    buf->u.recv.handler_name_length = NX_NTOHL(tmp_longs[5]);
#ifdef NEXUS_PROFILE		    
		    buf->u.recv.node_id = (int) NX_NTOHL(tmp_longs[6]);
		    buf->u.recv.context_id = (int) NX_NTOHL(tmp_longs[7]);
#endif		    
		    if(buf->use_xdr) buf->funcs = &tcp_xdr_buffer_funcs;
		    else buf->funcs = &tcp_buffer_funcs;

		    nexus_debug_printf(3, ("read_incoming(): got size=%d context=%x address=%x handler_id=%ld handler_name_length=%ld\n",
					   buf->u.recv.size,
					   buf->u.recv.context,
					   buf->u.recv.address,
					   (unsigned long) buf->u.recv.handler_id,
					   buf->u.recv.handler_name_length) );

		    /*
		     * Allocate memory to hold the buffer.
		     */
		    buf->u.recv.size += buf->u.recv.handler_name_length;
		    NexusMalloc(read_incoming(),
				buf->storage,
				char *,
				buf->u.recv.size);
		    buf->u.recv.handler_name = buf->storage;
		    buf->u.recv.current_pointer
			= buf->storage + buf->u.recv.handler_name_length;
		    
		    /*
		     * Reset the incoming's recv_state
		     * to INCOMING_RECV_BODY
		     */
		    incoming->recv_state = INCOMING_RECV_BODY;
		    incoming->recv_buf = buf;
		    incoming->recv_buf_current = buf->storage;
		    incoming->recv_n_left = buf->u.recv.size;
		    nexus_debug_printf(3, ("read_incoming(): State now: INCOMING_RECV_BODY  expecting: %d bytes\n", incoming->recv_n_left));
		}
		else /* (incoming->recv_state == INCOMING_RECV_BODY) */
		{
		    tcp_buffer_t *buf = incoming->recv_buf;

		    /*
		     * Lookup the handler function.
		     */
		    nexus_debug_printf(3, ("read_incoming(): got call to handler: %s\n", buf->u.recv.handler_name));
		    _nx_lookup_handler(buf->u.recv.context,
				       buf->u.recv.handler_name,
				       buf->u.recv.handler_id,
				       &(buf->u.recv.handler_type),
				       &(buf->u.recv.handler_func) );
		    if (buf->u.recv.handler_type ==NEXUS_HANDLER_TYPE_THREADED)
		    {
			buf->u.recv.stashed = NEXUS_TRUE;
		    }
		    
		    /*
		     * Reset the incoming's recv_state
		     * to INCOMING_RECV_HEADER.
		     * This needs to be done before the
		     * call to _nx_handle_message(), so
		     * that everything is in a consistent
		     * state when we tcp_exit() for
		     * the message handle.
		     */
		    ResetIncoming(incoming);
		    if( buf->use_xdr )
		    {
			xdrmem_create(&(buf->xdrs),
				      buf->u.recv.current_pointer,
				      buf->u.recv.size, XDR_DECODE );
		    }
			
		    /* Enqueue or handle the message */
		    if (handle_or_enqueue == HANDLE_MESSAGES)
		    {
			handle_in_progress = NEXUS_TRUE;
			tcp_exit();
			_nx_handle_message(buf->u.recv.handler_name,
					   buf->u.recv.handler_id,
					   buf->u.recv.handler_type,
					   buf->u.recv.handler_func,
					   buf->u.recv.context,
					   buf->u.recv.address,
#ifdef NEXUS_PROFILE
					   buf->u.recv.node_id,
					   buf->u.recv.context_id,
					   buf->u.recv.size,
#endif
					   (void *) buf);
			tcp_enter();
			handle_in_progress = NEXUS_FALSE;
		    }
		    else
		    {
			EnqueueMessage(buf);
		    }

		    /* 
		     * We have received an entire message and
		     * either handled or enqueued it.
		     * So set n_read=0 to break out of the read loop.
		     * If there are any more messages coming in on 
		     * this fd, we will catch them on the next
		     * select() go-around.
		     */
		    n_read = 0;
		}
	    }
	    else
	    {
		/*
		 * Else we didn't get a close message, and we didn't
		 * finish reading anything.  So we're in the middle of
		 * a message header, handler_name, or body, so just
		 * bump the recv pointer and keep trying...
		 */
		incoming->recv_buf_current += n_read;
	    }
	}
	else if (n_read == 0)
	{
	    /*
	     * Got EOF on the read, so die...
	     *
	     * We should never get an EOF on read under normal
	     * circumstances.  We should always get (and handle) a
	     * close message (the leading byte of a message) before
	     * a fd is closed, so we can handle the closed fd
	     * before we ever get an EOF.  This allows us to
	     * detect abnormal termination of any process we
	     * are connected to.
	     */
	    int len;
	    struct sockaddr_in addr;
	    
	    len = sizeof(addr);
	    if (getpeername(incoming->fd, (struct sockaddr *) &addr, &len) == 0)
	    {
		tcp_fatal("read_incoming(): got EOF for fd %d connected to %s/%d\n",
			  incoming->fd,
			  inet_ntoa(addr.sin_addr),
			  htons(addr.sin_port));
	    }
	    else
	    {
		tcp_fatal("read_incoming(): got EOF for fd %d\n",
			  incoming->fd);
	    }
	}
	else /* n_read < 0 */
	{
	    /* Got an error on the read */
	    nexus_debug_printf(5, ("read_incoming(): got -1 on read. errno:%d\n",read_errno));
	    if (read_errno == EINTR)
	    {
		/*
		 * The read() got interrupted (for example, by a signal),
		 * so retry the read.
		 */
		nexus_debug_printf(5, ("read_incoming(): error:EINTR retrying\n"));
		continue;
	    }
	    else if (read_errno == EAGAIN || read_errno == EWOULDBLOCK)
	    {
		/*
		 * The read() would have blocked without reading anything.
		 * So quit trying and return.
		 */
		nexus_debug_printf(5, ("read_incoming(): error:EWOULDBLOCK or EAGAIN quitting\n"));
	    }
	    else
	    {
		/* JGG How do I fix this? JGG */
		tcp_fatal("read_incoming(): Read failed (errno=%d): %s\n",
			  read_errno, _nx_md_system_error_string());
	    }
	}

    } while (n_read > 0);
    
} /* read_incoming() */


/*
 * handle_enqueued_messages()
 *
 * Handle all messages that are enqueued.
 */
static void handle_enqueued_messages()
{
    tcp_buffer_t *buf;

    handle_in_progress = NEXUS_TRUE;
    while (MessagesEnqueued())
    {
	DequeueMessage(buf);
	tcp_exit();
	_nx_handle_message(buf->u.recv.handler_name,
			   buf->u.recv.handler_id,
			   buf->u.recv.handler_type,
			   buf->u.recv.handler_func,
			   buf->u.recv.context,
			   buf->u.recv.address,
#ifdef NEXUS_PROFILE
			   buf->u.recv.node_id,
			   buf->u.recv.context_id,
			   buf->u.recv.size,
#endif
			   (void *) buf);
	tcp_enter();
    }
    handle_in_progress = NEXUS_FALSE;

} /* handle_enqueued_messages() */


/*
 * accept_new_connection()
 *
 * Do an accept() on the passed 'fd', and then setup
 * a new tcp_incoming_t for this connection.
 *
 * Note_enqueue: This routine could cause messages to be enqueued.
 */
static void accept_new_connection(tcp_listener_t *listener)
{
    int new_fd;
    struct sockaddr_in from;
    int len;
    tcp_incoming_t *incoming;
    int pid;
    int rc;
#ifdef NEXUS_DEBUG
    struct hostent *hp;
#if defined(NEXUS_ARCH_SOLARIS)
    struct hostent hp2;
    char hp_tsdbuffer[500];
    int hp_errnop;
#endif
#if defined(NEXUS_ARCH_AIX_DCE_THREADS)
    struct hostent hp2;
    struct hostent_data hp_data;
#endif
#endif

    nexus_debug_printf(4,("accept_new_connection(): entering\n"));
	
    make_room_for_new_fd();
    
    len = sizeof(from);
    nexus_debug_printf(4,("accept_new_connection(): calling accept()\n"));
    new_fd = accept(listener->fd, (struct sockaddr *) &from, &len);
    nexus_debug_printf(4,("accept_new_connection(): returned from accept()\n"));
    if (new_fd < 0)
    {
	tcp_fatal("accept_new_connection(): accept failed: %s\n",
		  _nx_md_system_error_string());
    }

#ifdef NEXUS_DEBUG
    if (NexusDebug(3))
    {
#if defined(NEXUS_ARCH_SOLARIS)
#define NEXUS_ACCEPT_NEW_CONNECTION_GETHOSTBYADDR    
	hp = gethostbyaddr_r((char *) &from.sin_addr,
			     sizeof(from.sin_addr),
			     from.sin_family,
			     &hp2, hp_tsdbuffer, 500, &hp_errnop);
#endif
#if defined(NEXUS_ARCH_AIX_DCE_THREADS)
#define NEXUS_ACCEPT_NEW_CONNECTION_GETHOSTBYADDR    
	gethostbyaddr_r(( char *) &from.sin_addr,
			sizeof(from.sin_addr),
			from.sin_family,
			&hp2, &hp_data);
	hp = &hp2;
#endif
#ifndef NEXUS_ACCEPT_NEW_CONNECTION_GETHOSTBYADDR    
	hp = gethostbyaddr((char *) &from.sin_addr,
			   sizeof(from.sin_addr),
			   from.sin_family);
#endif    
	nexus_printf("Got incoming from %s/%hu on %d\n",
		     hp->h_name, ntohs(from.sin_port), new_fd);
    }
#endif

    /* Read the pid of the connecting process */
    nexus_debug_printf(4,("accept_new_connection(): reading pid\n"));
    rc = read(new_fd, &pid, 4);
    pid = NX_NTOHL(pid);
    nexus_debug_printf(4,("accept_new_connection(): read pid=%d\n", pid));

    config_socket(new_fd);
    
    if (listener->internal)
    {
	incoming = construct_incoming(new_fd);
#ifdef NEXUS_DEBUG
	incoming->connected_pid = pid;
#endif
	add_proto_with_fd(incoming, new_fd);
    }
    else
    {
	accept_attachment(listener, new_fd);
    }

} /* accept_new_connection() */


/*
 * config_socket()
 *
 * Configure the passed socket (a file descriptor) with:
 *	- no delay: Keep the TCP driver from delaying
 *			outgoing packets to coalescing
 *	- keepalive: Periodically send messages down the socket
 *			so as to be able to detect if the process
 *			on the other end of the socket has died
 *	- non-blocking: read() and write() operations will not block
 */
static void config_socket(int s)
{
    int one = 1;

#ifndef NEXUS_NO_TCP_NODELAY
    if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *) &one,
		   sizeof(one)) < 0)
    {
	nexus_warning("config_socket(): setsockopt TCP_NODELAY failed: %s\n",
		     _nx_md_system_error_string());
    }
#endif
    
#ifdef SO_KEEPALIVE
    if (setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *) &one,
		   sizeof(one)) < 0)
    {
	nexus_warning("config_socket(): setsockopt SO_KEEPALIVE failed: %s\n",
		      _nx_md_system_error_string());
    }
#endif

    set_nonblocking(s);
} /* config_socket() */


/*
 * set_nonblocking()
 *
 * Set the passed file descriptor, fd, to non-blocking.  (Calls
 * to read() and write() on that fd will not block.)
 */
static void set_nonblocking(int fd)
{
    int flags;
    
    if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
    {
	tcp_fatal("set_nonblocking(): fcntl F_GETFL 1: %s\n",
		  _nx_md_system_error_string());
    }

#if defined(NEXUS_ARCH_AIX) || defined(NEXUS_ARCH_HPUX)
    flags |= O_NONBLOCK;
#else
    flags |= FNDELAY;
#endif

    if (fcntl(fd, F_SETFL, flags) < 0)
    {
	tcp_fatal("set_nonblocking(): fcntl F_SETFL: %s\n",
		  _nx_md_system_error_string());
    }

#ifdef F_SETFD
    if (fcntl(fd, F_SETFD, 1) < 0)
    {
	tcp_fatal("set_nonblocking(): fcntl F_SETFD: %s\n",
		  _nx_md_system_error_string());
    }
#endif
} /* set_nonblocking() */


/*
 * do_connect()
 *
 * Return: The fd of the new connection.
 */
static int do_connect(char *host, u_short port)
{
    struct sockaddr_in his_addr, use_his_addr;
    int s;
    nexus_bool_t connect_succeeded;
    int connect_failures = 0;
    struct hostent *hp;
    long pid, write_rc;
    char *interface;
    nexus_bool_t free_interface;
#if defined(NEXUS_ARCH_SOLARIS)
    struct hostent hp2;
    char hp_tsdbuffer[500];
    int  hp_errnop;
#endif    
#if defined(NEXUS_ARCH_AIX_DCE_THREADS)
    int rc;
    struct hostent hp2;
    struct hostent_data hp_data;
#endif

    nexus_debug_printf(2, ("do_connect(): Connecting to %s/%hu\n",host,port));

    /*
     * First see if there is an alternate interface in the database,
     * and if so then try it.
     */
    if ((interface = nexus_database_lookup(host, 0, "tcp-interface")))
    {
	nexus_debug_printf(2, ("do_connect(): Using alternate interface %s\n", interface));
	free_interface = NEXUS_TRUE;
    }
    else
    {
	interface = host;
	free_interface = NEXUS_FALSE;
    }

#if defined(NEXUS_ARCH_SOLARIS)
#define NEXUS_DO_CONNECT_GETHOSTBYNAME
    hp = gethostbyname_r( interface, &hp2, hp_tsdbuffer, 500, &hp_errnop );
#endif
#if defined(NEXUS_ARCH_AIX_DCE_THREADS)
#define NEXUS_DO_CONNECT_GETHOSTBYNAME
    rc = gethostbyname_r( interface, &hp2, &hp_data );
    if (rc == 0)
    {
	hp = &hp2;
    }
    else
    {
	hp = (struct hostent *) NULL;
    }
#endif
#ifndef NEXUS_DO_CONNECT_GETHOSTBYNAME
    hp = gethostbyname(interface);
#endif
    
    if (hp == (struct hostent *) NULL)
    {
	tcp_fatal("do_connect(): gethostbyname failed for: %s\n", interface);
    }
    
    ZeroOutMemory(&his_addr, sizeof(his_addr));

    his_addr.sin_port = htons(port);
    memcpy(&his_addr.sin_addr, hp->h_addr, hp->h_length);
    his_addr.sin_family = hp->h_addrtype;

    connect_succeeded = NEXUS_FALSE;
    while (!connect_succeeded)
    {
	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    tcp_fatal("do_connect(): Failed to create socket: %s\n",
		      _nx_md_system_error_string());
	}
	use_his_addr = his_addr;

	if (connect(s, (struct sockaddr *) &use_his_addr,
		    sizeof(use_his_addr)) == 0)
	{
	    connect_succeeded = NEXUS_TRUE;
	}
	else
	{
	    if (errno == EINPROGRESS)
	    {
		/*
		 * man connect: EINPROGRESS:
		 *   The socket is non-blocking and the connection
		 *   cannot be completed immediately. It is possible to
		 *   select(2) for completion by selecting the socket
		 *   for writing.
		 * So this connect has, for all practical purposes, succeeded.
		 */
                connect_succeeded = NEXUS_TRUE;
		nexus_debug_printf(2, ("do_connect(): Connect to %s/%hu got EINPROGRESS: %s\n",
				       interface, port,
				       _nx_md_system_error_string()));
            }
	    else if (errno == EINTR)
	    {
#ifdef NEXUS_ARCH_SUNOS41
		/*
		 * With SunOS 4.1.3 (at least when used
		 * with FSU pthreads), connect() has
		 * an annoying little bug.  It may return EINTR, but
		 * still succeed.  When using round robin scheduling,
		 * where SIGALRM is used to timeslice between threads,
		 * this bug actually occurs with surprising frequency.
		 *
		 * My first attempted solution was to call getpeername(),
		 * with the idea that if it succeeds then the connect()
		 * succeeded, and if it failed then the connect() failed.
		 * This didn't work; getpeername() would sometime fail
		 * yet the connect() would still succeed.
		 *
		 * But it appears that selecting the file descriptor
		 * for writing causes it to wait until things are sane.
		 * (This idea comes from EINPROGRESS.)  And it also
		 * appears that when connect() returns EINTR it
		 * nonetheless always succeeds.
		 *
		 * Note that sometimes the select() returns EINTR,
		 * so we need to retry the select() in that case.
		 * And the EBADF check is just for good measure, as
		 * this would seem to catch the (as yet unobserved)
		 * case where the interrupted connect() actually
		 * failed.
		 */
		fd_set my_fd_set;
		int select_rc;
		int select_loop_done = NEXUS_FALSE;
		while (!select_loop_done)
		{
		    FD_ZERO(&my_fd_set);
		    FD_SET(s, &my_fd_set);
		    select_rc = select(s+1,
				       NEXUS_FD_SET_CAST NULL,
				       NEXUS_FD_SET_CAST &my_fd_set,
				       NEXUS_FD_SET_CAST NULL,
				       NULL);
		    if (select_rc == 1)
		    {
			select_loop_done = NEXUS_TRUE;
			connect_succeeded = NEXUS_TRUE;
		    }
		    else
		    {
			if (errno == EBADF)
			{
			    select_loop_done = NEXUS_TRUE;
			}
			else if (errno != EINTR)
			{
			    nexus_fatal("do_connect(): Ack! connect() returned EINTR, but selecting on the socket failed (select returned %d, errno=%d).  Guess its time to try another hack around this connect() bug.\n", select_rc, errno);
			}
		    }
		}
		
#else  /* NEXUS_ARCH_SUNOS41 */
		
		/*
		 * Do nothing.  Just try again.
		 */
		nexus_debug_printf(2, ("do_connect(): Connect to %s/%hu, got EINTR.  Retrying...\n", interface, port));
		
#endif /* NEXUS_ARCH_SUNOS41 */
	    }
	    else if (errno == ECONNREFUSED)
	    {
		nexus_debug_printf(2, ("do_connect(): Connect to %s/%hu refused.  Retrying...\n", interface, port));
		close(s);

		/*
		 * If the first few attempts fail, then backoff for a
		 * bit before trying again
		 */
		connect_failures++;
		if (connect_failures > CONNECT_BACKOFF_FAST_RETRY)
		{
#ifdef NEXUS_LITE
		    nexus_usleep(CONNECT_BACKOFF_USEC);
#else  /* NEXUS_LITE */
		    nexus_thread_yield();
#endif /* NEXUS_LITE */
		}
	    }
	    else
	    {
		tcp_fatal("do_connect(): Connect to %s/%hu failed: %s\n",
			  interface, port, _nx_md_system_error_string());
	    }

	}
    }
    
#ifdef NEXUS_DEBUG

    if (NexusDebug(2))
    {
	struct sockaddr_in my_addr;
	int len;

	len = sizeof(my_addr);
	if (getsockname(s, (struct sockaddr *) &my_addr, &len) == 0)
	    nexus_printf("New conn: remote=%s/%hu local=%hu\n",
			hp->h_name, port, htons(my_addr.sin_port));
	else
	    nexus_printf("New conn: remote=%s/%hu local=?? <%s>\n",
			hp->h_name, port, _nx_md_system_error_string());
    }
#endif

    config_socket(s);

    /*
     * Send our pid down the pipe so that
     * we can identify the incoming connection on the other
     * end.
     */
    pid = NX_HTONL((long) getpid());
    write_rc = write(s, &pid, 4);

    if (free_interface)
    {
	NexusFree(interface);
    }
    
    return (s);
} /* do_connect() */


/*
 * setup_listener()
 *
 * Return the file descriptor of this listener.  Also, return the
 * listener port in 'port'.
 */
static int setup_listener(u_short *port)
{
    struct sockaddr_in my_addr;
    int len;
    int listen_fd;
    
    ZeroOutMemory(&my_addr, sizeof(my_addr));

    my_addr.sin_addr.s_addr = INADDR_ANY;
    my_addr.sin_family = AF_INET;
    my_addr.sin_port = ntohs(*port);

    listen_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_fd < 0)
    {
	tcp_fatal("setup_listener(): socket() call failed: %s\n",
		  _nx_md_system_error_string());
    }

#if defined(SO_REUSEADDR) && !defined(NEXUS_ARCH_SUNOS41)
    /*
     * Set the port so it can be reused immediately if this
     * process dies.
     *
     * Under SunOS4.1 there is apparently a bug in this.
     * With this option set, a bind() will succeed on a port
     * even if that port is still in active use.  In fact, it will
     * succeed even if that port is still in active use by another
     * process.
     */
    if (*port != 0)
    {
	int one = 1;
	if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, (char *) &one,
		       sizeof(one)) < 0)
	{
	    nexus_warning("setup_listener(): setsockopt SO_REUSEADDR failed: %s\n", _nx_md_system_error_string());
	}
    }
#endif /* SO_REUSEADDR */
    
    if (bind(listen_fd, (struct sockaddr *) &my_addr, sizeof(my_addr)) < 0)
    {
	if (errno == EADDRINUSE) /* Not sure if this is right */
	{
	    return(-1);
	}
	else
	{
	    tcp_fatal("setup_listener(): bind() call failed: %s\n",
		      _nx_md_system_error_string());
	}
    }

    if (listen(listen_fd, SOMAXCONN) < 0)
    {
	tcp_fatal("setup_listener(): listen() call failed: %s\n",
		  _nx_md_system_error_string());
    }

    len = sizeof(my_addr);
    if (getsockname(listen_fd, (struct sockaddr *) &my_addr, &len) < 0)
    {
	tcp_fatal("setup_listener(): getsockname() call failed: %s\n",
		  _nx_md_system_error_string());
    }

    *port = ntohs(my_addr.sin_port);

    return (listen_fd);
} /* setup_listener() */
    

/*
 * proto_table_init()
 *
 * Initialize the protocol table.
 */
static void proto_table_init()
{
    int i;

    for (i = 0; i < PROTO_TABLE_SIZE; i++)
    {
	proto_table[i].proto = (tcp_proto_t *) NULL;
	proto_table[i].next = (proto_table_entry_t *) NULL;
    }
} /* proto_table_init() */


/*
 * proto_table_hash()
 *
 * Hash the hostname and port for the proto table.
 */
static int proto_table_hash(char *host, u_short port)
{
    unsigned int hval = 0;
    char *s;

    for (s = host; *s != '\0'; s++)
    {
	hval <<= 1;
	hval += *s;
    }
    hval <<= 1;
    hval += port;
    return (hval % PROTO_TABLE_SIZE);
} /* proto_table_hash() */


/*
 * proto_table_insert()
 *
 * Insert the given proto into the table, hashing on its internal
 * hostname and port number.
 *
 * We assume that the entry is not present in the table.
 */
static void proto_table_insert(tcp_proto_t *proto)
{
    int bucket;
    proto_table_entry_t *new_ent;

    bucket = proto_table_hash(proto->host, proto->port);
    if (proto_table[bucket].proto == (tcp_proto_t *) NULL)
    {
	/* Drop it into the preallocated table entry */
	proto_table[bucket].proto = proto;
    }
    else
    {
	/*
	 * Need to allocate a new proto_table_entry_t and add it
	 * to the bucket
	 */
	NexusMalloc(proto_table_insert(), new_ent, proto_table_entry_t *,
		    sizeof(struct _proto_table_entry_t));

	new_ent->proto = proto;
	new_ent->next = proto_table[bucket].next;

	proto_table[bucket].next = new_ent;
    }
    nexus_debug_printf(2, ("proto_table_inserted(): Inserted proto=%x for %s/%hu bucket=%d\n",
			   proto, proto->host, proto->port, bucket));
} /* proto_table_insert() */


/*
 * proto_table_lookup()
 *
 * Look up and return the tcp_proto_t for the given hostname
 * and port. Return NULL if none exists.
 */
static tcp_proto_t *proto_table_lookup(char *host, u_short port)
{
    proto_table_entry_t *ent;
    int bucket;

    bucket = proto_table_hash(host, port);

    for (ent = &(proto_table[bucket]);
	 ent != (proto_table_entry_t *) NULL;
	 ent = ent->next)
    {
	if (   (ent->proto != (tcp_proto_t *) NULL)
	    && (ent->proto->port == port)
	    && (strcmp(ent->proto->host, host) == 0) )
	{
	    nexus_debug_printf(2, ("proto_table_lookup(): Found entry %x proto=%x for %s/%hu bucket=%d\n",
				   ent, ent->proto, host, port, bucket));
	    return (ent->proto);
	}
    }
    
    nexus_debug_printf(2, ("proto_table_lookup(): Didn't find entry for %s/%hu bucket=%d\n",
			   host, port, bucket));

    return ((tcp_proto_t *) NULL);
} /* proto_table_lookup() */


#ifdef NEXUS_DEBUG
static void print_fd_set(char *fmt_string, fd_set *set)
{
    char fd_buf[1024], fd_tmp_buf[10];
    int fd_i;

    fd_buf[0] = '\0';
    for (fd_i = 0; fd_i < fd_tablesize; fd_i++)
    {
	if (FD_ISSET(fd_i, set))
	{
	    sprintf(fd_tmp_buf, "%d ", fd_i);
	    strcat(fd_buf, fd_tmp_buf);
	}
    }
    nexus_printf(fmt_string, fd_buf);
} /* print_fd_set() */
#endif /* NEXUS_DEBUG */



/*********************************************************************
 * 		Buffer management code
 *********************************************************************/

#ifdef NEXUS_SANITY_CHECK
#define SANITY_TYPE_CHECK_SIZE (sizeof(unsigned long) + sizeof(int))
#else  /* NEXUS_SANITY_CHECK */
#define SANITY_TYPE_CHECK_SIZE 0
#endif /* NEXUS_SANITY_CHECK */

/*
 * tcp_set_buffer_size()
 * 
 * Allocate message buffer space for 'size' bytes, that will be used to
 * hold 'n_elements'.
 */
static void tcp_set_buffer_size(nexus_buffer_t *buffer,
				int size, int n_elements)
{
    tcp_buffer_t *tcp_buffer;
    char *storage;

    NexusBufferMagicCheck(tcp_check_buffer_size, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == SEND_BUFFER),
		 ("tcp_set_buffer_size(): Internal error: Expected a send buffer\n"));

#ifdef NEXUS_SANITY_CHECK
    if (n_elements != -1) {
	size += SANITY_TYPE_CHECK_SIZE * n_elements;
    }
#endif

    if (size > 0 )
    {
	NexusMalloc(tcp_set_buffer_size(), storage, char *,
		    (size + tcp_buffer->u.send.header_size));

	tcp_buffer->u.send.size = size;
	tcp_buffer->u.send.n_elements = n_elements;
	tcp_buffer->storage = storage;
	SetBufferPointers(tcp_buffer);
	tcp_buffer->u.send.current_pointer = tcp_buffer->u.send.base_pointer;
    }

} /* tcp_set_buffer_size() */


/*
 * tcp_check_buffer_size()
 *
 * Check that that passed message 'buffer' has at least 'slack'
 * bytes remaining; if not, increase size by 'increment' bytes
 * until there are enough bytes remaining.
 *
 * If no resizing is necessary, leave 'buffer' unchanged and
 * return NEXUS_TRUE.
 * If resizing is successful, modify 'buffer' to a new, larger
 * buffer and return NEXUS_TRUE.
 * Otherwise, if 'increment' is 0 and 'slack' bytes are not
 * available in the buffer, then leave 'buffer' unchanged and
 * return NEXUS_FALSE.
 */
static int tcp_check_buffer_size(nexus_buffer_t *buffer,
				 int slack, int increment)
{
    tcp_buffer_t *tcp_buffer;
    int used;
    int needed;

    NexusBufferMagicCheck(tcp_check_buffer_size, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == SEND_BUFFER),
		 ("tcp_check_buffer_size(): Internal error: Expected a send buffer\n"));
    
    used = (tcp_buffer->u.send.current_pointer
	    - tcp_buffer->u.send.base_pointer);
    needed = used + slack;

    if (tcp_buffer->u.send.size == 0)
    {
	tcp_set_buffer_size(buffer, slack, -1);
    }
    else if (needed > tcp_buffer->u.send.size)
    {
	char *new_storage;
	int new_size;

	if (increment <= 0)
	    return(NEXUS_FALSE);

	new_size = tcp_buffer->u.send.size;
	while (new_size < needed)
	{
	    new_size += increment;
	}

	NexusMalloc(tcp_check_buffer_size(), new_storage, char *,
		    (new_size + tcp_buffer->u.send.header_size));
	
	memcpy(new_storage + tcp_buffer->u.send.header_size,
	       tcp_buffer->u.send.base_pointer,
	       used);
	    
	NexusFree(tcp_buffer->storage);

	tcp_buffer->u.send.size = new_size;
	tcp_buffer->storage = new_storage;
	SetBufferPointers(tcp_buffer);
	tcp_buffer->u.send.current_pointer = (tcp_buffer->u.send.base_pointer
					      + used);
    }
    
    return(NEXUS_TRUE);
} /* tcp_check_buffer_size() */


/*
 * tcp_free_buffer()
 *
 * Free the passed nexus_buffer_t.
 *
 * This should be called on the receiving end, after the handler
 * has completed.
 *
 * Note: The stashed flag could be set, since tcp_stash_buffer()
 * just sets this flag and typecasts a buffer to a stashed buffer.
 * In this case, do not free the buffer.
 */
static void tcp_free_buffer(nexus_buffer_t *buffer)
{
    tcp_buffer_t *tcp_buffer;
    
    NexusAssert2((buffer),
		 ("tcp_free_buffer(): Passed a NULL nexus_buffer_t *\n") );
    
    /* If the buffer was stashed, *buffer will have been set to NULL */
    if (!(*buffer))
	return;

    NexusBufferMagicCheck(tcp_free_buffer, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == RECV_BUFFER),
		 ("tcp_free_buffer(): Expected a receive buffer\n"));
    NexusAssert2((!tcp_buffer->u.recv.stashed),
		 ("tcp_free_buffer(): Expected a non-stashed buffer\n"));

    FreeTcpBuffer(tcp_buffer);

    *buffer = (nexus_buffer_t) NULL;
    
} /* tcp_free_buffer() */


/*
 * tcp_stash_buffer()
 *
 * Convert 'buffer' to a stashed buffer.
 */
static void tcp_stash_buffer(nexus_buffer_t *buffer,
			     nexus_stashed_buffer_t *stashed_buffer)
{
    tcp_buffer_t *tcp_buffer;
    
    NexusBufferMagicCheck(tcp_stash_buffer, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    NexusAssert2((tcp_buffer->buffer_type == RECV_BUFFER),
		 ("tcp_stash_buffer(): Expected a receive buffer\n"));
    NexusAssert2((!tcp_buffer->u.recv.stashed),
		 ("tcp_stash_buffer(): Expected an un-stashed buffer\n"));
    
    tcp_buffer->u.recv.stashed = NEXUS_TRUE;
		  
    *stashed_buffer = (nexus_stashed_buffer_t) *buffer;
    
    *buffer = (nexus_buffer_t) NULL;
} /* tcp_stash_buffer() */


/*
 * tcp_free_stashed_buffer()
 *
 * Free the passed nexus_stashed_buffer_t that was stashed
 * by tcp_stash_buffer().
 */
static void tcp_free_stashed_buffer(nexus_stashed_buffer_t *stashed_buffer)
{
    tcp_buffer_t *tcp_buffer;

    NexusAssert2((stashed_buffer),
		 ("tcp_free_stashed_buffer(): Passed a NULL nexus_stashed_buffer_t *\n") );
    NexusBufferMagicCheck(tcp_free_stashed_buffer,
			  (nexus_buffer_t *) stashed_buffer);
    
    tcp_buffer = (tcp_buffer_t *) *stashed_buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == RECV_BUFFER),
		 ("tcp_free_stashed_buffer(): Expected a receive buffer\n"));
    NexusAssert2((tcp_buffer->u.recv.stashed),
		 ("tcp_free_stashed_buffer(): Expected a stashed buffer\n"));

    FreeTcpBuffer(tcp_buffer);

    *stashed_buffer = (nexus_stashed_buffer_t) NULL;
    
} /* tcp_free_stashed_buffer() */



/*
 * tcp_sizeof_*()
 *
 * Return the size (in bytes) that 'count' elements of the given
 * type will require to be put into the 'buffer'.
 */
static int tcp_sizeof_float(nexus_buffer_t *buffer, int count)
{
    return(sizeof(float) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_double(nexus_buffer_t *buffer, int count)
{
    return(sizeof(double) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_short(nexus_buffer_t *buffer, int count)
{
    return(sizeof(short) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_u_short(nexus_buffer_t *buffer, int count)
{
    return(sizeof(unsigned short) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_int(nexus_buffer_t *buffer, int count)
{
    return(sizeof(int) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_u_int(nexus_buffer_t *buffer, int count)
{
    return(sizeof(unsigned int) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_long(nexus_buffer_t *buffer, int count)
{
    return(sizeof(long) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_u_long(nexus_buffer_t *buffer, int count)
{
    return(sizeof(unsigned long) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_char(nexus_buffer_t *buffer, int count)
{
    return(sizeof(char) * count + SANITY_TYPE_CHECK_SIZE);
}

static int tcp_sizeof_u_char(nexus_buffer_t *buffer, int count)
{
    return(sizeof(unsigned char) * count + SANITY_TYPE_CHECK_SIZE);
}


#ifdef NEXUS_ARCH_AIX
#define DO_TRACEBACK _nx_traceback()
#else
#define DO_TRACEBACK
#endif


/*
 * tcp_put_*()
 *
 * Put 'count' elements, starting at 'data', of the given type,
 * into 'buffer'.
 */

#ifdef NEXUS_DEBUG
#define DO_PUT_ASSERTIONS(TYPE) \
    NexusBufferMagicCheck(tcp_put_ ## TYPE, buf); \
    if (((tcp_buf->u.send.current_pointer - tcp_buf->u.send.base_pointer) + count * sizeof(TYPE) + SANITY_TYPE_CHECK_SIZE) > tcp_buf->u.send.size) \
    { \
	tcp_fatal("nexus_put_" #TYPE ": Buffer overrun\n"); \
    } \
    if(NexusDebug(3)) \
    { \
        nexus_printf("tcp_put_" #TYPE "(): base:%x size:%lu current:%x limit:%x count:%d\n", \
		     tcp_buf->u.send.base_pointer, \
		     tcp_buf->u.send.size, \
		     tcp_buf->u.send.current_pointer, \
		     tcp_buf->u.send.base_pointer + tcp_buf->u.send.size, \
		     count); \
    }

#else /* NEXUS_DEBUG */
#define DO_PUT_ASSERTIONS(TYPE)
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_SPOOF_COPY
#define PUT(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_PUT_ASSERTIONS(TYPE); \
    if (!_nx_spoof_copy) \
        memcpy(tcp_buf->u.send.current_pointer, data, count * sizeof(TYPE)); \
    tcp_buf->u.send.current_pointer = \
        (char *) tcp_buf->u.send.current_pointer + (count * sizeof(TYPE)); \
}
#else

#ifdef NEXUS_SANITY_CHECK
#define PUT(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_PUT_ASSERTIONS(TYPE); \
    { \
        unsigned long sanity_flag_long   = 0xfcb1b00a;  \
        unsigned long sanity_flag_int    = 0xfcb1b00b;  \
        unsigned long sanity_flag_char   = 0xfcb1b00c;  \
        unsigned long sanity_flag_double = 0xfcb1b00d;  \
        unsigned long sanity_flag_short  = 0xfcb1b00e;  \
        unsigned long sanity_flag_float  = 0xfcb1b00f;  \
                                                        \
        unsigned long sanity_flag_u_long  = 0xfcb1b01a; \
        unsigned long sanity_flag_u_int   = 0xfcb1b01b; \
        unsigned long sanity_flag_u_char  = 0xfcb1b01c; \
                                                        \
        unsigned long sanity_flag_u_short = 0xfcb1b01e; \
                                                        \
        memcpy(tcp_buf->u.send.current_pointer, & sanity_flag_ ##TYPE, sizeof(unsigned long)); \
        tcp_buf->u.send.current_pointer = \
            (char *) tcp_buf->u.send.current_pointer + (sizeof(unsigned long)); \
        memcpy(tcp_buf->u.send.current_pointer, &count, sizeof(int)); \
        tcp_buf->u.send.current_pointer = \
            (char *) tcp_buf->u.send.current_pointer + (sizeof(int)); \
    } \
    memcpy(tcp_buf->u.send.current_pointer, data, count * sizeof(TYPE)); \
    tcp_buf->u.send.current_pointer = \
        (char *) tcp_buf->u.send.current_pointer + (count * sizeof(TYPE)); \
} 
#else  /* NEXUS_SANITY_CHECK */
#define PUT(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_PUT_ASSERTIONS(TYPE); \
    memcpy(tcp_buf->u.send.current_pointer, data, count * sizeof(TYPE)); \
    tcp_buf->u.send.current_pointer = \
        (char *) tcp_buf->u.send.current_pointer + (count * sizeof(TYPE)); \
}
#endif /* NEXUS_SANITY_CHECK */
#endif

static void tcp_put_float(nexus_buffer_t *buf, float *data, int count)
{
    PUT(float);
}

static void tcp_put_double(nexus_buffer_t *buf, double *data, int count)
{
    PUT(double);
}

static void tcp_put_short(nexus_buffer_t *buf, short *data, int count)
{
    PUT(short);
}

static void tcp_put_u_short(nexus_buffer_t *buf, unsigned short *data,
			    int count)
{
    PUT(u_short);
}

static void tcp_put_int(nexus_buffer_t *buf, int *data, int count)
{
    PUT(int);
}

static void tcp_put_u_int(nexus_buffer_t *buf, unsigned int *data, int count)
{
    PUT(u_int);
}

static void tcp_put_long(nexus_buffer_t *buf, long *data, int count)
{
    PUT(long);
}

static void tcp_put_u_long(nexus_buffer_t *buf, unsigned long *data, int count)
{
    PUT(u_long);
}

static void tcp_put_char(nexus_buffer_t *buf, char *data, int count)
{
    PUT(char);
}

static void tcp_put_u_char(nexus_buffer_t *buf, unsigned char *data, int count)
{
    PUT(u_char);
}



/*
 * tcp_get_*()
 *
 * Get 'count' elements of the given type from 'buffer' and store
 * them into 'data'.
 */

#ifdef NEXUS_DEBUG
#define DO_GET_ASSERTIONS(TYPE) \
    NexusBufferMagicCheck(tcp_get_ ## TYPE, buf); \
    NexusAssert2(!tcp_buf->u.recv.stashed, \
		 ("tcp_get_*(): Expected an un-stashed buffer\n")); \
    if (tcp_buf->u.recv.current_pointer > (tcp_buf->storage + tcp_buf->u.recv.size)) \
    { \
	tcp_fatal("nexus_get_" #TYPE ": Buffer overrun\n"); \
    } \
    if(NexusDebug(3)) \
    { \
        nexus_printf("tcp_get_" #TYPE "(): base:%x size:%lu current:%x limit:%x count:%d\n", \
		     tcp_buf->storage, \
		     tcp_buf->u.recv.size, \
		     tcp_buf->u.recv.current_pointer, \
		     tcp_buf->storage + tcp_buf->u.recv.size, \
		     count); \
    }

#else /* NEXUS_DEBUG */
#define DO_GET_ASSERTIONS(TYPE)
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_SPOOF_COPY
#define GET(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_GET_ASSERTIONS(TYPE); \
    if (!_nx_spoof_copy) \
        memcpy(dest, tcp_buf->u.recv.current_pointer, sizeof(TYPE) * count); \
    tcp_buf->u.recv.current_pointer += count * sizeof(TYPE); \
}
#else
#ifdef NEXUS_SANITY_CHECK
#define GET(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_GET_ASSERTIONS(TYPE); \
    { \
        unsigned long sanity_flag_long   = 0xfcb1b00a;  \
        unsigned long sanity_flag_int    = 0xfcb1b00b;  \
        unsigned long sanity_flag_char   = 0xfcb1b00c;  \
        unsigned long sanity_flag_double = 0xfcb1b00d;  \
        unsigned long sanity_flag_short  = 0xfcb1b00e;  \
        unsigned long sanity_flag_float  = 0xfcb1b00f;  \
                                                        \
        unsigned long sanity_flag_u_long  = 0xfcb1b01a; \
        unsigned long sanity_flag_u_int   = 0xfcb1b01b; \
        unsigned long sanity_flag_u_char  = 0xfcb1b01c; \
                                                        \
        unsigned long sanity_flag_u_short = 0xfcb1b01e; \
                                                        \
        unsigned long buffer_sanity_value; \
        int buffer_sanity_count; \
        \
        memcpy( &buffer_sanity_value, tcp_buf->u.recv.current_pointer, sizeof(unsigned long)); \
        if(  buffer_sanity_value != sanity_flag_ ## TYPE ) { \
    nexus_printf("tcp_get_" #TYPE ": warning: expected type of %x but found %x\n", \
		         sanity_flag_ ## TYPE , \
		         buffer_sanity_value); \
	   DO_TRACEBACK; \
        } \
        tcp_buf->u.recv.current_pointer += sizeof(unsigned long); \
        memcpy( &buffer_sanity_count, tcp_buf->u.recv.current_pointer, sizeof(unsigned long)); \
        if( buffer_sanity_count != count ) { \
         nexus_printf("tcp_get_" #TYPE ": warning: expected count of %x but found %x\n", \
		         count, \
		         buffer_sanity_count); \
	   DO_TRACEBACK; \
        } \
        tcp_buf->u.recv.current_pointer += sizeof(int); \
    } \
    memcpy(dest, tcp_buf->u.recv.current_pointer, sizeof(TYPE) * count); \
    tcp_buf->u.recv.current_pointer += count * sizeof(TYPE); \
}
#else /* NEXUS_SANITY_CHECK */
#define GET(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_GET_ASSERTIONS(TYPE); \
    memcpy(dest, tcp_buf->u.recv.current_pointer, sizeof(TYPE) * count); \
    tcp_buf->u.recv.current_pointer += count * sizeof(TYPE); \
}
#endif /* NEXUS_SANITY_CHECK */
#endif


static void tcp_get_float(nexus_buffer_t *buf, float *dest, int count)
{
    GET(float);
}

static void tcp_get_double(nexus_buffer_t *buf, double *dest, int count)
{
    GET(double);
}

static void tcp_get_short(nexus_buffer_t *buf, short *dest, int count)
{
    GET(short);
}

static void tcp_get_u_short(nexus_buffer_t *buf, unsigned short *dest,
			    int count)
{
    GET(u_short);
}

static void tcp_get_int(nexus_buffer_t *buf, int *dest, int count)
{
    GET(int);
}

static void tcp_get_u_int(nexus_buffer_t *buf, unsigned int *dest, int count)
{
    GET(u_int);
}

static void tcp_get_long(nexus_buffer_t *buf, long *dest, int count)
{
    GET(long);
}

static void tcp_get_u_long(nexus_buffer_t *buf, unsigned long *dest, int count)
{
    GET(u_long);
}

static void tcp_get_char(nexus_buffer_t *buf, char *dest, int count)
{
    GET(char);
}

static void tcp_get_u_char(nexus_buffer_t *buf, unsigned char *dest, int count)
{
    GET(u_char);
}



/*
 * tcp_get_stashed_*()
 *
 * Get 'count' elements of the given type from the stashed 'buffer' and store
 * them into 'data'.
 */

#ifdef NEXUS_DEBUG
#define DO_GET_STASHED_ASSERTIONS(TYPE) \
    NexusBufferMagicCheck(tcp_get_stashed_ ## TYPE, (nexus_buffer_t *) buf); \
    NexusAssert2(tcp_buf->u.recv.stashed, \
		 ("tcp_get_stashed_*(): Expected a stashed buffer\n"));   \
    if (tcp_buf->u.recv.current_pointer > (tcp_buf->storage + tcp_buf->u.recv.size)) \
    { \
	tcp_fatal("nexus_get_stashed_" #TYPE ": Buffer overrun\n"); \
    } \
    if(NexusDebug(3)) \
    { \
        nexus_printf("tcp_get_stashed_" #TYPE "(): base:%x size:%lu current:%x limit:%x count:%d\n", \
		     tcp_buf->storage, \
		     tcp_buf->u.recv.size, \
		     tcp_buf->u.recv.current_pointer, \
		     tcp_buf->storage + tcp_buf->u.recv.size, \
		     count); \
    }

#else /* NEXUS_DEBUG */
#define DO_GET_STASHED_ASSERTIONS(TYPE)
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_SPOOF_COPY
#define GET_STASHED(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_GET_STASHED_ASSERTIONS(TYPE); \
    if (!_nx_spoof_copy) \
        memcpy(dest, tcp_buf->u.recv.current_pointer, sizeof(TYPE) * count); \
    tcp_buf->u.recv.current_pointer += count * sizeof(TYPE); \
}
#else
#ifdef NEXUS_SANITY_CHECK
#define GET_STASHED(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_GET_STASHED_ASSERTIONS(TYPE); \
    { \
        unsigned long sanity_flag_long   = 0xfcb1b00a;  \
        unsigned long sanity_flag_int    = 0xfcb1b00b;  \
        unsigned long sanity_flag_char   = 0xfcb1b00c;  \
        unsigned long sanity_flag_double = 0xfcb1b00d;  \
        unsigned long sanity_flag_short  = 0xfcb1b00e;  \
        unsigned long sanity_flag_float  = 0xfcb1b00f;  \
                                                        \
        unsigned long sanity_flag_u_long  = 0xfcb1b01a; \
        unsigned long sanity_flag_u_int   = 0xfcb1b01b; \
        unsigned long sanity_flag_u_char  = 0xfcb1b01c; \
                                                        \
        unsigned long sanity_flag_u_short = 0xfcb1b01e; \
                                                        \
        unsigned long buffer_sanity_value; \
        int buffer_sanity_count; \
        \
        memcpy( &buffer_sanity_value, tcp_buf->u.recv.current_pointer, sizeof(unsigned long)); \
        if(  buffer_sanity_value != sanity_flag_ ## TYPE ) { \
    nexus_printf("tcp_get_stashed_" #TYPE ": warning: expected type of %x but found %x\n", \
		         sanity_flag_ ## TYPE , \
		         buffer_sanity_value); \
	   DO_TRACEBACK; \
        } \
        tcp_buf->u.recv.current_pointer += sizeof(unsigned long); \
        memcpy( &buffer_sanity_count, tcp_buf->u.recv.current_pointer, sizeof(unsigned long)); \
        if( buffer_sanity_count != count ) { \
         nexus_printf("tcp_get_stashed_" #TYPE ": warning: expected count of %x but found %x\n", \
		         count, \
		         buffer_sanity_count); \
	   DO_TRACEBACK; \
        } \
        tcp_buf->u.recv.current_pointer += sizeof(int); \
    } \
    memcpy(dest, tcp_buf->u.recv.current_pointer, sizeof(TYPE) * count); \
    tcp_buf->u.recv.current_pointer += count * sizeof(TYPE); \
}
#else  /* NEXUS_SANITY_CHECK */
#define GET_STASHED(TYPE) \
{ \
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf; \
    DO_GET_STASHED_ASSERTIONS(TYPE); \
    memcpy(dest, tcp_buf->u.recv.current_pointer, sizeof(TYPE) * count); \
    tcp_buf->u.recv.current_pointer += count * sizeof(TYPE); \
}
#endif /* NEXUS_SANITY_CHECK */
#endif


static void tcp_get_stashed_float(nexus_stashed_buffer_t *buf,
				  float *dest, int count)
{
    GET_STASHED(float);
}

static void tcp_get_stashed_double(nexus_stashed_buffer_t *buf,
				   double *dest, int count)
{
    GET_STASHED(double);
}

static void tcp_get_stashed_short(nexus_stashed_buffer_t *buf,
				  short *dest, int count)
{
    GET_STASHED(short);
}

static void tcp_get_stashed_u_short(nexus_stashed_buffer_t *buf,
				    unsigned short *dest, int count)
{
    GET_STASHED(u_short);
}

static void tcp_get_stashed_int(nexus_stashed_buffer_t *buf,
				int *dest, int count)
{
    GET_STASHED(int);
}

static void tcp_get_stashed_u_int(nexus_stashed_buffer_t *buf,
				  unsigned int *dest, int count)
{
    GET_STASHED(u_int);
}

static void tcp_get_stashed_long(nexus_stashed_buffer_t *buf,
				 long *dest, int count)
{
    GET_STASHED(long);
}

static void tcp_get_stashed_u_long(nexus_stashed_buffer_t *buf,
				   unsigned long *dest, int count)
{
    GET_STASHED(u_long);
}

static void tcp_get_stashed_char(nexus_stashed_buffer_t *buf,
				 char *dest, int count)
{
    GET_STASHED(char);
}

static void tcp_get_stashed_u_char(nexus_stashed_buffer_t *buf,
				   unsigned char *dest, int count)
{
    GET_STASHED(u_char);
}


/*********************************************************************
 * 		Buffer management code XDR
 *********************************************************************/

/*
 * tcp_xdr_set_buffer_size()
 * 
 * Allocate message buffer space for 'size' bytes, that will be used to
 * hold 'n_elements'.
 */
static void tcp_xdr_set_buffer_size(nexus_buffer_t *buffer,
				int size, int n_elements)
{
    tcp_buffer_t *tcp_buffer;
    char *storage;

    NexusBufferMagicCheck(tcp_check_buffer_size, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == SEND_BUFFER),
		 ("tcp_xdr_set_buffer_size(): Internal error: Expected a send buffer\n"));

    if (size > 0 )
    {
	NexusMalloc(tcp_set_buffer_size(), storage, char *,
		    (size + tcp_buffer->u.send.header_size));

	tcp_buffer->u.send.size = size;
	tcp_buffer->u.send.n_elements = n_elements;
	tcp_buffer->storage = storage;
	SetBufferPointers(tcp_buffer);
	tcp_buffer->u.send.current_pointer = tcp_buffer->u.send.base_pointer;

	/* do XDR stuff here */
	/* xdrmem_create */
	xdrmem_create(&(tcp_buffer->xdrs), tcp_buffer->u.send.base_pointer,
		      tcp_buffer->u.send.size, XDR_ENCODE );
    }

} /* tcp_xdr_set_buffer_size() */


/*
 * tcp_xdr_check_buffer_size()
 *
 * Check that that passed message 'buffer' has at least 'slack'
 * bytes remaining; if not, increase size by 'increment' bytes
 * until there are enough bytes remaining.
 *
 * If no resizing is necessary, leave 'buffer' unchanged and
 * return NEXUS_TRUE.
 * If resizing is successful, modify 'buffer' to a new, larger
 * buffer and return NEXUS_TRUE.
 * Otherwise, if 'increment' is 0 and 'slack' bytes are not
 * available in the buffer, then leave 'buffer' unchanged and
 * return NEXUS_FALSE.
 */
static int tcp_xdr_check_buffer_size(nexus_buffer_t *buffer,
				 int slack, int increment)
{
    tcp_buffer_t *tcp_buffer;
    unsigned int used;
    int needed;

    NexusBufferMagicCheck(tcp_xdr_check_buffer_size, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == SEND_BUFFER),
		 ("tcp_xdr_check_buffer_size(): Internal error: Expected a send buffer\n"));

    /* do XDR stuff here */
    /* used should be based on xdr_getpos */

    used = xdr_getpos(&(tcp_buffer->xdrs));
    
    needed = used + slack;

    if (tcp_buffer->u.send.size == 0)
	{
	    tcp_xdr_set_buffer_size(buffer, slack, -1);
	}
    else if (needed > tcp_buffer->u.send.size)
	{
	    char *new_storage, *old_storage;
	    int new_size;

	if (increment <= 0)
	    return(NEXUS_FALSE);

	new_size = tcp_buffer->u.send.size;
	while (new_size < needed)
	{
	    new_size += increment;
	}

	NexusMalloc(tcp_xdr_check_buffer_size(), new_storage, char *,
		    (new_size + tcp_buffer->u.send.header_size));
	
	memcpy(new_storage + tcp_buffer->u.send.header_size,
	       tcp_buffer->u.send.base_pointer,
	       used);
	    
	old_storage = tcp_buffer->storage;

	tcp_buffer->u.send.size = new_size;
	tcp_buffer->storage = new_storage;
	SetBufferPointers(tcp_buffer);
	tcp_buffer->u.send.current_pointer = (tcp_buffer->u.send.base_pointer
					      + used);

	/* do XDR stuff here */
	/* xdr_destroy */
	xdr_destroy(&(tcp_buffer->xdrs));
	NexusFree(old_storage);
	/* xdrmem_create */
	xdrmem_create(&(tcp_buffer->xdrs), tcp_buffer->storage,
		      new_size, XDR_ENCODE);
	/* xdr_setpos */
	xdrrc = xdr_setpos(&(tcp_buffer->xdrs), used);
	NexusAssert2((xdrrc), 
	      ("tcp_xdr_check_buffer_size(): xdr_setpos() returned error\n"));
    }
    
    return(NEXUS_TRUE);
} /* tcp_xdr_check_buffer_size() */


/*
 * tcp_xdr_free_buffer()
 *
 * Free the passed nexus_buffer_t.
 *
 * This should be called on the receiving end, after the handler
 * has completed.
 *
 * Note: The stashed flag could be set, since tcp_xdr_stash_buffer()
 * just sets this flag and typecasts a buffer to a stashed buffer.
 * In this case, do not free the buffer.
 */
static void tcp_xdr_free_buffer(nexus_buffer_t *buffer)
{
    tcp_buffer_t *tcp_buffer;
    
    NexusAssert2((buffer),
		 ("tcp_xdr_free_buffer(): Passed a NULL nexus_buffer_t *\n") );
    
    /* If the buffer was stashed, *buffer will have been set to NULL */
    if (!(*buffer))
	return;

    NexusBufferMagicCheck(tcp_xdr_free_buffer, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == RECV_BUFFER),
		 ("tcp_xdr_free_buffer(): Expected a receive buffer\n"));
    NexusAssert2((!tcp_buffer->u.recv.stashed),
		 ("tcp_xdr_free_buffer(): Expected a non-stashed buffer\n"));

    xdr_destroy(&(tcp_buffer->xdrs));

    FreeTcpBuffer(tcp_buffer);

    *buffer = (nexus_buffer_t) NULL;
    
} /* tcp_xdr_free_buffer() */


/*
 * tcp_xdr_stash_buffer()
 *
 * Convert 'buffer' to a stashed buffer.
 */
static void tcp_xdr_stash_buffer(nexus_buffer_t *buffer,
			     nexus_stashed_buffer_t *stashed_buffer)
{
    tcp_buffer_t *tcp_buffer;
    
    NexusBufferMagicCheck(tcp_xdr_stash_buffer, buffer);

    tcp_buffer = (tcp_buffer_t *) *buffer;
    NexusAssert2((tcp_buffer->buffer_type == RECV_BUFFER),
		 ("tcp_xdr_stash_buffer(): Expected a receive buffer\n"));
    NexusAssert2((!tcp_buffer->u.recv.stashed),
		 ("tcp_xdr_stash_buffer(): Expected an un-stashed buffer\n"));
    
    tcp_buffer->u.recv.stashed = NEXUS_TRUE;
		  
    *stashed_buffer = (nexus_stashed_buffer_t) *buffer;
    
    *buffer = (nexus_buffer_t) NULL;
} /* tcp_xdr_stash_buffer() */


/*
 * tcp_xdr_free_stashed_buffer()
 *
 * Free the passed nexus_stashed_buffer_t that was stashed
 * by tcp_xdr_stash_buffer().
 */
static void tcp_xdr_free_stashed_buffer(nexus_stashed_buffer_t *stashed_buffer)
{
    tcp_buffer_t *tcp_buffer;

    NexusAssert2((stashed_buffer),
		 ("tcp_xdr_free_stashed_buffer(): Passed a NULL nexus_stashed_buffer_t *\n") );
    NexusBufferMagicCheck(tcp_xdr_free_stashed_buffer,
			  (nexus_buffer_t *) stashed_buffer);
    
    tcp_buffer = (tcp_buffer_t *) *stashed_buffer;
    
    NexusAssert2((tcp_buffer->buffer_type == RECV_BUFFER),
		 ("tcp_xdr_free_stashed_buffer(): Expected a receive buffer\n"));
    NexusAssert2((tcp_buffer->u.recv.stashed),
		 ("tcp_xdr_free_stashed_buffer(): Expected a stashed buffer\n"));
    xdr_destroy(&(tcp_buffer->xdrs));

    FreeTcpBuffer(tcp_buffer);

    *stashed_buffer = (nexus_stashed_buffer_t) NULL;
    
} /* tcp_xdr_free_stashed_buffer() */



/*
 * tcp_xdr_sizeof_*()
 *
 * Return the size (in bytes) that 'count' elements of the given
 * type will require to be put into the 'buffer'.
 */
#define XDR_SIZEOF(TYPE,cnt)        \
{                            \
    int i;                    \
    i=sizeof(TYPE) * cnt;    \
    while (i%4) i++;         \
    return i;                \
}

#define XDR_SIZEOF_CHARS(TYPE,cnt)  \
{                                       \
    int i;                               \
    i=sizeof(TYPE) * cnt;               \
    while (i%4) i++;                    \
    return i+4;                         \
}

static int tcp_xdr_sizeof_float(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(float,count);
}

static int tcp_xdr_sizeof_double(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(float,count);
}

static int tcp_xdr_sizeof_short(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(short,count);
}

static int tcp_xdr_sizeof_u_short(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(unsigned short,count);
}

static int tcp_xdr_sizeof_int(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(int,count);
}

static int tcp_xdr_sizeof_u_int(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(unsigned int,count);
}

static int tcp_xdr_sizeof_long(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(long,count);
}

static int tcp_xdr_sizeof_u_long(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF(unsigned long,count);
}

static int tcp_xdr_sizeof_char(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF_CHARS(char,count);
}

static int tcp_xdr_sizeof_u_char(nexus_buffer_t *buffer, int count)
{
    XDR_SIZEOF_CHARS(unsigned char,count);
}




/*
 * tcp_xdr_put_*()
 *
 * Put 'count' elements, starting at 'data', of the given type,
 * into 'buffer'.
 */


static void tcp_xdr_put_float(nexus_buffer_t *buf, float *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(float),
		       xdr_float);
    NexusAssert2((xdrrc), ("tcp_xdr_put_float(): xdr_vector failed\n"));
}

static void tcp_xdr_put_double(nexus_buffer_t *buf, double *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(double),
		       xdr_double);
    NexusAssert2((xdrrc), ("tcp_xdr_put_double(): xdr_vector failed\n"));
}

static void tcp_xdr_put_short(nexus_buffer_t *buf, short *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(short),
		       xdr_short);
    NexusAssert2((xdrrc), ("tcp_xdr_put_short(): xdr_vector failed\n"));
}

static void tcp_xdr_put_u_short(nexus_buffer_t *buf, unsigned short *data,
			    int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(unsigned short),
		       xdr_u_short);
    NexusAssert2((xdrrc), ("tcp_xdr_put_u_short(): xdr_vector failed\n"));
}

static void tcp_xdr_put_int(nexus_buffer_t *buf, int *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(int),
		       xdr_int);
    NexusAssert2((xdrrc), ("tcp_xdr_put_int(): xdr_vector failed\n"));
}

static void tcp_xdr_put_u_int(nexus_buffer_t *buf, unsigned int *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(unsigned int),
		       xdr_u_int);
    NexusAssert2((xdrrc), ("tcp_xdr_put_u_int(): xdr_vector failed\n"));
}

static void tcp_xdr_put_long(nexus_buffer_t *buf, long *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(long),
		       xdr_long);
    NexusAssert2((xdrrc), ("tcp_xdr_put_long(): xdr_vector failed\n"));
}

static void tcp_xdr_put_u_long(nexus_buffer_t *buf, unsigned long *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)data,
		       count, sizeof(unsigned long),
		       xdr_u_long);
    NexusAssert2((xdrrc), ("tcp_xdr_put_u_long(): xdr_vector failed\n"));
}

static void tcp_xdr_put_char(nexus_buffer_t *buf, char *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_bytes(&(tcp_buf->xdrs), &data, 
		      (unsigned int *)&count, (unsigned int)(count));
    NexusAssert2((xdrrc), ("tcp_xdr_put_char(): xdr_bytes failed\n"));
}

static void tcp_xdr_put_u_char(nexus_buffer_t *buf, unsigned char *data, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_bytes(&(tcp_buf->xdrs), (char **)&data,
		      (unsigned int *)&count, (unsigned int)(count));
    NexusAssert2((xdrrc), ("tcp_xdr_put_u_char(): xdr_bytes failed\n"));
}



/*
 * tcp_xdr_get_*()
 *
 * Get 'count' elements of the given type from 'buffer' and store
 * them into 'data'.
 */



static void tcp_xdr_get_float(nexus_buffer_t *buf, float *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(float),
		       xdr_float);
    NexusAssert2((xdrrc), ("tcp_xdr_get_float(): xdr_vector failed\n"));
}

static void tcp_xdr_get_double(nexus_buffer_t *buf, double *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(double),
		       xdr_double);
    NexusAssert2((xdrrc), ("tcp_xdr_get_double(): xdr_vector failed\n"));
}

static void tcp_xdr_get_short(nexus_buffer_t *buf, short *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(short),
		       xdr_short);
    NexusAssert2((xdrrc), ("tcp_xdr_get_short(): xdr_vector failed\n"));
}

static void tcp_xdr_get_u_short(nexus_buffer_t *buf, unsigned short *dest,
			    int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(unsigned short),
		       xdr_u_short);
    NexusAssert2((xdrrc), ("tcp_xdr_get_u_short(): xdr_vector failed\n"));
}

static void tcp_xdr_get_int(nexus_buffer_t *buf, int *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(int),
		       xdr_int);
    NexusAssert2((xdrrc), ("tcp_xdr_get_int(): xdr_vector failed\n"));
}

static void tcp_xdr_get_u_int(nexus_buffer_t *buf, unsigned int *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(unsigned int),
		       xdr_u_int);
    NexusAssert2((xdrrc), ("tcp_xdr_get_u_int(): xdr_vector failed\n"));
}

static void tcp_xdr_get_long(nexus_buffer_t *buf, long *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(long),
		       xdr_u_long);
    NexusAssert2((xdrrc), ("tcp_xdr_get_long(): xdr_vector failed\n"));
}

static void tcp_xdr_get_u_long(nexus_buffer_t *buf, unsigned long *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(unsigned long),
		       xdr_u_long);
    NexusAssert2((xdrrc), ("tcp_xdr_get_u_long(): xdr_vector failed\n"));
}

static void tcp_xdr_get_char(nexus_buffer_t *buf, char *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_bytes(&(tcp_buf->xdrs), &dest, 
		      (unsigned int *)&count, (unsigned int)count);
    NexusAssert2((xdrrc), ("tcp_xdr_get_char(): xdr_bytes failed\n"));
}

static void tcp_xdr_get_u_char(nexus_buffer_t *buf, unsigned char *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_bytes(&(tcp_buf->xdrs), (char **)&dest,
		      (unsigned int *)&count, (unsigned int)count);
    NexusAssert2((xdrrc), ("tcp_xdr_get_u_char(): xdr_bytes failed\n"));
}



/*
 * tcp_xdr_get_stashed_*()
 *
 * Get 'count' elements of the given type from the stashed 'buffer' and store
 * them into 'data'.
 */



static void tcp_xdr_get_stashed_float(nexus_stashed_buffer_t *buf,
				  float *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
	       count, sizeof(float),
	       xdr_float);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_float(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_double(nexus_stashed_buffer_t *buf,
				   double *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(double),
		       xdr_double);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_double(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_short(nexus_stashed_buffer_t *buf,
				  short *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(short),
		       xdr_short);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_short(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_u_short(nexus_stashed_buffer_t *buf,
				    unsigned short *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(unsigned short),
		       xdr_u_short);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_u_short(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_int(nexus_stashed_buffer_t *buf,
				int *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(int),
		       xdr_int);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_int(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_u_int(nexus_stashed_buffer_t *buf,
				  unsigned int *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(unsigned int),
		       xdr_u_int);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_u_int(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_long(nexus_stashed_buffer_t *buf,
				 long *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(long),
		       xdr_long);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_long(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_u_long(nexus_stashed_buffer_t *buf,
				   unsigned long *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_vector(&(tcp_buf->xdrs), (char *)dest,
		       count, sizeof(unsigned long),
		       xdr_u_long);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_u_long(): xdr_vector failed\n"));
}

static void tcp_xdr_get_stashed_char(nexus_stashed_buffer_t *buf,
				 char *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_bytes(&(tcp_buf->xdrs), &dest, 
		      (unsigned int *)&count, (unsigned int)count);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_char(): xdr_bytes failed\n"));
}

static void tcp_xdr_get_stashed_u_char(nexus_stashed_buffer_t *buf,
				   unsigned char *dest, int count)
{
    tcp_buffer_t *tcp_buf = (tcp_buffer_t *) *buf;
    xdrrc = xdr_bytes(&(tcp_buf->xdrs), (char **)&dest,
		      (unsigned int *)&count, (unsigned int)count);
    NexusAssert2((xdrrc), ("tcp_xdr_get_stashed_u_char(): xdr_bytes failed\n"));
}


/*********************************************************************
 *	Attachment code
 *********************************************************************/

/*
 * nexus_allow_attach()
 *
 * Allow other nexus clients to attach this one.
 *
 * If *port==0, then a port number is allocated and returned in *port.
 * If *port!=0, then this port number will be used.
 *
 * '*host' is set to point to a string that contains the hostname
 * that should be used.  This string should _not_ be freed.
 *
 * This function may be called multiple times to allow multiple ports to
 * be used for attachment.
 *
 * Return:	0 on success
 *		non-0 on failure:
 *			1: the port is in use
 */
int nexus_allow_attach(unsigned short *port,
		       char **host,
		       int (*approval_func)(char *url,
					    nexus_global_pointer_t *gp)
		       )
{
    int fd;
    tcp_listener_t *listener;
    
    if ((fd = setup_listener(port)) < 0)
    {
	if (fd == -1)
	{
	    /* The port is in use */
	    return(1);
	}
	else
	{
	    tcp_fatal("nexus_allow_attach(): setup_listener() failed\n");
	}
    }
    
    listener = construct_listener(fd, *port, NEXUS_FALSE);
    listener->approval_func = approval_func;
    add_proto_with_fd((void *) listener, fd);

    *host = tcp_local_host;

    return(0);
} /* nexus_allow_attach() */


/*
 * nexus_disallow_attach()
 *
 * Disallow other nexus clients to attach this one using the passed 'port'.
 */
void nexus_disallow_attach(unsigned short port)
{
    tcp_listener_t *listener;

    for (listener = all_listeners;
	 listener && listener->port != port;
	 listener = listener->next)
	;
    if (listener)
    {
	close_listener(listener);
    }
} /* nexus_disallow_attach() */


/*
 * nexus_attach_with_url()
 *
 * Use 'url', of the form "x-nexus://<host>:<port>/...", and
 * use it to attach this nexus client to the nexus server
 * specified by this url.
 *
 * Return:	0 if successful, and fill in 'gp' with a
 *			global pointer to the nexus server.
 *		non-0 on failure:
 *			1: invalid url
 *			2: failed to connect
 */
int nexus_attach_with_url(char *url, nexus_global_pointer_t *gp)
{
    char *host;
    unsigned short port;
    int fd;
    int url_length;
    int rc;

    if (nexus_split_nexus_url(url, &host, &port, NULL) != 0)
    {
	/* Invalid url */
	return(1);
    }

    fd = do_connect(host, port);
    nexus_split_nexus_url_free(&host, NULL);
    if (fd < 0)
    {
	/* Failed to connect */
	return(2);
    }

    /* Write the url length to the fd */
    url_length = strlen(url);
    url_length = NX_HTONL(url_length);
    _nx_write(fd, (char *) &url_length, 4);
	       
    /* Write the url to the fd */
    _nx_write(fd, url, url_length);

    /* Read the return code */
    _nx_read(fd, (char *) &rc, 4);
    rc = NX_NTOHL(rc);

    if (rc == 0)
    {
	_nx_read_global_pointer(fd, gp);
    }
    
    return (rc);
    
} /* nexus_attach_with_url() */


/*
 * accept_attachment()
 *
 * A client just attached through 'new_fd' file descriptor.
 * Read the url from this fd and decide if the connections should
 * be accepted.  If so, then pass it a valid global pointer
 * and close 'new_fd'.
 */
static void accept_attachment(tcp_listener_t *listener,
			      int new_fd)
{
    nexus_global_pointer_t gp;
    int rc;
    int rc_n;
    char *url;
    int url_length;
    
    /* Read the url length */
    _nx_read(new_fd, (char *) &url_length, 4);
    url_length = NX_NTOHL(url_length);
    nexus_debug_printf(1,("accept_attachment(): read url_length=%d\n", url_length));

    NexusMalloc(accept_attachment(),
		url,
		char *,
		url_length + 1);
    
    /* Read the url */
    _nx_read(new_fd, url, url_length);
    url[url_length] = '\0';
    nexus_debug_printf(1,("accept_attachment(): read url=%s\n", url));
    
    rc = 3;
    nexus_null_global_pointer(&gp);
    if (listener->approval_func)
    {
	nexus_debug_printf(1,("accept_attachment(): calling approval_func\n"));
	rc = (*listener->approval_func)(url, &gp);
	nexus_debug_printf(1,("accept_attachment(): approval_func returned %d\n", rc));
    }
    if (nexus_is_null_global_pointer(&gp))
    {
	rc = 4;
    }
    
    /* Write the return code */
    rc_n = NX_HTONL(rc);
    _nx_write(new_fd, (char *) &rc_n, 4);
    
    if (rc == 0)
    {
	/* Write the global pointer */
	nexus_debug_printf(1,("accept_attachment(): writing global pointer\n"));
	_nx_write_global_pointer(new_fd, &gp);
	nexus_debug_printf(1,("accept_attachment(): done writing global pointer\n"));
    }
    
    NexusFree(url);

    close(new_fd);
    
} /* accept_attachment() */


/*
 * nexus_split_nexus_url()
 *
 * Take 'url', of the form:
 *	x-nexus://<host>:<port>/specifier1/specifier2/...
 * and split out the various components, and fill them into
 * the 'host', 'port', and 'specifiers' arguments.
 *
 * The 'host' is pointer to a string with the hostname.
 * The 'specifiers' is an array of character pointers, pointing
 * to the list of specifiers.
 *
 * 'host', 'port', and/or 'specifiers' may be NULL pointers,
 * indicating that the caller does not care about those fields.
 *
 * The values returned in 'host' and 'specifiers' should be freed
 * by calling nexus_split_nexus_url_free(host,specifiers).
 *
 * Return:	0 if successful
 *		non-0 if it is an invalid url
 */
int nexus_split_nexus_url(char *url,
			  char **host,
			  unsigned short *port,
			  char ***specifiers)
{
    char *start;
    char *end;
    char *specifier_start;
    char *s;
    int length;
    char tmp[16];
    int i;
    int n_specifiers;
    int done;

    if (host)
    {
	*host = (char *) NULL;
    }
    if (specifiers)
    {
	*specifiers = (char **) NULL;
    }

    /*
     * Check the url for a valid prefix
     */
    if (strncmp(url, "x-nexus://", 10) != 0)
    {
	/* invalid url: invalid prefix */
	goto badurl;
    }
    start = url + 10;

    /*
     * Pull the host out of the url
     */
    if ((end = strchr(start, ':')) == (char *) NULL)
    {
	/* invalid url: no ':' after host */
	goto badurl;
    }
    if (host)
    {
	length = end - start;
	NexusMalloc(nexus_attach_with_url(),
		    *host,
		    char *,
		    length + 1);
	strncpy(*host, start, length);
	(*host)[length] = '\0';
    }

    /* Pull the port out of the url */
    start = end + 1;
    if ((end = strchr(start, '/')) == (char *) NULL)
    {
	end = strchr(start, '\0');
    }
    if (   ((end - start) > 15)
	|| ((end - start) == 0) )
    {
	/*
	 * invalid url: the port section is either empty or is
	 *		way to big to be a valid port
	 */
	goto badurl;
    }
    for (i = 0; start+i < end; i++)
    {
	tmp[i] = *(start + i);
	if (!isdigit(tmp[i]))
	{
	    /* invalid url: port section contains a non-digit */
	    goto badurl;
	}
    }
    tmp[i] = '\0';
    if (port)
    {
	*port = (short) atoi(tmp);
    }

    if (*end == '\0')
    {
	/* There is no specifiers section, so we're done */
	return(0);
    }
    
    if (!specifiers)
    {
	/* We don't care about the specifiers, so we're done */
	return(0);
    }

    /* Count the number of specifiers */
    n_specifiers = 0;
    start = end + 1;
    specifier_start = start;
    for (done = NEXUS_FALSE; !done; )
    {
	end = strchr(start, '/');
	if (end == (char *) NULL)
	{
	    /* We've hit the end of the url, without a trailing / */
	    n_specifiers++;
	    done = NEXUS_TRUE;
	}
	else if (*(end - 1) == '\\')
	{
	    /* This is an escaped slash, so keep looking */
	    start = end + 1;
	}
	else
	{
	    /* We found the end of a specifier */
	    start = end + 1;
	    n_specifiers++;
	    if (*start == '\0')
	    {
		/* We've hit the end of the url, with a trailing / */
		done = NEXUS_TRUE;
	    }
	}
    }

    if (n_specifiers == 0)
    {
	/* No specifiers, so we're done */
	return(0);
    }
    
    /* Allocate the specifiers array */
    NexusMalloc(nexus_split_nexus_url(),
		*specifiers,
		char **,
		(sizeof(char *) * (n_specifiers + 1)) );
    for (i = 0; i <= n_specifiers; i++)
    {
	(*specifiers)[i] = (char *) NULL;
    }

    /* Split out the specifiers */
    start = specifier_start;
    i = 0;
    for (done = NEXUS_FALSE; !done; )
    {
	end = strchr(start, '/');
	if (end == (char *) NULL)
	{
	    /* We've hit the end of the url, without a trailing / */
	    length = strlen(specifier_start);
	    NexusMalloc(nexus_split_nexus_url(),
			s,
			char *,
			length + 1);
	    strncpy(s, specifier_start, length);
	    s[length] = '\0';
	    (*specifiers)[i] = s;
	    i++;
	    
	    done = NEXUS_TRUE;
	}
	else if (*(end - 1) == '\\')
	{
	    /* This is an escaped slash, so keep looking */
	    start = end + 1;
	}
	else
	{
	    /* We found the end of a specifier */
	    length = end - specifier_start;
	    NexusMalloc(nexus_split_nexus_url(),
			s,
			char *,
			length + 1);
	    strncpy(s, specifier_start, length);
	    s[length] = '\0';
	    (*specifiers)[i] = s;
	    i++;
	    
	    start = end + 1;
	    specifier_start = start;

	    if (*start == '\0')
	    {
		/* We've hit the end of the url, with a trailing / */
		done = NEXUS_TRUE;
	    }
	}
    }

    return(0);

 badurl:
    nexus_split_nexus_url_free(host, specifiers);
    return(1);
    
} /* nexus_split_nexus_url() */


/*
 * nexus_split_nexus_url_free()
 *
 * Free the 'host' and 'specifiers' values that were
 * previously returned by nexus_split_nexus_url().
 */
void nexus_split_nexus_url_free(char **host,
				char ***specifiers)
{
    int i;
    
    if (host && *host)
    {
	NexusFree(*host);
    }

    if (specifiers && *specifiers)
    {
	for (i = 0; (*specifiers)[i] != (char *) NULL; i++)
	{
	    NexusFree((*specifiers)[i]);
	}
	NexusFree(*specifiers);
    }
} /* nexus_split_nexus_url_free() */
