/*
 * st_mpl.c
 *
 * Remote node startup using IBM SP2 MPL
 */

static char *rcsid = "$Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/st_mpl.c,v 1.7 1995/05/23 18:24:10 patton Exp $";

#include "internal.h"
#include "mpproto.h"

#ifndef NEXUS_MPL_NODE_NAME
#define NEXUS_MPL_NODE_NAME "sp2_mpl"
#endif

#define INITIAL_MSG_TYPE   67
#define _NX_MP_ACQUIRE_NODE_HANDLER_HASH 786

static int my_node;
static int n_nodes;
static nexus_bool_t nodes_started = NEXUS_FALSE;

static nexus_bool_t	st_mpl_preinit(nexus_startup_node_t **nl_head,
			       nexus_startup_node_t **nl_tail);
static char *	st_mpl_get_master_node_name();
static void	st_mpl_start_node(nexus_startup_node_t *node,
				  unsigned long context,
				  unsigned long address,
				  unsigned long class_type,
				  int first_checkin_number);
#ifndef NEXUS_ARCH_STARTUP_IBMDS
static nexus_bool_t	st_mpl_start_context(char *executable_path,
				     int context_size,
				     unsigned long reply_context,
				     unsigned long reply_address,
				     unsigned long reply_class_type,
				     int checkin_number,
				     int *return_code,
				     nexus_context_t **new_local_context);
#endif /* !NEXUS_ARCH_STARTUP_IBMDS */
static void     st_mpl_register_handlers();
static void _nx_mp_acquire_node_handler( void * address,
					 nexus_buffer_t *buffer);

static nexus_startup_funcs_t st_mpl_funcs =
{
    st_mpl_preinit,
    st_mpl_get_master_node_name,
    NULL /* st_mpl_init */,
    NULL /* st_mpl_shutdown */,
    NULL /* st_mpl_abort */,
    st_mpl_start_node,
#ifdef NEXUS_ARCH_STARTUP_IBMDS    
    NULL /* st_mpl_start_context */,
#else    
    st_mpl_start_context,
#endif
    st_mpl_register_handlers,
};

static nexus_handler_t mpl_startup_handlers[] =
{ \
    {"_nx_mp_acquire_node_handler",
     _NX_MP_ACQUIRE_NODE_HANDLER_HASH,
     NEXUS_HANDLER_TYPE_NON_THREADED,
     (nexus_handler_func_t) _nx_mp_acquire_node_handler},
  {(char *) NULL,
   0,
   NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t) NULL},
};

/*
 * _nx_st_mpl_info()
 *
 * Return the function table for this module.
 */
void *_nx_st_mpl_info()
{
    return ((void *) (&st_mpl_funcs));
}


/*
 * st_mpl_preinit()
 */
static nexus_bool_t st_mpl_preinit(nexus_startup_node_t **nl_head,
			     nexus_startup_node_t **nl_tail)
{
    int source, type;
    size_t n_bytes, n_bytes_out;

    *nl_head = *nl_tail = (nexus_startup_node_t *) NULL;
    
    /*
     * TODO: Need a way to sense if MPL is being used
     */
    mpc_environ(&n_nodes, &my_node);

    if (n_nodes == 1)
	return;
    
    if (my_node == 0)
    {
	/*
	 * This is the master node.
	 * Setup a nexus_startup_node_t structure for the rest
	 * of the MPL nodes.  That will cause st_mpl_start_node
	 * to be called during Nexus initialization.
	 */
	nexus_startup_node_t *initial_node;
	NexusMalloc(st_mpl_preinit(),
		    initial_node,
		    nexus_startup_node_t *,
		    sizeof(nexus_startup_node_t));
	initial_node->name = _nx_copy_string(NEXUS_MPL_NODE_NAME);
	initial_node->number = 1;
	initial_node->count = n_nodes - 1;
	initial_node->startup_module = _nx_copy_string("mpl");
	initial_node->startup_funcs = &st_mpl_funcs;
	initial_node->next = (nexus_startup_node_t *) NULL;
	initial_node->directory_path = (char *) NULL;
	initial_node->executable_path = (char *) NULL;
	initial_node->st_ptr = NULL;

	*nl_head = *nl_tail = initial_node;
    }
    else
    {
	/*
	 * This is a non-master node.
	 * Wait for a message from the master node that contains
	 * _nx_my_process_params for my node.
	 * That message is sent from st_mpl_start_node.
	 */
	n_bytes = NEXUS_MAX_COMMAND_LENGTH;
	NexusMalloc(st_mpl_preinit(),
		    _nx_my_process_params,
		    char *,
		    n_bytes);
	source = DONTCARE;
	type = DONTCARE;
	mpc_brecv(_nx_my_process_params,
		  n_bytes,
		  &source,
		  &type,
		  &n_bytes_out);

	if (   (source != 0)
	    && (type != INITIAL_MSG_TYPE) )
	{
	    fprintf(_nx_stdout,
		    "ERROR: MPL node %d expected initial message\n");
	    mpc_stopall(1);
	    exit(1);
	}
	nodes_started = NEXUS_TRUE;
    }

    return(NEXUS_FALSE);
} /* st_mpl_preinit() */


/*
 * st_mpl_get_master_node_name()
 */
static char *st_mpl_get_master_node_name()
{
    return (_nx_copy_string(NEXUS_MPL_NODE_NAME));
} /* st_mpl_get_master_node_name() */


/*
 * _nx_start_nodes()
 */
static void st_mpl_start_node(nexus_startup_node_t *node,
			      unsigned long context,
			      unsigned long address,
			      unsigned long class_type,
			      int first_checkin_number)
{
    int i;
    char cmd[NEXUS_MAX_COMMAND_LENGTH], cmd2[NEXUS_MAX_COMMAND_LENGTH];
    int cmd2_len;
    int msglen;
    int type = INITIAL_MSG_TYPE;

    /* Cannot start dynamic nodes, but must call NexusAcquiredAsNode
     * on requested node
     */
    if (nodes_started == NEXUS_TRUE) {
	/* hack together a global pointer to the node we want
	 * to acquire.  It should be within the current
	 * set of nodes that we have.  Send an RSR to that node
	 * to invoke a handler which will return a "good" global
	 * pointer to itself and also call NexusAcquiredAsNode
	 */
	nexus_global_pointer_t mygp;
	nexus_global_pointer_t remotegp;
	nexus_proto_t *remote_proto;
	nexus_mi_proto_t *remote_mi_proto;
	nexus_proto_funcs_t  *my_mp_proto_funcs;

	nexus_buffer_t buffer;
	int buf_len = 0;
	int n_elements = 0;

	nexus_global_pointer( &mygp, (void *) address );

	my_mp_proto_funcs = (nexus_proto_funcs_t *)_nx_pr_mpl_info();

	remote_mi_proto = (*my_mp_proto_funcs->get_my_mi_proto)();

	remote_mi_proto->ints[0] = node->number;
	remote_mi_proto->next = NULL;

	(*my_mp_proto_funcs->construct_from_mi_proto)(&remote_proto,
						      remote_mi_proto);

	remotegp.proto = remote_proto;
	remotegp.all_protos = remote_mi_proto;
	remotegp.context = 0;
	remotegp.address = address;
	remotegp.class_type = mygp.class_type;
	
	nexus_init_remote_service_request( &buffer,
					   &remotegp,
					   "_nx_mp_acquire_node_handler",
					   _NX_MP_ACQUIRE_NODE_HANDLER_HASH);

	buf_len = nexus_sizeof_global_pointer(&buffer, &mygp, 1,
					      &n_elements);

	nexus_set_buffer_size(&buffer, (buf_len + nexus_sizeof_int(&buffer,1)),
			      (n_elements + 1));
	nexus_put_global_pointer(&buffer, &mygp, 1);
	nexus_put_int(&buffer, &first_checkin_number, 1);

	nexus_send_remote_service_request(&buffer);

	nexus_destroy_global_pointer(&mygp);
	nexus_destroy_global_pointer(&remotegp);

	return;
    }
    
    /* Start the nodes on the machines listed in arg_host_list */
    for (i = 0; i < node->count; i++)
    {
	/* Get all arguments that need to go to this process */
	_nx_new_process_params(cmd2, NEXUS_MAX_COMMAND_LENGTH);
	cmd2_len = strlen(cmd2);
	_nx_proto_get_creator_gp_params(cmd2 + cmd2_len,
					NEXUS_MAX_COMMAND_LENGTH - cmd2_len,
					node->name,
					node->number + i,
					"mpl",
					context,
					address,
					class_type);
	
	nexus_stdio_lock();
	sprintf(cmd, "%s -nx_node %s %d y -nx_node_id %d -nx_checkin %d %s %s",
		_nx_get_package_id_start(),
		node->name,
		node->number + i,
		node->id + i,
		first_checkin_number + i,
		cmd2,
		_nx_get_package_id_end() );
	nexus_stdio_unlock();
	
	msglen = strlen(cmd) + 1;
	mpc_bsend(cmd, msglen, node->number + i, type);

	nodes_started = NEXUS_TRUE;
    }
} /* _nx_start_nodes() */


#ifndef NEXUS_ARCH_STARTUP_IBMDS
/*
 * st_mpl_start_context()
 */
static nexus_bool_t st_mpl_start_context(char *executable_path,
				   int context_size,
				   unsigned long reply_context,
				   unsigned long reply_address,
				   unsigned long reply_class_type,
				   int checkin_number,
				   int *return_code,
				   nexus_context_t **new_local_context)
{
    /*
     * TODO: Should get rid of this fatal error, so that other
     * startup and protocol modules can work simultaneously.
     */
    nexus_fatal("st_mpl_start_context(): ERROR: Cannot create other processes with MPL\n");
    
    return (NEXUS_FALSE);
    
} /* st_mpl_start_context() */
#endif /* !NEXUS_ARCH_STARTUP_IBMDS */

static void st_mpl_register_handlers()
{
    nexus_register_handlers(mpl_startup_handlers);
}

static void _nx_mp_acquire_node_handler(void *address,
					nexus_buffer_t *buffer)
{
    nexus_global_pointer_t replygp;
    int first_checkin_number;
    nexus_node_t my_fake_node;

    nexus_debug_printf(2, ("_nx_mp_acquire_node_handler(): this node being acquired\n"));

    my_fake_node = _nx_my_node;
    my_fake_node.return_code = NEXUS_NODE_OLD_DIFF_EXE;

    nexus_global_pointer(&(my_fake_node.gp), NULL);

    nexus_get_global_pointer(buffer, &replygp, 1);
    nexus_get_int(buffer, &first_checkin_number, 1);

    _nx_node_checkin( &my_fake_node, &replygp, first_checkin_number
#ifdef NEXUS_PROFILE
		      , NEXUS_TRUE
#endif
		      );

    nexus_destroy_global_pointer(&replygp);
    nexus_destroy_global_pointer(&(my_fake_node.gp));
    nexus_debug_printf(2, ("_nx_mp_acquire_node_handler(): returning\n"));

}
