/*
 * types.h
 *
 * rcsid = "$Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/types.h,v 1.33 1995/06/06 23:52:54 tuecke Exp $"
 */

#ifndef _NEXUS_INCLUDE_TYPES_H
#define _NEXUS_INCLUDE_TYPES_H

/*
 * nexus_mi_proto_t
 *
 * This structure contains a "machine independent" representation
 * of a protocol.  This is passed as the node part of a global
 * pointer, and is used to decode protocols for accessing nodes.
 * Protocols need to be represented in a machine independent
 * form so that they can be passed through nodes that may
 * speak a different set of protocols.  (That is, a particular
 * protocol may not be used by a particular node, but that node
 * must be able to store and pass on the protocol.)
 *
 * A nexus_mi_proto_t is always passed in a message as:
 *	1) The protocol type
 *	2) The number of integers in the protocol
 *	3) The number of character stings in the protocol
 *	4) The integers
 *	5) The lengths of each character string (not including
 *		NULL termination)
 *	6) The character strings (not necessarily NULL terminated)
 */
struct _nexus_mi_proto_t
{
    int				proto_type;
    int				n_ints;
    int				n_strings;
    int *			ints;
    int *			string_lengths;
    char **			strings;
    struct _nexus_mi_proto_t *	next;
};

#define MiProtoInts(mi_proto, n)		mi_proto->ints[n]
#define MiProtoStringLengths(mi_proto, n)	mi_proto->string_lengths[n]
#define MiProtoStrings(mi_proto, n)		mi_proto->strings[n]


#ifdef NEXUS_PROFILE

#ifndef NEXUS_PROFILE_RSR_HASH_TABLE_SIZE
#define NEXUS_PROFILE_RSR_HASH_TABLE_SIZE	16
#endif

/*
 * nexus_profile_rsr_record_t
 *
 * This is used to keep cumulative totals of handler invocations
 */
typedef struct _nexus_profile_rsr_record_t
{
    int						node_id;
    int						context_id;
    int						count;
    int						size;
    struct _nexus_profile_rsr_record_t *	next;
} nexus_profile_rsr_record_t;

#endif /* NEXUS_PROFILE */


/*
 * nexus_handler_record_t
 *
 * One record in the handler table...
 */
typedef struct _nexus_handler_record_t
{
    char *				name;
    int					id;
    nexus_handler_type_t		type;
    nexus_handler_func_t		func;
    struct _nexus_handler_record_t *	next;
#ifdef NEXUS_PROFILE
    nexus_profile_rsr_record_t **	profile_table;
#endif    
} nexus_handler_record_t;


/*
 * nexus_segment_t
 *
 * Holds one data segment in a context.
 */
typedef struct _nexus_segment_t
{
    void *			data;
    unsigned int		size;
    struct _nexus_context_t *	context;
    struct _nexus_segment_t *	prev;
    struct _nexus_segment_t *	next;
} nexus_segment_t;

#ifdef NEXUS_ALIGN
#define NEXUS_SEGMENT_HEADER_SIZE \
    (sizeof(nexus_segment_t) \
     + ((NEXUS_ALIGN - (sizeof(nexus_segment_t) % NEXUS_ALIGN)) % NEXUS_ALIGN))
#else  /* NEXUS_ALIGN */
#define NEXUS_SEGMENT_HEADER_SIZE \
    sizeof(nexus_segment_t)
#endif /* NEXUS_ALIGN */


/*
 * nexus_context_t
 *
 * Hold context info.
 *
 * The segment_list is a circular, doubly linked list, with a dummy
 * first segment.
 */
typedef struct _nexus_context_t
{
#ifndef NEXUS_LITE
    nexus_mutex_t		mutex;
#endif /* NEXUS_LITE */
    int				id;
    void *                      handle;
    int                         (*entry_point)(void *);
    void                        (*switchingfunc)(void *);
    int                         (*destructofunc)(struct _nexus_context_t *);
    int                         (*pfnNexusBoot)();
    void                        (*pfnNexusExit)();
    void                        (*pfnNexusUnknownHandler)(void *,
							  nexus_buffer_t *,
							  char *, int );
    int                         (*pfnNexusAcquiredAsNode)();
    int				n_segments;
    struct _nexus_segment_t *	segment_list;
    void *			initial_segment;
    nexus_handler_record_t **	handler_table;
    nexus_bool_t		is_node;
    struct _nexus_context_t *	context_list_next;
#ifdef NEXUS_PROFILE
    int				snapshot_id;
    int				rsr_profile_count;
#ifndef NEXUS_LITE
    nexus_global_pointer_t	master_gp;
    nexus_bool_t		waiting_on_master_gp;
    nexus_cond_t		cond;
#endif /* NEXUS_LITE */
#endif /* NEXUS_PROFILE */
} nexus_context_t;


#ifndef NEXUS_LITE
/*
 * nexus_new_nodes_monitor_t
 *
 * This monitor is used to keep track of new processes as they checkin.
 */
typedef struct _nexus_new_nodes_monitor_t
{
    nexus_mutex_t	mutex;
    nexus_cond_t	cond;
    int			total;
    int			checked_in;
    nexus_node_t *	nodes;
} nexus_new_nodes_monitor_t;
#endif /* NEXUS_LITE */


#ifndef NEXUS_LITE

typedef struct _nexus_startup_node_t nexus_startup_node_t;
typedef struct _nexus_startup_funcs_t nexus_startup_funcs_t;

/*
 * nexus_startup_funcs_t
 *
 * This is the function indirection table used by startup modules.
 */
struct _nexus_startup_funcs_t
{
    nexus_bool_t	(*preinit)(nexus_startup_node_t **nl_head,
				   nexus_startup_node_t **nl_tail);
    char *		(*get_master_node_name)();
    void		(*init)();
    void		(*shutdown)(nexus_bool_t shutdown_others);
    void		(*abort)();
    void		(*start_node)(nexus_startup_node_t *node,
				      unsigned long context,
				      unsigned long address,
				      unsigned long class_type,
				      int first_checkin_number);
    nexus_bool_t	(*start_context)(char *executable_path,
					 int context_size,
					 unsigned long reply_context,
					 unsigned long reply_address,
					 unsigned long reply_class_type,
					 int checkin_number,
					 int *return_code,
					 nexus_context_t **new_local_context);
    void                (*register_handlers)();
};


/*
 * nexus_startup_node_t.
 *
 * Startup nodes structure.  st_iface.c creates this structure and
 * passes it to the start_node function of the startup module, to
 * tell it what start a node or set of nodes.
 *
 * If 'count' is > 1, then 'number' and 'id' fields are the _first_
 * of a continguous set of numbers and ids that the set of nodes should use.  
 */
struct _nexus_startup_node_t
{
    char *			name;
    int				number;
    int				count;
    int				id;
    char *			directory_path;
    char *			executable_path;
    char *			startup_module;
    nexus_startup_funcs_t *	startup_funcs;
    nexus_startup_node_t *	next;
    void *			st_ptr;
};

#endif /* NEXUS_LITE */


typedef struct _nexus_database_funcs_t nexus_database_funcs_t;

/* 
 * This is the function indirection table that gets loaded in
 * db_iface
 */
struct _nexus_database_funcs_t
{
    nexus_bool_t                (*init)(int *argc, char ***argv);
    void                        (*usage)();
    int                         (*new_params)(char *buf,
					     int size);
    char *                      (*lookup)(char *node_name,
				 	 int node_number,
					 char *key);
    void                        (*shutdown)();
    void                        (*abort)();
};

typedef struct _nexus_list_t nexus_list_t;

/*
 * This is a generic linked list
 */
struct _nexus_list_t
{
    void *                    value;
    struct _nexus_list_t *    next;
};

typedef struct _nexus_db_hash_entry_t nexus_db_hash_entry_t;

/*
 * This is the hash table stuff for database/startup file lookup
 * values.
 */
struct _nexus_db_hash_entry_t
{
    char *name;
    int number;
    int count;
    nexus_list_t *attr;
    struct _nexus_db_hash_entry_t *next;
};

#endif /* _NEXUS_INCLUDE_TYPES_H */
